COMMENT

A "simple" implementation of the Izhikevich neuron with AMPA, NMDA,
GABA_A, and GABA_B receptor dynamics. Most parameter values are taken
from the appendix to
   Izhikevich EM, Edelman GM (2008).
   "Large-scale model of mammalian thalamocortical systems." 
   PNAS 105(9) 3593-3598.

Example usage (in Python):
  from neuron import h
  dummycell = h.Section() # Since Izhi is a point process, it needs to be in a section
  cells = [0,1] # Initialize cells
  for c in cells: cells[c] = h.Izhi(0) # Create two new Izhikevich cells
  connection = h.NetCon(cells[0], cells[1]) # Connect them
  
See also (if available) izhi.py, a Python wrapper for different types of
Izhikevich neuron.

Version: 2013sep12 by Cliff Kerr (cliffk@neurosim.downstate.edu)

ENDCOMMENT


: Declare name of object and variables
NEURON {
  POINT_PROCESS Izhi
  RANGE C, k, vr, vt, vpeak, a, b, c, d, tauAMPA, tauNMDA, tauGABAA, tauGABAB, xloc, yloc, zloc, flag
}


: Specify units that have physiological interpretations (NB: ms is already declared)
UNITS {
	(mV) = (millivolt)
	(uM) = (micrometer)
}


: Parameters from Izhikevich 2008, PNAS for P2/3 cell, no dendrites
PARAMETER {
  C = 100 : Capacitance
  k = 3
  vr = -60 (mV) : Resting membrane potential
  vt = -50 (mV) : Membrane threhsold
  vpeak = 30 (mV) : Peak voltage
  a = 0.01
  b = 5
  c = -60
  d = 400
  tauAMPA = 5 (ms) : Receptor time constant, AMPA
  tauNMDA = 150 (ms) : Receptor time constant, NMDA
  tauGABAA = 6 (ms) : Receptor time constant, GABAA
  tauGABAB = 150 (ms) : Receptor time constant, GABAB
  xloc = 0 (uM) : Spatial location, x
  yloc = 0 (uM) : Spatial location, y
  zloc = 0 (uM) : Spatial location, z
  flag = -1 : An unused parameter for storing any kind of flag desired -- e.g. cell type, active/inactive...
}


: Variables used for internal calculations
ASSIGNED {
  t0 : Previous time
  factor : Voltage factor used for calculating the current
}


: State variables
STATE {
  V (mV) : Membrane voltage
  u (mV) : Slow current/recovery variable
  gAMPA : AMPA conductance
  gNMDA : NMDA conductance
  gGABAA : GABAA conductance
  gGABAB : GABAB conductance
  I : Total current
  delta : Time step
}


: Initial conditions -- from, like, every Izhikevich paper
INITIAL {
  V = vr
  u = 0.2*V
  t0 = t
  gAMPA = 0
  gNMDA = 0
  gGABAA = 0
  gGABAB = 0
  I = 0
  delta = 0
  net_send(0,1) : Required for the WATCH statement to be active
}

: Define neuron dynamics
BREAKPOINT {
  delta = t-t0 : Find time difference

  : Receptor dynamics
  gAMPA = gAMPA - delta*gAMPA/tauAMPA : Exponential decays
  gNMDA = gNMDA - delta*gNMDA/tauNMDA
  gGABAA = gGABAA - delta*gGABAA/tauGABAA
  gGABAB = gGABAB - delta*gGABAB/tauGABAB
  
  : Calculate current
  factor = ((V+80)/60)*((V+80)/60)
  I = gAMPA*(V-0) + gNMDA*factor/(1+factor)*(V-0) + gGABAA*(V+70) + gGABAB*(V+90)
  
  : Calculate neuronal dynamics; -I since I = -I_{syn}, which is really what I is as I've defined it above
  V = V + delta*(V + k*(V-vr)*(V-vt) - u - I)/C
  u = u + delta*a*(b*(V-vr)-u)

  t0=t : Reset last time so delta can be calculated in the next time step
}

: Input received
NET_RECEIVE (wAMPA, wNMDA, wGABAA, wGABAB) {  
  INITIAL { wAMPA=wAMPA wNMDA=wNMDA wGABAA=wGABAA wGABAB=wGABAB } : Insanely stupid but required, arrrgggh

  : Check if spike occurred
  if (flag == 1) { : Fake event from INITIAL block
    WATCH (V>vpeak) 2 : Check if threshold has ben crossed, and if so, set flag=2
  } 
  
  : Event created by WATCH statement -- i.e. threshold crossed
  else if (flag == 2) { 
    net_event(t) : Send spike event
    V = c : Reset voltage
    u = u+d : Reset recovery variable
  } 
  
  : Actual input, calculate receptor dynamics
  else { 
    gAMPA = gAMPA + wAMPA
    gNMDA = gNMDA + wNMDA
    gGABAA = gGABAA + wGABAA
    gGABAB = gGABAB + wGABAB
  }
  
}
