COMMENT
STDP + RL Weight Adjuster POINT_PROCESS
Spike Timing-Dependent Weight Adjuster with Reinforcement Learning
Last update: 7/16/13 (georgec)

original STDP code adapted from:
http://senselab.med.yale.edu/modeldb/showmodel.asp?model=64261&file=\bfstdp\stdwa_songabbott.mod
based on (Song and Abbott, 2001)
by Andrew Davison, UNIC, CNRS, 2003-2004

adapted to implement a "nearest-neighbor spike-interaction" model (see 
Scholarpedia article on STDP) that just looks at the last-seen pre- and 
post-synaptic spikes, and implementing a reinforcement learning algorithm based
on (Chadderdon et al., 2012):
http://www.plosone.org/article/info%3Adoi%2F10.1371%2Fjournal.pone.0047251
by George Chadderdon, SUNY Downstate Medical Center
ENDCOMMENT

NEURON {
    : Definition of mechanism
    POINT_PROCESS StdpRLWtAdj
    
    : LTP/LTD decay time constants (in ms) for the Hebbian (pre-before-post- 
    : synaptic spikes), and anti-Hebbian (post-before-pre-synaptic) cases. 
    : Typically, e.g. for excitatory connections, the Hebbian part of the curve 
    : triggers LTP and the anti-Hebbian part LTD.  For other cases, e.g. 
    : inhibitory connections, the anti-Hebbian part may trigger LTP and the 
    : Hebbian part LTD.  Sometimes the signs may even be the same for 
    : the two parts of the curve, depending on the location of the synapse.
    : See the Scholarpedia article on STDP (Diversity of STDP section):
    :   http://www.scholarpedia.org/article/Spike-timing_dependent_plasticity
    RANGE tau_Hebb_STDP, tau_antiHebb_STDP
    
    : Maximal adjustment (can be positive or negative) for Hebbian and 
    : anti-Hebbian cases (i.e., as inter-spike interval approaches zero).  This
    : should be set positive for LTP and negative for LTD.
    RANGE maxadj_Hebb, maxadj_antiHebb  
    
    : Maximum interval between pre- and post-synaptic events for an 
    : starting an eligibility trace.  There are separate ones for the Hebbian 
    : and anti-Hebbian events.
    RANGE maxeligwind_Hebb, maxeligwind_antiHebb
    
    : Use exponentially decaying eligibility traces?  If 0, then the eligibility
    : traces are binary, turning on at the beginning and completely off after 
    : time has passed corresponding to eliglength.  If exponential decay 
    : is turned on, then the eliglength values are the decay time constants.
    RANGE expdecay_elig
    
    : Length of the eligibility Hebbian and anti-Hebbian eligibility traces, 
    : or the decay time constants if the traces are decaying exponentials.
    RANGE eliglength_Hebb, eliglength_antiHebb  
    
    : Maximum gains to be applied to the reward or punishing signal by 
    : Hebbian and anti-Hebbian eligibility traces.  
    RANGE maxeliggain_Hebb, maxeliggain_antiHebb
    
    : The maximum weight for the synapse.
    RANGE wmax
    
    : Flag turning on "soft thresholding" for the maximal adjustment parameters.
    RANGE adj_soft_thresh
    
    : Pointer to the weight (in a NetCon object) to be adjusted.
    POINTER wsyn    
        
    : Flag for turning the weight adjustment on / off.
    RANGE on
    
    : Flag for turning STDP adjustment on / off.
    RANGE STDPon
    
    : Flag for turning RL adjustment on / off.
    RANGE RLon
 
     : Flags for turning Hebbian and anti-Hebbian eligibility traces on / off.
    RANGE elig_Hebb_on, elig_antiHebb_on
      
    : Flag for whether the weight pointed to must be negative.
    RANGE negweight
    
    : Flag for turning off prints of weight update events for debugging.
    RANGE verbose
            
    : Remembered times for last pre- and post-synaptic spikes.
    RANGE tlast_pre, tlast_post
    
    : Remembered times for Hebbian anti-Hebbian eligibility traces.
    RANGE tlast_Hebb_elig, tlast_antiHebb_elig
    
    : Interval between current time t and previous spike.
    RANGE interval
    
    : The calculated weight change.
    RANGE deltaw
    
    : New calculated weight.
    RANGE newwt
}

ASSIGNED {
    wsyn        
    tlast_pre   (ms)    
    tlast_post  (ms)   
    tlast_Hebb_elig   (ms)    
    tlast_antiHebb_elig  (ms)        
    interval    (ms)    
    deltaw
    newwt          
}

INITIAL {
    tlast_pre = -1                     : no spike yet
    tlast_post = -1                   : no spike yet
    tlast_Hebb_elig = -1        : no eligibility yet
    tlast_antiHebb_elig = -1  : no eligibility yet   
    interval = 0
    deltaw = 0
    newwt = 0
}

PARAMETER {
    tau_Hebb_STDP  = 10  (ms)   
    tau_antiHebb_STDP  = 10  (ms)    
    maxadj_Hebb = 1.0
    maxadj_antiHebb = -1.0
    maxeligwind_Hebb = 10 (ms)
    maxeligwind_antiHebb = 10 (ms)
    expdecay_elig = 0   : default to using binary eligibility traces
    eliglength_Hebb = 100 (ms)
    eliglength_antiHebb = 100 (ms)
    maxeliggain_Hebb = 1.0
    maxeliggain_antiHebb = -1.0
    wmax  = 15.0
    adj_soft_thresh = 0
    on = 1  
    STDPon = 1
    RLon = 1
    elig_Hebb_on  = 1
    elig_antiHebb_on = 0  : turn off anti-Hebbian eligibility traces by default
    negweight = 0
    verbose = 0
}

NET_RECEIVE (w) {
    : Default the weight change to 0.
     deltaw = 0.0
     
    : If we receive a non-negative weight value, we are receiving a pre-synaptic
    : spike (and thus need to check for an anti-Hebbian event, since the post-
    : synaptic weight must be earlier).
    if (w >= 0) {           
        : Get the interval.
        interval = tlast_post - t   : interval is negative
        
        : If we had a post-synaptic spike and a non-zero interval...
        if  ((tlast_post > -1) && (interval != 0)) {
            : If STDP learning is turned on...
            if (STDPon == 1) {
                : Use the anti-Hebbian decay to set the anti-Hebbian weight adjustment.
                deltaw = maxadj_antiHebb * exp(interval / tau_antiHebb_STDP)
                
               : If we have soft-thresholding on, apply it.
                if (adj_soft_thresh == 1) {
                    deltaw = soft_threshold(deltaw)
                }
                
                : Show weight update information if debugging on.       
                if (verbose > 0) {
                    printf("anti-Hebbian STDP event: t = %f ms; deltaw = %f\n",t,deltaw)
                }
            }
            
           : If RL and anti-Hebbian eligibility traces are turned on, and 
           : the interval falls within the maximum window for eligibility, 
           : remember the eligibilty trace start at the current time.
            if ((RLon == 1) && (elig_antiHebb_on == 1) &&
                (-interval <= maxeligwind_antiHebb)) {
                tlast_antiHebb_elig = t
            }
        }
        
        : Remember the current spike time for next NET_RECEIVE.
        tlast_pre = t       
        
    : Else, if we receive a negative weight value, we are receiving a post-
    : synaptic spike (and thus need to check for an anti-Hebbian event, since 
    : the post-synaptic weight must be earlier).        
    } else {                   
        : Get the interval.        
        interval = t - tlast_pre    : interval is positive
                     
        : If we had a pre-synaptic spike and a non-zero interval...
        if  ((tlast_pre > -1) && (interval != 0)) {
            : If STDP learning is turned on...
            if (STDPon == 1) {
                : Use the Hebbian decay to set the Hebbian weight adjustment.   
                deltaw = maxadj_Hebb * exp(-interval / tau_Hebb_STDP)
                          
                : If we have soft-thresholding on, apply it.
                if (adj_soft_thresh == 1) {
                    deltaw = soft_threshold(deltaw)
                }    
                
                : Show weight update information if debugging on.
                if (verbose > 0) {
                    printf("Hebbian STDP event: t = %f ms; deltaw = %f\n",t,deltaw)
                }
            }
            
           : If RL and Hebbian eligibility traces are turned on, and the 
           : interval falls within the maximum window for eligibility, remember 
           : the eligibilty trace start at the current time.
            if ((RLon == 1) && (elig_Hebb_on == 1) &&
                (interval <= maxeligwind_Hebb)) {
                tlast_Hebb_elig = t
            }
        }
        
        : Remember the current spike time for next NET_RECEIVE.
        tlast_post = t  
    }
    
    : Adjust the weight.
    adjust_weight(deltaw)
}

PROCEDURE reward_punish(reinf) {
    : If RL is turned on...
    if (RLon == 1) {
        : Start the weight change as being 0.
        deltaw = 0.0
        
        : If we have the Hebbian eligibility traces on, add their effect in.
        if (elig_Hebb_on == 1) {
            deltaw = deltaw + reinf * Hebb_elig_gain()
        }       
        
        : If we have the anti-Hebbian eligibility traces on, add their effect in.
        if (elig_antiHebb_on == 1) {
            deltaw = deltaw + reinf * antiHebb_elig_gain()
        }
    
        : If we have soft-thresholding on, apply it.
        if (adj_soft_thresh == 1) {
            deltaw = soft_threshold(deltaw)
        }    
        
        : Adjust the weight.
        adjust_weight(deltaw)
        
        : Show weight update information if debugging on.
        if (verbose > 0) {
            printf("RL event: t = %f ms; deltaw = %f\n",t,deltaw)
        }        
    }
}

FUNCTION Hebb_elig_gain() {
    : If RL is turned off or eligibility has not occurred yet, return 0.0.
    if ((RLon == 0) || (tlast_Hebb_elig < 0.0)) {
            Hebb_elig_gain = 0.0    
    : If we are using a binary (i.e. square-wave) eligibility traces...
    } else if (expdecay_elig == 0) {
        : If we are within the length of the eligibility trace...
        if (t - tlast_Hebb_elig <= eliglength_Hebb) {
            Hebb_elig_gain = maxeliggain_Hebb
        : Otherwise (outside the length), return 0.0.
        } else {
            Hebb_elig_gain = 0.0
        }
    : Otherwise (if we're using an exponential decay traces)...
    } else {
        : Use the Hebbian decay to calculate the gain.   
        Hebb_elig_gain = maxeliggain_Hebb * 
            exp((tlast_Hebb_elig - t) / eliglength_Hebb)          
    }
}

FUNCTION antiHebb_elig_gain() {
    : If RL is turned off or eligibility has not occurred yet, return 0.0.
    if ((RLon == 0) || (tlast_antiHebb_elig < 0.0)) {
            antiHebb_elig_gain = 0.0    
    : If we are using a binary (i.e. square-wave) eligibility traces...
    } else if (expdecay_elig == 0) {
        : If we are within the length of the eligibility trace...
        if (t - tlast_antiHebb_elig <= eliglength_antiHebb) {
            antiHebb_elig_gain = maxeliggain_antiHebb                
        : Otherwise (outside the length), return 0.0.
        } else {
            antiHebb_elig_gain = 0.0
        }
    : Otherwise (if we're using an exponential decay traces)...
    } else {
        : Use the anti-Hebbian decay to calculate the gain.  
        antiHebb_elig_gain = maxeliggain_antiHebb * 
            exp((tlast_antiHebb_elig - t) / eliglength_antiHebb)              
    }
}

FUNCTION soft_threshold(rawwc) {
    : If the weight change is non-negative, scale by 1 - weight / wmax.
    if (rawwc >= 0) {
        soft_threshold = rawwc * (1.0 - get_weight() / wmax)
     : Otherwise (the weight change is negative), scale by weight / wmax.               
    } else {
        soft_threshold = rawwc * get_weight() / wmax
    }
}

PROCEDURE adjust_weight(wc) {
    : If the unit is on, apply the synaptic modification, and then clip the
    : weight if necessary to make sure it's between 0 and wmax.
    if (on) {
        newwt = get_weight() + wc
        if (newwt > wmax) {
            newwt = wmax
        }
        if (newwt < 0) {
            newwt = 0
        }
        set_weight(newwt)
    }
}

FUNCTION get_weight() {
    : If we are using a negative weight, get the positive version.
    if (negweight > 0) {
        get_weight = -wsyn
    } else {
        get_weight = wsyn
    }
}

PROCEDURE set_weight(wtval) {
    : If we are using a negative weight, set the value to the negative of the 
    : input.
    if (negweight > 0) {
        wsyn = -wtval
    } else {
        wsyn = wtval
    }
}
