COMMENT
STDP Weight Adjuster POINT_PROCESS
Spike Timing-Dependent Weight Adjuster
Last update: 7/17/13 (georgec)

original STDP code adapted from:
http://senselab.med.yale.edu/modeldb/showmodel.asp?model=64261&file=\bfstdp\stdwa_songabbott.mod
based on (Song and Abbott, 2001)
by Andrew Davison, UNIC, CNRS, 2003-2004

adapted to implement a "nearest-neighbor spike-interaction" model (see 
Scholarpedia article on STDP) that just looks at the last-seen pre- and 
post-synaptic spikes
George Chadderdon, SUNY Downstate Medical Center
ENDCOMMENT

NEURON {
    : Definition of mechanism
    POINT_PROCESS StdpWtAdjLastSpks
    
    : LTP/LTD decay time constants (in ms) for the Hebbian (pre-before-post- 
    : synaptic spikes), and anti-Hebbian (post-before-pre-synaptic) cases. 
    : Typically, e.g. for excitatory connections, the Hebbian part of the curve 
    : triggers LTP and the anti-Hebbian part LTD.  For other cases, e.g. 
    : inhibitory connections, the anti-Hebbian part may trigger LTP and the 
    : Hebbian part LTD.  Sometimes the signs may even be the same for 
    : the two parts of the curve, depending on the location of the synapse.
    : See the Scholarpedia article on STDP (Diversity of STDP section):
    :   http://www.scholarpedia.org/article/Spike-timing_dependent_plasticity
    RANGE tau_Hebb, tau_antiHebb
    
    : Maximal adjustment (can be positive or negative) for Hebbian and 
    : anti-Hebbian cases (i.e., as inter-spike interval approaches zero).  This
    : should be set positive for LTP and negative for LTD.
    RANGE maxadj_Hebb, maxadj_antiHebb  
    
    : The maximum weight for the synapse.
    RANGE wmax
    
    : Flag turning on "soft thresholding" for the maximal adjustment parameters.
    RANGE adj_soft_thresh
    
    : Pointer to the weight (in a NetCon object) to be adjusted.
    POINTER wsyn    
        
    : Flag for turning the weight adjustment on / off.
    RANGE on
    
    : Flag for whether the weight pointed to must be negative.
    RANGE negweight
    
    : Flag for turning off prints of weight update events for debugging.
    RANGE verbose
            
    : Remembered times for last pre- and post-synaptic spikes.
    RANGE tlast_pre, tlast_post
    
    : Interval between current time t and previous spike.
    RANGE interval
    
    : The calculated weight change.
    RANGE deltaw
    
    : New calculated weight.
    RANGE newwt
}

ASSIGNED {
    wsyn        
    tlast_pre   (ms)    
    tlast_post  (ms)    
    interval    (ms)    
    deltaw
    newwt          
}

INITIAL {
    tlast_pre = -1     : no spike yet
    tlast_post = -1   : no spike yet
    interval = 0
    deltaw = 0
    newwt = 0
}

PARAMETER {
    tau_Hebb  = 10  (ms)   
    tau_antiHebb  = 10  (ms)    
    maxadj_Hebb = 1.0
    maxadj_antiHebb = -1.0
    wmax  = 15.0
    adj_soft_thresh = 0
    on = 1  
    negweight = 0
    verbose = 0
}

NET_RECEIVE (w) {
    : Default the weight change to 0.
    deltaw = 0.0
     
    : If we receive a non-negative weight value, we are receiving a pre-synaptic
    : spike (and thus need to check for an anti-Hebbian event, since the post-
    : synaptic weight must be earlier).
    if (w >= 0) {           
        : Get the interval.
        interval = tlast_post - t   : interval is negative
        
        : If we had a post-synaptic spike and a non-zero interval...
        if  ((tlast_post > -1) && (interval != 0)) {
            : Use the anti-Hebbian decay to set the anti-Hebbian weight adjustment.
            deltaw = maxadj_antiHebb * exp(interval / tau_antiHebb)
            
            : If we have soft-thresholding on...
            if (adj_soft_thresh == 1) {
                deltaw = soft_threshold(deltaw)
            }
            
            : Show weight update information if debugging on.       
            if (verbose > 0) {
                printf("anti-Hebbian event: t = %f ms; deltaw = %f\n",t,deltaw)
            }  
        }
        
        : Remember the current spike time for next NET_RECEIVE.
        tlast_pre = t       
        
    : Else, if we receive a negative weight value, we are receiving a post-
    : synaptic spike (and thus need to check for an anti-Hebbian event, since 
    : the post-synaptic weight must be earlier).        
    } else {                   
        : Get the interval.        
        interval = t - tlast_pre    : interval is positive
                     
        : If we had a pre-synaptic spike and a non-zero interval...
        if  ((tlast_pre > -1) && (interval != 0)) {      
            : Use the Hebbian decay to set the Hebbian weight adjustment.   
            deltaw = maxadj_Hebb * exp(-interval / tau_Hebb)
                      
             : If we have soft-thresholding on...
            if (adj_soft_thresh == 1) {
                deltaw = soft_threshold(deltaw)
            }    
            
            : Show weight update information if debugging on.
            if (verbose > 0) {
                printf("Hebbian event: t = %f ms; deltaw = %f\n",t,deltaw)
            }           
        }
        
        : Remember the current spike time for next NET_RECEIVE.
        tlast_post = t  
    }
    
    : Adjust the weight.
    adjust_weight(deltaw)
}

FUNCTION soft_threshold(rawwc) {
    : If the weight change is non-negative, scale by 1 - weight / wmax.
    if (rawwc >= 0) {
        soft_threshold = rawwc * (1.0 - get_weight() / wmax)
     : Otherwise (the weight change is negative), scale by weight / wmax.               
    } else {
        soft_threshold = rawwc * get_weight() / wmax
    }
}

PROCEDURE adjust_weight(wc) {
    : If the unit is on, apply the synaptic modification, and then clip the
    : weight if necessary to make sure it's between 0 and wmax.
    if (on) {
        newwt = get_weight() + wc
        if (newwt > wmax) {
            newwt = wmax
        }
        if (newwt < 0) {
            newwt = 0
        }
        set_weight(newwt)
    }
}

FUNCTION get_weight() {
    : If we are using a negative weight, get the positive version.
    if (negweight > 0) {
        get_weight = -wsyn
    } else {
        get_weight = wsyn
    }
}

PROCEDURE set_weight(wtval) {
    : If we are using a negative weight, set the value to the negative of the 
    : input.
    if (negweight > 0) {
        wsyn = -wtval
    } else {
        wsyn = wtval
    }
}
