"""
genunittest.py -- Script for testing a NeuralUnit.
Usage: ipython --pylab -i scripts/genunittest.py
Last revision: 9/26/13 (georgec)
"""

# Run the bulk of the module imports.
execfile('sim/sim.py') 

# Import neuralnetwork module.
import neuralnetwork as nnet

# Import neuralsim module.
import neuralsim as nsim

# Import the timed callbacks module.
import timedcallback as tcb

# Set up recording interval (in ms).
#recinterval = 0.025
recinterval = 0.1

#
# Class for the network
#

class NeuralUnitTestModel(nnet.NeuralNetwork):
    """
    Class for the entire NeuralUnit tester model.

    Methods:
      __init__(self): constructor
      prenrnrun(self): code needed pre-run() by INTF6 units
      postnrnrun(self): code needed post-run() by INTF6 units

    Attributes:
      classlabel (class): text label for the class
      netparams: dictionary of parameters for the model   
        'testunittype': type of unit for tst (set to classlabel)
          'IntFire1Unit' (default setting)
          'INTF6Unit'
          'SPICellUnit'
          'IzhiUnit'        
      nstexc: excitatory NetStimUnit for driving the unit to be tested
      nstinh: inhibitory NetStimUnit for driving the unit to be tested      
      tst: NeuralUnit to be tested
      ic: INTF6Unit for receiving the output from tst
      nstexcnlc: a connection from nstexc to tst
      nstinhnlc: a connection from nstinh to tst
      tstnlc: a connection from tst to ic

    Usage:
    >>>thenet = NeuralUnitTestModel(label='NeuralUnitTestModel')
    """

    classlabel = 'NeuralUnitTestModel'

    def __init__(self,  params={}, label='NeuralUnitTestModel'):
        # Set superclass parameters.
        super(NeuralUnitTestModel,self).__init__(params, label)
                                
        # Set up two NetStimUnits for excitation and inhibition.
        self.nstexc = stim.NetStimUnit(noise=0.0, label='nstexc')
        self.nstinh = stim.NetStimUnit(noise=0.0, NTs=['GABA'], label='nstinh')        
            
        # Set up global parameters for INTF6 units (including ic).
        intf6.INTF6Unit.set_EAM(65.0)
        intf6.INTF6Unit.set_ENM(90.0)
        intf6.INTF6Unit.set_EGA(-15.0)
        intf6.INTF6Unit.set_mg(1.6)
            
        # Set up the test unit.                    
        if self.netparams['testunittype'] == 'INTF6Unit':                    
            self.tst = intf6.INTF6Unit(celltype=h.E5B,VRMP=-65.0,VTH=-37.5,\
                Vblock=-25.0,ahpwt=1.0,tauahp=10.0,refrac=5.0,RRWght=0.25,\
                tauRR=8.0,tauAM=10,tauAM2=20.0,tauNM=300.0,tauNM2=300.0,\
                tauGA=10.0,tauGA2=20.0,label='tst')   
        elif self.netparams['testunittype'] == 'SPICellUnit':        
            self.tst = spic.SPICellUnit(label='tst')  
        elif self.netparams['testunittype'] == 'IzhiUnit':        
            self.tst = izhi.IzhiUnit(C=100, k=3, vr=-60, vt=-50, vpeak=30, \
                a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
                tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
                verbose=0, label='tst')                                  
        else:   # IntFire1Unit
            self.tst = if1.IntFire1Unit(tau=10.0, refrac=5.0, label='tst')
                     
        # Set up a single INTF6Unit to receive the output of the test-unit.
        self.ic = intf6.INTF6Unit(celltype=h.E5B,VRMP=-65.0,VTH=-37.5,\
            Vblock=-25.0,ahpwt=1.0,tauahp=10.0,refrac=5.0,RRWght=0.25,\
            tauRR=8.0,tauAM=10,tauAM2=20.0,tauNM=300.0,tauNM2=300.0,\
            tauGA=10.0,tauGA2=20.0,label='ic')   
                
        # Attach a NeuralConn from the excitatory NetStimUnit to the 
        # test-unit.
        if self.netparams['testunittype'] == 'INTF6Unit':
            self.nstexcnlc = intf6.INTF6Conn(self.nstexc,self.tst,weight=30.0, \
                delay=0.0,label='nstexcnlc')
        elif self.netparams['testunittype'] == 'SPICellUnit':
            self.nstexcnlc = nconn.NeuralConn(self.nstexc,self.tst,weight=30.0, \
                delay=0.0,label='nstexcnlc')
        elif self.netparams['testunittype'] == 'IzhiUnit':
            self.nstexcnlc = izhi.IzhiConn(self.nstexc,self.tst,weight=30.0, \
                delay=0.0,label='nstexcnlc')                               
        else:   # IntFire1Unit
            self.nstexcnlc = if1.IntFire1Conn(self.nstexc,self.tst,weight=1.1, \
                delay=0.0,label='nstexcnlc')
                
        # Attach a NeuralConn from the inhibitory NetStimUnit to the 
        # test-unit.                
        if self.netparams['testunittype'] == 'INTF6Unit':
            self.nstinhnlc = intf6.INTF6Conn(self.nstinh,self.tst,weight=30.0, \
                delay=0.0,label='nstinhnlc')
        elif self.netparams['testunittype'] == 'SPICellUnit':
            self.nstinhnlc = nconn.NeuralConn(self.nstinh,self.tst,weight=30.0, \
                delay=0.0,label='nstinhnlc') 
        elif self.netparams['testunittype'] == 'IzhiUnit':
            self.nstinhnlc = izhi.IzhiConn(self.nstinh,self.tst,weight=30.0, \
                delay=0.0,label='nstinhnlc')                            
        else:   # IntFire1Unit
            self.nstinhnlc = if1.IntFire1Conn(self.nstinh,self.tst,weight=1.1, \
                delay=0.0,label='nstinhnlc')
                        
        # Attach a NeuralConn from the test-unit to the INTF6Unit.
        self.tstnlc = nconn.NeuralConn(self.tst,self.ic,weight=30.0, \
            delay=0.0,label='tstnlc')    
                                                                           
    def setdefaultparams(self):    
        # Set up what type of unit is used for prx.
        self.netparams['testunittype'] = 'IntFire1Unit'
        
    def prenrnrun(self):
        self.ic.hobj.global_init()   # just need to call once
        
    def postnrnrun(self):
        self.ic.hobj.global_fini()   # just need to call once
        self.ic.hobj.spkstats2(1)   # will need for each cell /  column? not sure
        print 'TMAX: %f' % h.tmax_INTF6                
                                      
class NeuralUnitTestSim(nsim.NeuralSimulation):
    """
    Class for the a sim for the NeuralUnit tester model.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      simparams: dictionary of simulation (as opposed to network) parameters
        'simduration': simulation duration (in ms)
        'nstexcon': is the excitatory stimulus on?
        'nstinhon': is the inhibitory stimulus on?
        'stimstarttime': time of first stimulus spike of the sim
        'stimspkinterval': interval of repeated spiking of NetStimUnits 
                        
    Usage:
    >>>thesim = NeuralUnitTestSim(label='NeuralUnitTestSim')
    """

    classlabel = 'NeuralUnitTestSim'

    def __init__(self, thenet=None, params={}, filename='', \
        label='NeuralUnitTestSim'):
        # Set superclass parameters.
        super(NeuralUnitTestSim,self).__init__(thenet, params, filename, label)
        
        # If we need to initialize a new network with (read in) params, do so.
        if self.initnewnetwork:
            self.thenetwork = NeuralUnitTestModel(self.networkparams)
            
        # Set up the stimuli.
        self.setupstims()
        
        # Set up the recordings.
        self.setuprecordings()           

    def setupstims(self):
        # Set the time that the first spike happens.
        firstspktime = self.simparams['stimstarttime'] 
        
        # Set the repeated spiking interval.
        spkInterv = self.simparams['stimspkinterval'] 
            
        # Set up NetStimUnit firing parameters.
        self.thenetwork.nstexc.set_interval(spkInterv)
        self.thenetwork.nstexc.set_number(1e9)
        self.thenetwork.nstexc.set_start(firstspktime)
        self.thenetwork.nstinh.set_interval(spkInterv)  
        self.thenetwork.nstinh.set_number(1e9)
        self.thenetwork.nstinh.set_start(firstspktime)
        
        # Disable the excitatory NetStimUnit if we want it off.   
        if not self.simparams['nstexcon']:
            self.thenetwork.nstexc.disable()
            
        # Disable the inhibitory NetStimUnit if we want it off.   
        if not self.simparams['nstinhon']:            
            self.thenetwork.nstinh.disable()
                                              
    def setuprecordings(self):
        # Turn on spike recording for all of the NeuralUnits.
        for unit in nun.NeuralUnit.allUnits:
            unit.turnonrecspks()
            
        # Turn on spike recording for all of the NeuralConns.
        for conn in nconn.NeuralConn.allConns:
            conn.turnonrecspks()
           
        # Turn on membrane potential recording for tst.
        if self.thenetwork.tst.classlabel == 'INTF6Unit':
            self.thenetwork.tst.turnonrecVms(self.simparams['simduration'])
        elif self.thenetwork.tst.classlabel == 'SPICellUnit':
            self.thenetwork.tst.turnonrecvs(recinterval)
        elif self.thenetwork.tst.classlabel == 'IzhiUnit':
            self.thenetwork.tst.turnonrecVs(self, recinterval)                           
        else:  # IntFire1Unit
            self.thenetwork.tst.turnonrecms(self, recinterval)
                         
        # Set up the recording calls to be called post-run().
        self.reccalls['nstexcspks'] = self.thenetwork.nstexc.getrecspks
        self.reccalls['nstinhspks'] = self.thenetwork.nstinh.getrecspks         
        self.reccalls['tstspks'] = self.thenetwork.tst.getrecspks
        self.reccalls['icspks'] = self.thenetwork.ic.getrecspks
        self.reccalls['nstexcnlcspks'] = self.thenetwork.nstexcnlc.getrecspks
        self.reccalls['nstinhnlcspks'] = self.thenetwork.nstinhnlc.getrecspks
        self.reccalls['tstnlcspks'] = self.thenetwork.tstnlc.getrecspks
        if self.thenetwork.tst.classlabel == 'INTF6Unit':         
            self.reccalls['tstVms'] = self.thenetwork.tst.getrecVms
        elif self.thenetwork.tst.classlabel == 'SPICellUnit':         
            self.reccalls['tstvs'] = self.thenetwork.tst.getrecvs 
        elif self.thenetwork.tst.classlabel == 'IzhiUnit':         
            self.reccalls['tstVs'] = self.thenetwork.tst.getrecVs                         
        else:  # IntFire1Unit
            self.reccalls['tstms'] = self.thenetwork.tst.getrecms
            
    def setdefaultparams(self):   
        # Set the simulation duration (in ms).
        self.simparams['simduration'] = 1000.0 
         
        # Set whether the excitatory NetStimUnit is on.
        self.simparams['nstexcon'] = True

        # Set whether the inhibitory NetStimUnit is on.
        self.simparams['nstinhon'] = False
        
        # Set the stimulus start time (in ms).
        self.simparams['stimstarttime'] = 400.0
        
        # Set the spiking interval for the NetStimUnits (in ms).
        self.simparams['stimspkinterval'] = 50.0        
                                
    def globalfinits(self):
        # Run superclass method.
        super(NeuralUnitTestSim,self).globalfinits()
                                                                                                                                                                               
    def showresults(self):
        #
        # Show the spike and membrane value results
        #
        print 'Showing the results...\n'
        
        print 'Model spikes:'
        print '      nstexc ->' , self.recdata['nstexcspks']
        print '      nstinh ->' , self.recdata['nstinhspks']        
        print '         tst ->' , self.recdata['tstspks']
        print '          ic ->' , self.recdata['icspks']
        print '   nstexcnlc ->' , self.recdata['nstexcnlcspks']
        print '   nstinhnlc ->' , self.recdata['nstinhnlcspks']        
        print '      tstnlc ->' , self.recdata['tstnlcspks']
        
        # Plot unit spike times.
        #figure()
        #self.plotunitspikes('nstexc')
        #figure()
        #self.plotunitspikes('nstinh')        
        #figure()
        #self.plotunitspikes('tst')
        #figure()
        #self.plotunitspikes('ic')
                
        # Show the SpikeCellUnit voltages.
        figure()
        #tvec, vec = self.recdata['tstvs']
        #plot(tvec, vec)
        self.plotunitvoltages('tst', stimname='nstexc')
                
    def plotunitspikes(self, unitname):
        print type(unitname)        
        xmarg = 10   # x display margin
        spktimes = self.recdata['%sspks' % unitname]
        if len(spktimes) > 0:
            plot_spike_times(spktimes)
        title('%s Spike Times' % unitname)
        xlabel('Time (ms)')
        xlim((-xmarg, self.simparams['simduration'] + xmarg))
 
    def plotunitvoltages(self, unitname, stimname=None):
        xmarg = 10    # x display margin
        ymarg = 0.1   # y display margin
        if unitname == 'tst':
            unittype = self.networkparams['testunittype']      
        else:   
            unittype = eval('self.thenetwork.%s.classlabel' % unitname)
        #unitobj = eval('self.thenetwork.%s' % unitname)
        if unittype == 'INTF6Unit':
            tvec, vec = self.recdata['%sVms' % unitname]
        elif unittype == 'SPICellUnit':
            tvec, vec = self.recdata['%svs' % unitname]  
        elif unittype == 'IzhiUnit':
            tvec, vec = self.recdata['%sVs' % unitname]                       
        else:  # IntFire1Unit
            tvec, vec = self.recdata['%sms' % unitname]         
        plot(tvec, vec)
        title('%s Membrane Potentials' % unitname)
        xlabel('Time (ms)')
        if unittype == 'INTF6Unit':
            ylabel('Vm (mV)')
        elif unittype == 'SPICellUnit':
            ylabel('v (mV)')  
        elif unittype == 'IzhiUnit':
            ylabel('V (mV)')                  
        else:  # IntFire1Unit
            ylabel('m value')
            ylim(-ymarg - 1.0, 2.0 + ymarg)
        xlim((-xmarg, self.simparams['simduration'] + xmarg))
        spktimes = self.recdata['%sspks' % unitname]
        #for ind in range(len(spktimes)):
            #axvline(spktimes[ind], color='r')   
        if stimname != None:
            for stimtime in self.recdata['%sspks' % stimname]:
                if unittype == 'INTF6Unit':
                    plot(stimtime, -70.0, 'go')  
                elif unittype == 'SPICellUnit':
                    plot(stimtime, -110.0, 'go')    
                elif unittype == 'IzhiUnit':
                    plot(stimtime, -65.0, 'go')                                     
                else:  # IntFire1Unit
                    plot(stimtime, -0.02, 'go') 
        # This is commented out because it doesn't work properly with saved 
        # NeuralSimulations.  To make it work correctly, I may need to save the
        # values that are to be plotted.
        #if unittype == 'INTF6Unit':      
            #axhline(unitobj.get_VRMP(), color='k',linestyle=':')
            #axhline(unitobj.get_VTH(), color='k',linestyle='--')
            #axhline(unitobj.get_Vblock(), color='r',linestyle='--')
                                                  
#
# Globals
#

# Set up directory and file names.
datadir = 'data'
simdir =  '%s/testsim' % datadir
simfstem = '%s/testsim-' % simdir
simfilename = '%ssim' % simfstem

#
# Script code to run.
#

#
# Set up network and simulation parameters.
#

# Create an empty dictionary for all of the parameters.
theparams = {}

# Set whether we turn off even initializing the network on initial run.
theparams['noinit'] = False

# Set whether we turn off running the sim.
theparams['norun'] = False

# Set whether we turn off showing the results of the sim.
theparams['noshow'] = False

# Set whether we turn off showing the results of the sim.
theparams['dosave'] = False

# Set file to load a sim from.
theparams['loadsim'] = None

# Read in what parameters exist on the command line.
cmdlineparams(theparams)

# Set up what type of NeuralUnit is being tested.
#theparams['testunittype'] = 'SPICellUnit' # 'INTF6Unit'

# Set the simulation duration (in ms).
#theparams['simduration'] = 1500.0

# Set the stimulus start time (in ms).
#theparams['stimstarttime'] = 400.0

# Set the spiking interval for the NetStimUnit (in ms).
#theparams['stimspkinterval'] = 50.0

# Set whether the excitatory NetStimUnit is on.
#theparams['nstexcon'] = False # True

# Set whether the inhibitory NetStimUnit is on.
#theparams['nstinhon'] = True  # False

# If we are loading from an existing sim...
if theparams['loadsim'] != None:
    pass
    ## Load the simulation from the specified file.
    ##thesimfile = theparams['loadsim']
    #thesimfile = '%s/%s' % (simdir, theparams['loadsim'])
    #thesim = STDPRLTestSim(params=theparams, filename=thesimfile)
    
    ## Show the results, if we haven't disabled.
    #if not theparams['noshow']: 
        #thesim.showresults()
    
# If we're not disabling initialization...
elif not theparams['noinit']:
    # Initialize the network.
    print 'Initializing network...'
    #thenet = NeuralUnitTestModel()  # just using default parameters
    thenet = NeuralUnitTestModel(theparams)  # passing in parameters
    print
    
    # Initialize a new simulation.
    #thesim = NeuralUnitTestSim(thenet)  # just using default parameters
    thesim = NeuralUnitTestSim(thenet, theparams)
    
    # If we're not disabling running...
    if not theparams['norun']:    
        # Run the simulation
        print 'Running the simulation...\n'
        thesim.run()
        print
        
        # Show the results, if we haven't disabled.
        if not theparams['noshow']: 
            thesim.showresults()
            
        # Do a save, if we want to.
        if theparams['dosave']:
            thesim.save()
