"""
simpleizhinet.py -- Script implementing Cliff's simple.py under the 
framework.

Usage: ipython --pylab -i scripts/simpleizhinet.py
Last revision: 9/30/13 (georgec)
"""

# Run the bulk of the module imports.
execfile('sim/sim.py') 

# Import neuralnetwork module.
import neuralnetwork as nnet

# Import neuralsim module.
import neuralsim as nsim

# Import the timed callbacks module.
import timedcallback as tcb

# Set up recording interval (in ms).
#recinterval = 0.025
recinterval = 0.1

#
# Class for the network
#

class SimpleIzhiNetModel(nnet.NeuralNetwork):
    """
    Class for the entire Izhikevich network model.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      netparams: dictionary of parameters for the model   
        'ncells': number of IzhiUnits in the model
        'efraction': fraction of cells that are excitatory
      allUnits: UnitGroup containing all of the units
      allConns: ConnGroup containing all of the connections

    Usage:
    >>>thenet = SimpleIzhiNetModel(label='SimpleIzhiNetModel')
    """

    classlabel = 'SimpleIzhiNetModel'

    def __init__(self,  params={}, label='SimpleIzhiNetModel'):
        # Set superclass parameters.
        super(SimpleIzhiNetModel,self).__init__(params, label)
        
        #
        # Create the unit and connection groups.
        #
        
        # Set up empty groups for all of the units and connections.
        self.allUnits = ugs.UnitGroup(label='AllUnits')
        self.allConns = nconn.ConnGroup(label='AllConns')
        
        # Set up empty UnitGroups for the unit populations.
        self.Estimpop = ugs.UnitGroup(label='EStimPop')
        self.Istimpop = ugs.UnitGroup(label='IStimPop')                         
        self.Epop = ugs.UnitGroup(label='EPop')
        self.Ipop = ugs.UnitGroup(label='IPop')
        
        # Set up empty ConnGroups for the inter-population connections.
        self.stimtoEnlcs = nconn.ConnGroup(label='StimtoEnlcs')
        self.stimtoInlcs = nconn.ConnGroup(label='StimtoInlcs')
        self.EtoEnlcs = nconn.ConnGroup(label='EtoEnlcs')
        self.EtoInlcs = nconn.ConnGroup(label='EtoInlcs')        
        self.ItoEnlcs = nconn.ConnGroup(label='ItoEnlcs')
        self.ItoInlcs = nconn.ConnGroup(label='ItoInlcs')
        
        #
        # Create all of the NeuralUnits and the NetStimUnit->IzhiUnit 
        # connections.
        #
        
        # Set the number of excitatory and inhibitory IzhiUnits.
        self.numexc = int(self.netparams['efraction'] * \
            self.netparams['ncells'])
        self.numinh = self.netparams['ncells'] - self.numexc

        # Set the pylab rand() seed using the wiring random seed.
        seed(self.netparams['wirerandseed'])
        
        # Create the excitatory cells and their driving stimuli.
        for ii in range(self.numexc):
            # Create a NetStimUnit and add it to the appropriate populations.
            newlabel = 'Estim_%d' % ii
            newstim = stim.NetStimUnit(noise=0.0, label=newlabel, \
                randseed=ii+1)
            self.allUnits.append(newstim)
            self.Estimpop.append(newstim)
            
            # Create an IzhiUnit and add it to the appropriate populations.
            newlabel = 'E_%d' % ii
            newunit = izhi.IzhiUnit(C=100, k=3, vr=-60, vt=-50, vpeak=30, \
                    a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
                    tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
                    verbose=0, label=newlabel)
            newunit.loc = nun.UnitLocation(self.netparams['networksize'] * rand(), \
                self.netparams['networksize'] * rand(), 0.0)   
            self.allUnits.append(newunit)
            self.Epop.append(newunit)  
            
            # Create a (non-plastic excitatory) connection from newstim to 
            # newunit and add to the appropriate ConnGroups.
            newlabel = 'EstimtoE_%d' % ii
            newconn = izhi.IzhiConn(newstim, newunit, weight=0.0, \
                delay=2.0, label=newlabel)                       
            self.allConns.append(newconn)
            self.stimtoEnlcs.append(newconn)
            
        # Create the inhibitory cells and their driving stimuli.
        for ii in range(self.numinh):
            # Create a NetStimUnit and add it to the appropriate populations.
            newlabel = 'Istim_%d' % ii
            newstim = stim.NetStimUnit(noise=0.0, label=newlabel, \
                randseed=ii+self.numexc+1)
            self.allUnits.append(newstim)
            self.Istimpop.append(newstim)
            
            # Create an IzhiUnit and add it to the appropriate populations.
            newlabel = 'I_%d' % ii
            newunit = izhi.IzhiUnit(C=100, k=3, vr=-60, vt=-50, vpeak=30, \
                    a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
                    tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
                    verbose=0, NTs=['GABA'], label=newlabel)
            newunit.loc = nun.UnitLocation(self.netparams['networksize'] * rand(), \
                self.netparams['networksize'] * rand(), 0.0)   
            self.allUnits.append(newunit)
            self.Ipop.append(newunit)  
            
            # Create a (non-plastic excitatory) connection from newstim to 
            # newunit and add to the appropriate ConnGroups.
            newlabel = 'IstimtoI_%d' % ii
            newconn = izhi.IzhiConn(newstim, newunit, weight=0.0, \
                delay=2.0, label=newlabel)                       
            self.allConns.append(newconn)
            self.stimtoInlcs.append(newconn)
        
        #    
        # Create the connections between the IzhiUnits.
        #
        
                                                                                                           
    def setdefaultparams(self):    
        # Set up how many IzhiUnits are in the model.
        self.netparams['ncells'] = 1000
        
        # Set up the fraction of cells that are excitatory.
        self.netparams['efraction'] = 0.8
        
        # Set up the random seed for wiring.
        self.netparams['wirerandseed'] = 0
        
        # Set up the network size (width in microns).
        self.netparams['networksize'] = 2000.0
                                              
class SimpleIzhiNetSim(nsim.NeuralSimulation):
    """
    Class for the a sim for the Izhikevich model.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      simparams: dictionary of simulation (as opposed to network) parameters
        'simduration': simulation duration (in ms)
        ...
                        
    Usage:
    >>>thesim = SimpleIzhiNetSim(label='SimpleIzhiNetSim')
    """

    classlabel = 'SimpleIzhiNetSim'

    def __init__(self, thenet=None, params={}, filename='', \
        label='SimpleIzhiNetSim'):
        # Set superclass parameters.
        super(SimpleIzhiNetSim,self).__init__(thenet, params, filename, label)
        
        # If we need to initialize a new network with (read in) params, do so.
        if self.initnewnetwork:
            self.thenetwork = SimpleIzhiNetModel(self.networkparams)
            
        # Set up the stimuli.
        self.setupstims()
        
        # Set up the recordings.
        self.setuprecordings()           

    def setupstims(self):
        # Set the appropriate parameters for all of the Estims.
        for unit in self.thenetwork.Estimpop:
            unit.set_interval(self.simparams['stimfreq'] ** -1 * 1e3)
            unit.set_number(self.simparams['stimnumber'])
            unit.set_noise(self.simparams['stimnoise'])
            
        # Set the appropriate parameters for all of the Istims.
        for unit in self.thenetwork.Istimpop:
            unit.set_interval(self.simparams['stimfreq'] ** -1 * 1e3)
            unit.set_number(self.simparams['stimnumber'])
            unit.set_noise(self.simparams['stimnoise'])           
           
        # Set all of the stimulus weights (the NMDA weights) to the desired 
        # value.
        for conn in self.thenetwork.stimtoEnlcs:
            conn.setwt2(self.simparams['stimweight'])
        for conn in self.thenetwork.stimtoInlcs:
            conn.setwt2(self.simparams['stimweight'])
                                               
    def setuprecordings(self):
        # Turn on spike recording for all of the NeuralUnits.
        self.thenetwork.allUnits.turnonrecspks() 
         
        # Set up the recording calls to be called post-run().
        self.reccalls['Estimspks'] = self.thenetwork.Estimpop.getrecspks
        self.reccalls['Istimspks'] = self.thenetwork.Istimpop.getrecspks
        self.reccalls['Espks'] = self.thenetwork.Epop.getrecspks
        self.reccalls['Ispks'] = self.thenetwork.Ipop.getrecspks          
                    
    def setdefaultparams(self):   
        # Set the simulation duration (in ms).
        self.simparams['simduration'] = 1000.0 
        
        # Set the stimulus frequency (in Hz).
        self.simparams['stimfreq'] = 2.0
        
        # Set the number of spikes to be emitted by the stims.
        self.simparams['stimnumber'] = 1e9
        
        # Set the noise value for the NetStimUnits.
        self.simparams['stimnoise'] = 1.0
        
        # Set the weight value between each NetStimUnit and its corresponding 
        # IzhiUnit.
        self.simparams['stimweight'] = 20.0
                                
    def showresults(self):
        # For now, just show separate rasters for each population.
        figure()
        plot_spike_raster(self.recdata['Estimspks'])
        xlim(0, self.simparams['simduration'])   
        title('Estimspks')    
        figure()
        plot_spike_raster(self.recdata['Istimspks'])  
        xlim(0, self.simparams['simduration'])   
        title('Istimspks')     
        figure()
        plot_spike_raster(self.recdata['Espks'])   
        xlim(0, self.simparams['simduration'])   
        title('Espks')    
        figure()
        plot_spike_raster(self.recdata['Ispks'])  
        xlim(0, self.simparams['simduration'])   
        title('Ispks')     
                
        #self.plot_raster()
                
    # This is going to need some work.  Generally, we need to work on some
    # better functions for raster plotting, probably methods in UnitGroup.
    def plot_raster(self, startt=0, endt=1000, xmarg=10, ymarg=10, \
        drawpopbounds=True, bwmode=False):
    
        def plotcells(ctyp, pcol, drawpopbounds=True, drawpoplabels=True):
            pass
            ## Get the raster.
            #raster = self.recdata['allspks']
            
            ## Init an empty cell index list.
            #cell_inds = []
            
            ## Loop over all of the cells in the raster...
            #for unitind, unitspks in enumerate(raster):
                ## If the type of the cell matches the desired type...
                #if self.thenetwork.allUnits[unitind].hobj.type() == get_ctyp_num(ctyp):
                    ## Add the cell index to the list.
                    #cell_inds.append(unitind)
                    
                    ## Get the spikes within the desired time window.
                    #tvec = unitspks[find((unitspks >= startt) & (unitspks <= endt))]
                    
                    ## Convert units to seconds.
                    #tvec = unitspks / 1000.0        
                    
                    ## Fill out the cell index vector.           
                    #vec = ones(len(tvec)) * unitind

                    ## Create the ticks for the cell spikes.
                    #marks = scatter(tvec, vec, marker='+')
                    #setp(marks, color=pcol)         

            ## If we want to draw population boundaries...
            #if (drawpopbounds):
                #axhline(cell_inds[0] - 0.5,color='k')
                #axhline(cell_inds[-1] + 0.5,color='k') 
    
            ## If we want to draw population labels...
            #if (drawpoplabels):
                #text((xmarg + endt) / 1000.0 + ((endt - startt) / 1000.0 / 80.0), \
                    #(cell_inds[0] + cell_inds[-1]) / 2.0 - 12, ctyp)
    
        # 
        # Main function begins here
        #
        
        # Get the number of cells.
        num_cells = len(thenet.allUnits)
        
        # If not black-and-white mode...
        if not bwmode:
            # Plot the cells with color-coding.
            plotcells('E','b',drawpopbounds)
            plotcells('I','r',drawpopbounds)
            plotcells('Es','g',drawpopbounds)
            plotcells('Is','m',drawpopbounds)
        else:
            # Plot the cells with black.
            plotcells('E','k',drawpopbounds)
            plotcells('I','k',drawpopbounds)
            plotcells('Es','k',drawpopbounds)
            plotcells('Is','k',drawpopbounds)
    
        # Set the x limits according to the min and max time.
        xlim((startt - xmarg) / 1000.0, (endt + xmarg) / 1000.0)
    
        # Set the y limits according to which cells exist.
        ylim(-ymarg, num_cells + ymarg - 1)
    
        # Label the axes.
        xlabel('Time (s)')
        ylabel('Cell #')
                                                  
#
# Globals
#

# Set up directory and file names.
datadir = 'data'
simdir =  '%s/testsim' % datadir
simfstem = '%s/testsim-' % simdir
simfilename = '%ssim' % simfstem

#
# Script code to run.
#

#
# Set up network and simulation parameters.
#

# Create an empty dictionary for all of the parameters.
theparams = {}

# Set whether we turn off even initializing the network on initial run.
theparams['noinit'] = False

# Set whether we turn off running the sim.
theparams['norun'] = False

# Set whether we turn off showing the results of the sim.
theparams['noshow'] = False

# Set whether we turn off showing the results of the sim.
theparams['dosave'] = False

# Set file to load a sim from.
theparams['loadsim'] = None

# Read in what parameters exist on the command line.
cmdlineparams(theparams)

# Set the simulation duration (in ms).
#theparams['simduration'] = 1500.0

# If we are loading from an existing sim...
if theparams['loadsim'] != None:
    pass
    ## Load the simulation from the specified file.
    ##thesimfile = theparams['loadsim']
    #thesimfile = '%s/%s' % (simdir, theparams['loadsim'])
    #thesim = SimpleIzhiNetSim(params=theparams, filename=thesimfile)
    
    ## Show the results, if we haven't disabled.
    #if not theparams['noshow']: 
        #thesim.showresults()
    
# If we're not disabling initialization...
elif not theparams['noinit']:
    # Initialize the network.
    print 'Initializing network...'
    #thenet = SimpleIzhiNetModel()  # just using default parameters
    thenet = SimpleIzhiNetModel(theparams)  # passing in parameters
    print
    
    # Initialize a new simulation.
    #thesim = SimpleIzhiNetSim(thenet)  # just using default parameters
    thesim = SimpleIzhiNetSim(thenet, theparams)
    
    # If we're not disabling running...
    if not theparams['norun']:    
        # Run the simulation
        print 'Running the simulation...\n'
        thesim.run()
        print
        
        # Show the results, if we haven't disabled.
        if not theparams['noshow']: 
            thesim.showresults()
            
        # Do a save, if we want to.
        if theparams['dosave']:
            thesim.save()
