"""
stdptestnet.py -- Script for testing STDP.
Usage: ipython --pylab -i scripts/stdptestnet.py
Last revision: 9/27/13 (georgec)
"""

# Run the bulk of the module imports.
execfile('sim/sim.py')

# Import neuralnetwork module.
import neuralnetwork as nnet

# Import neuralsim module.
import neuralsim as nsim

# Import the timed callbacks module.
import timedcallback as tcb

# Set up recording interval (in ms).
#recinterval = 0.025
recinterval = 0.1

#
# Class for the network
#

class STDPTestModel(nnet.NeuralNetwork):
    """
    Class for the entire STDP tester model.

    Methods:
      __init__(self): constructor
      prenrnrun(self): code needed pre-run() by INTF6 units
      postnrnrun(self): code needed post-run() by INTF6 units

    Attributes:
      classlabel (class): text label for the class
      netparams: dictionary of parameters for the model
        'prxunittype': type of unit for prx (set to classlabel)
          'IntFire1Unit' (default setting)
          'INTF6Unit'
          'SPICellUnit'
          'IzhiUnit'
        'poxunittype': same types as of above, but for pox
        'prxexcitatory': is the prx unit excitatory?
        'wtadjsoftthresh': do we use weight adjustment soft-thresholding?
      nst1, nst2: two NetStimUnits that fire at different phases
      prx: a pre-synaptic cell for the STDP test
      pox: a post-synaptic cell for the STDP test
      nst1nlc: a connection from nst1 to prx
      nst2nlc: a connection from nst2 to pox
      learnnlc: a connection from prx to pox

    Usage:
    >>>thenet = STDPTestModel(label='STDPTestModel')
    """

    classlabel = 'STDPTestModel'

    def __init__(self,  params={}, label='STDPTestModel'):
        # Set superclass parameters.
        super(STDPTestModel,self).__init__(params, label)
                       
        # Set up two NetStimUnits.
        self.nst1 = stim.NetStimUnit(noise=0.0, label='nst1')
        #self.nst1 = stim.SpikeCallbackStimUnit(label='nst1')            
        self.nst2 = stim.NetStimUnit(noise=0.0, label='nst2')
                
        # If we use INTF6 units, set up global parameters for them.
        if ((self.netparams['prxunittype'] == 'INTF6Unit') or \
            (self.netparams['poxunittype'] == 'INTF6Unit')):
            intf6.INTF6Unit.set_EAM(65.0)
            intf6.INTF6Unit.set_ENM(90.0)
            intf6.INTF6Unit.set_EGA(-15.0)
            intf6.INTF6Unit.set_mg(1.6)
                        
        # Create the pre-synaptic unit.
        if self.netparams['prxunittype'] == 'INTF6Unit':
            if self.netparams['prxexcitatory']:
                self.prx = intf6.INTF6Unit(celltype=h.E5B,VRMP=-65.0,VTH=-37.5,\
                    Vblock=-25.0,ahpwt=1.0,tauahp=10.0,refrac=5.0,RRWght=0.25,\
                    tauRR=8.0,tauAM=10,tauAM2=20.0,tauNM=300.0,tauNM2=300.0,\
                    tauGA=10.0,tauGA2=20.0,label='prx')   
            else:
                self.prx = intf6.INTF6Unit(celltype=h.E5B,VRMP=-65.0,VTH=-37.5,\
                    Vblock=-25.0,ahpwt=1.0,tauahp=10.0,refrac=5.0,RRWght=0.25,\
                    tauRR=8.0,tauAM=10,tauAM2=20.0,tauNM=300.0,tauNM2=300.0,\
                    tauGA=10.0,tauGA2=20.0,NTs=['GABA'],label='prx')  
        elif self.netparams['prxunittype'] == 'SPICellUnit': 
            if self.netparams['prxexcitatory']:
                self.prx = spic.SPICellUnit(label='prx')  
            else:
                self.prx = spic.SPICellUnit(NTs=['GABA'], label='prx')                            
        elif self.netparams['prxunittype'] == 'IzhiUnit': 
            if self.netparams['prxexcitatory']:
                self.prx = izhi.IzhiUnit(C=100, k=3, vr=-60, vt=-50, vpeak=30, \
                    a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
                    tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
                    verbose=0, label='prx')                 
            else:
                self.prx = izhi.IzhiUnit(C=100, k=3, vr=-60, vt=-50, vpeak=30, \
                    a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
                    tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
                    verbose=0, NTs=['GABA'], label='prx')                                                           
        else:   # IntFire1Unit
            if self.netparams['prxexcitatory']:
                self.prx = if1.IntFire1Unit(tau=10.0, refrac=5.0, label='prx')  
            else:
                self.prx = if1.IntFire1Unit(tau=10.0, refrac=5.0, \
                    NTs=['GABA'], label='prx')
                       
        # Create the post-synaptic unit.
        if self.netparams['poxunittype'] == 'INTF6Unit':                    
            # This cell is same as SPI cell, except:
            # - Vblock = -10.0 rather than -25.0  
            self.pox = intf6.INTF6Unit(celltype=h.E5B,VRMP=-65.0,VTH=-37.5,\
                Vblock=-10.0,ahpwt=1.0,tauahp=10.0,refrac=5.0,RRWght=0.25,\
                tauRR=8.0,tauAM=10,tauAM2=20.0,tauNM=300.0,tauNM2=300.0,\
                tauGA=10.0,tauGA2=20.0,label='pox')      
        elif self.netparams['poxunittype'] == 'SPICellUnit': 
            self.pox = spic.SPICellUnit(label='pox')                               
        elif self.netparams['poxunittype'] == 'IzhiUnit': 
            self.pox = izhi.IzhiUnit(C=100, k=3, vr=-60, vt=-50, vpeak=30, \
                a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
                tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
                verbose=0, label='pox')                 
        else:   # IntFire1Unit
            self.pox = if1.IntFire1Unit(tau=10.0, refrac=5.0, label='pox') 
                   
        # Attach a connection from nst1 to prx.     
        if self.netparams['prxunittype'] == 'INTF6Unit':
            self.nst1nlc = intf6.INTF6Conn(self.nst1,self.prx,weight=0.0, \
                delay=0.0,label='nst1nlc')
        elif self.netparams['prxunittype'] == 'SPICellUnit': 
            self.nst1nlc = nconn.NeuralConn(self.nst1,self.prx,weight=0.0, \
                delay=0.0,label='nst1nlc')           
        elif self.netparams['prxunittype'] == 'IzhiUnit': 
            self.nst1nlc = izhi.IzhiConn(self.nst1,self.prx,weight=0.0, \
                delay=0.0,label='nst1nlc')                                                        
        else:   # IntFire1Unit
            self.nst1nlc = if1.IntFire1Conn(self.nst1,self.prx,weight=0.0, \
                delay=0.0,label='nst1nlc')
                                            
        # Attach a connection from nst2 to pox.   
        if self.netparams['poxunittype'] == 'INTF6Unit':    
            self.nst2nlc = intf6.INTF6Conn(self.nst2,self.pox,weight=0.0, \
                delay=0.0,label='nst2nlc')
        elif self.netparams['poxunittype'] == 'SPICellUnit': 
            self.nst2nlc = nconn.NeuralConn(self.nst2,self.pox,weight=0.0, \
                delay=0.0,label='nst2nlc')                
        elif self.netparams['poxunittype'] == 'IzhiUnit': 
            self.nst2nlc = izhi.IzhiConn(self.nst2,self.pox,weight=0.0, \
                delay=0.0,label='nst2nlc')                                                        
        else:   # IntFire1Unit
            self.nst2nlc = if1.IntFire1Conn(self.nst2,self.pox,weight=0.0, \
                delay=0.0,label='nst2nlc')
                
        # Create the learning connection (learnnlc) between prx and pox.
        if self.netparams['poxunittype'] == 'INTF6Unit': 
            if self.netparams['prxexcitatory']:            
                self.learnnlc = intf6.INTF6STDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0,  tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=10.0, \
                    maxadj_antiHebb=-10.0, maxweight=30.0, \
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')
            else:
                self.learnnlc = intf6.INTF6STDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0,  tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=-10.0, \
                    maxadj_antiHebb=10.0, maxweight=30.0, \
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')        
        elif self.netparams['poxunittype'] == 'SPICellUnit': 
            if self.netparams['prxexcitatory']:
                self.learnnlc = stdp.STDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
                    maxadj_antiHebb=-1.0, maxweight=1.0, \
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')        
            else:  
                self.learnnlc = stdp.STDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=-1.0, \
                    maxadj_antiHebb=1.0, maxweight=1.0, 
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')                                                     
        elif self.netparams['poxunittype'] == 'IzhiUnit': 
            if self.netparams['prxexcitatory']:
                self.learnnlc = izhi.IzhiSTDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=10.0, \
                    maxadj_antiHebb=-10.0, maxweight=30.0, \
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')        
            else:  
                self.learnnlc = izhi.IzhiSTDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=-10.0, \
                    maxadj_antiHebb=10.0, maxweight=30.0, \
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')                                    
        else:   # IntFire1Unit
            if self.netparams['prxexcitatory']:
                self.learnnlc = if1.IntFire1STDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
                    maxadj_antiHebb=-1.0, maxweight=4.0, \
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')        
            else:  
                self.learnnlc = if1.IntFire1STDPConn(self.prx, self.pox, weight=0.0, \
                    delay=1.0, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=-1.0, \
                    maxadj_antiHebb=1.0, maxweight=4.0, 
                    soft_thresh=self.netparams['wtadjsoftthresh'], \
                    label='learnnlc')  
                            
    def setdefaultparams(self):
        # Set up what type of unit is used for prx.
        self.netparams['prxunittype'] = 'IntFire1Unit'
        
        # Set up what type of unit is used for pox.
        self.netparams['poxunittype'] = 'IntFire1Unit'

        # Set whether the pre-synaptic STDP test cell is excitatory.
        self.netparams['prxexcitatory'] = True
        
        # Set whether soft-thresholding is used for weight adjustments.
        self.netparams['wtadjsoftthresh'] = False
        
    def prenrnrun(self):
        if self.prx.classlabel == 'INTF6Unit':
            self.prx.hobj.global_init()   # just need to call once
        elif self.pox.classlabel == 'INTF6Unit':
            self.pox.hobj.global_init()   # just need to call once          
        
    def postnrnrun(self):
        if self.prx.classlabel == 'INTF6Unit':
            self.prx.hobj.global_fini()   # just need to call once
            self.prx.hobj.spkstats2(1)   # will need for each cell /  column? not sure
            print 'TMAX: %f' % h.tmax_INTF6                 
        elif self.pox.classlabel == 'INTF6Unit':
            self.pox.hobj.global_fini()   # just need to call once
            self.pox.hobj.spkstats2(1)   # will need for each cell /  column? not sure
            print 'TMAX: %f' % h.tmax_INTF6                          
                              
class STDPTestSim(nsim.NeuralSimulation):
    """
    Class for the a sim for the STDP tester model.

    Methods:
      __init__(self): constructor
      setupstims(self): set up the stimuli for the simulation
      setuprecordings(self): set up the recordings for a run
      setdefaultparams(self): set default values for the simulation parameters  
      globalfinits(self): FInitializeHandler called in beginning stages of run
      ...
      
    Attributes:
      classlabel (class): text label for the class
      simparams: dictionary of simulation (as opposed to network) parameters
        'simduration': simulation duration (in ms)    
        'trainLTP': is the network to be trained to do LTP?        
        'stimstarttime': time of first stimulus spike of the sim
        'stimspkinterval': interval of repeated spiking of NetStimUnits
        'stimsleadlaginterval': lead / lag phase interval for stims                        
                
    Usage:
    >>>thesim = STDPTestSim(label='STDPTestSim')
    """

    classlabel = 'STDPTestSim'

    def __init__(self, thenet=None, params={}, filename='', label='STDPTestSim'):
        # Set superclass parameters.
        super(STDPTestSim,self).__init__(thenet, params, filename, label)
        
        # If we need to initialize a new network with (read in) params, do so.
        if self.initnewnetwork:
            self.thenetwork = STDPTestModel(self.networkparams)

        # Set up the stimuli.
        self.setupstims()

        # Set up the recordings.
        self.setuprecordings()
               
    def setupstims(self):
        # Set the time that the first spike happens.
        firstspktime = self.simparams['stimstarttime'] 
        
        # Set the repeated spiking interval.
        spkInterv = self.simparams['stimspkinterval'] 
        
        # Set the lead / lag phase interval between the NetStimUnit spikes.
        diffInterv = self.simparams['stimsleadlaginterval'] 
                         
        # Set up the number of training spikes.
        numtrainspks = self.simparams['simduration'] / 3.0 / \
            self.simparams['stimspkinterval']
                    
        # If we are implementing the E->X test...
        if self.thenetwork.netparams['prxexcitatory']:
            # If we are doing an LTP training test...
            if self.simparams['trainLTP']:
                # Make the pre-synaptic NetStimUnit spike from beginnng to end.
                nst1spkstart = firstspktime
                nst1spknumber = 1e9
                if self.thenetwork.netparams['prxunittype'] == 'SPICellUnit': 
                    nst1wtval = 1.0
                else:
                    nst1wtval = 30.0
                
                # Make the post-synaptic NetStimUnit spike at training start time 
                # with lag offset.
                #nst2spkstart = self.simparams['simduration']  / 3.0 + diffInterv 
                nst2spkstart = firstspktime + diffInterv                 
                #nst2spknumber = numtrainspks
                nst2spknumber = 1e9
                if self.thenetwork.netparams['poxunittype'] == 'SPICellUnit': 
                    nst2wtval = 1.0
                else:
                    nst2wtval = 30.0 
                          
                # Start with the weight low.
                if self.thenetwork.netparams['poxunittype'] == 'INTF6Unit':           
                    startlearnwt = 0.5
                else:
                    startlearnwt = 0.0
            # Else, if we are doing an LTD test...
            else:
                # Make the pre-synaptic NetStimUnit spike from beginnng to end,
                # but with lag offset.
                nst1spkstart = firstspktime + diffInterv
                nst1spknumber = 1e9
                if self.thenetwork.netparams['prxunittype'] == 'SPICellUnit': 
                    nst1wtval = 1.0
                else:
                    nst1wtval = 30.0
                
                # Make the post-synaptic NetStimUnit spike at training start time 
                # with lag offset.
                #nst2spkstart = self.simparams['simduration']  / 3.0
                nst2spkstart = firstspktime
                #nst2spknumber = numtrainspks
                nst2spknumber = 1e9
                if self.thenetwork.netparams['poxunittype'] == 'SPICellUnit': 
                    nst2wtval = 1.0
                else:
                    nst2wtval = 30.0                
                
                # Start with the weight high. 
                if self.thenetwork.netparams['poxunittype'] == 'INTF6Unit':                              
                    startlearnwt = 30.0
                elif self.thenetwork.netparams['poxunittype'] == 'SPICellUnit':                              
                    startlearnwt = 1.0                    
                elif self.thenetwork.netparams['poxunittype'] == 'IzhiUnit':                              
                    startlearnwt = 30.0                    
                else:
                    startlearnwt = 4.0
        # Else (if we are implementing the I->X test...
        else:
            # If we are doing an LTP training test...           
            if self.simparams['trainLTP']:
                # Make the pre-synaptic NetStimUnit spike at training start time 
                # with lag offset.
                #nst1spkstart = self.simparams['simduration']  / 3.0 + diffInterv 
                nst1spkstart = firstspktime + diffInterv
                #nst1spknumber = numtrainspks  
                nst1spknumber = 1e9 
                if self.thenetwork.netparams['prxunittype'] == 'SPICellUnit': 
                    nst1wtval = 1.0
                else:
                    nst1wtval = 30.0           
                
                # Make the post-synaptic NetStimUnit spike from beginnng to end.
                nst2spkstart = firstspktime
                nst2spknumber = 1e9  
                if self.thenetwork.netparams['poxunittype'] == 'SPICellUnit': 
                    nst2wtval = 1.0
                else:
                    nst2wtval = 30.0     
                         
                # Start with the weight low.    
                if self.thenetwork.netparams['poxunittype'] == 'INTF6Unit':                            
                    startlearnwt = 0.5
                else:
                    startlearnwt = 0.0
            # Else, if we are doing an LTD test...              
            else:
                # Make the pre-synaptic NetStimUnit spike at training start time 
                # with lag offset.
                #nst1spkstart = self.simparams['simduration']  / 3.0
                nst1spkstart = firstspktime
                #nst1spknumber = numtrainspks
                nst1spknumber = 1e9
                if self.thenetwork.netparams['prxunittype'] == 'SPICellUnit': 
                    nst1wtval = 1.0
                else:
                    nst1wtval = 30.0
                                
                # Make the post-synaptic NetStimUnit spike from beginnng to end,
                # but with lag offset.
                nst2spkstart = firstspktime + diffInterv
                nst2spknumber = 1e9
                if self.thenetwork.netparams['poxunittype'] == 'SPICellUnit': 
                    nst2wtval = 2.0  # more weight needed to overcome inhibition
                else:
                    nst2wtval = 50.0  # more weight needed to overcome inhibition 
                         
                # Start with the weight high.          
                if self.thenetwork.netparams['poxunittype'] == 'INTF6Unit':                           
                    startlearnwt = 30.0
                elif self.thenetwork.netparams['poxunittype'] == 'SPICellUnit':                              
                    startlearnwt = 1.0                    
                elif self.thenetwork.netparams['poxunittype'] == 'IzhiUnit':                              
                    startlearnwt = 30.0                    
                else:
                    startlearnwt = 4.0
            
        # Set up NetStimUnit firing parameters.
        self.thenetwork.nst1.set_interval(spkInterv)
        self.thenetwork.nst1.set_number(nst1spknumber)
        self.thenetwork.nst1.set_start(nst1spkstart)
        self.thenetwork.nst2.set_interval(spkInterv)  
        self.thenetwork.nst2.set_number(nst2spknumber)
        self.thenetwork.nst2.set_start(nst2spkstart)
        
        # Set the initial weights as needed.
        self.thenetwork.nst1nlc.setwt(nst1wtval)
        self.thenetwork.nst2nlc.setwt(nst2wtval) 
        self.thenetwork.learnnlc.setwt(startlearnwt)      

        # Testing of SpikeCallbackStimUnit.
        #self.thenetwork.nst1.addspktimes(self, [300.0, 600.0])
        
        # Initially disable the nst2. 
        self.disablenst2cb1 = tcb.TimedCallback(self.disablenst2, self, 0.0, label='disablenst2')
        self.disablenst2cb1.verbose = 1
        
        # Enable nst2 at the start of the training segment.  
        self.enablenst2cb = tcb.TimedCallback(self.enablenst2, self, \
            self.simparams['simduration'] / 3.0, label='enablenst2')
        self.enablenst2cb.verbose = 1
                    
        # Disable nst2 at the end of the training segment.        
        self.disablenst2cb2 = tcb.TimedCallback(self.disablenst2, self, \
            self.simparams['simduration'] * 2.0 / 3.0, \
            label='disablenst2')
        self.disablenst2cb2.verbose = 1
                                      
    def setuprecordings(self):
        # Turn on spike recording for all of the NeuralUnits.
        for unit in nun.NeuralUnit.allUnits:
            unit.turnonrecspks()
            
        # Turn on spike recording for all of the NeuralConns.
        for conn in nconn.NeuralConn.allConns:
            conn.turnonrecspks()
                      
        # Turn on weights recording for the plastic synapse.
        self.thenetwork.learnnlc.turnonrecwts(recinterval)

        # Turn on membrane potential recording for prx.
        if self.thenetwork.prx.classlabel == 'INTF6Unit':
            self.thenetwork.prx.turnonrecVms(self.simparams['simduration'])
        elif self.thenetwork.prx.classlabel == 'SPICellUnit':
            self.thenetwork.prx.turnonrecvs(recinterval)            
        elif self.thenetwork.prx.classlabel == 'IzhiUnit':
            self.thenetwork.prx.turnonrecVs(self, recinterval)                           
        else:  # IntFire1Unit
            self.thenetwork.prx.turnonrecms(self, recinterval)
            
        # Turn on membrane potential recording for pox.
        if self.thenetwork.pox.classlabel == 'INTF6Unit':
            self.thenetwork.pox.turnonrecVms(self.simparams['simduration'])
        elif self.thenetwork.pox.classlabel == 'SPICellUnit':
            self.thenetwork.pox.turnonrecvs(recinterval)                
        elif self.thenetwork.pox.classlabel == 'IzhiUnit':
            self.thenetwork.pox.turnonrecVs(self, recinterval)                           
        else:  # IntFire1Unit
            self.thenetwork.pox.turnonrecms(self, recinterval)
            
        # Set up the recording calls to be called post-run().
        self.reccalls['nst1spks'] = self.thenetwork.nst1.getrecspks
        self.reccalls['nst2spks'] = self.thenetwork.nst2.getrecspks  
        self.reccalls['prxspks'] = self.thenetwork.prx.getrecspks
        self.reccalls['poxspks'] = self.thenetwork.pox.getrecspks
        self.reccalls['nst1nlcspks'] = self.thenetwork.nst1nlc.getrecspks
        self.reccalls['nst2nlcspks'] = self.thenetwork.nst2nlc.getrecspks
        self.reccalls['learnnlcspks'] = self.thenetwork.learnnlc.getrecspks
        self.reccalls['learnnlcwts'] = self.thenetwork.learnnlc.getrecwts
        if self.thenetwork.prx.classlabel == 'INTF6Unit':         
            self.reccalls['prxVms'] = self.thenetwork.prx.getrecVms
        elif self.thenetwork.prx.classlabel == 'SPICellUnit':         
            self.reccalls['prxvs'] = self.thenetwork.prx.getrecvs            
        elif self.thenetwork.prx.classlabel == 'IzhiUnit':         
            self.reccalls['prxVs'] = self.thenetwork.prx.getrecVs                         
        else:  # IntFire1Unit
            self.reccalls['prxms'] = self.thenetwork.prx.getrecms
        if self.thenetwork.pox.classlabel == 'INTF6Unit':         
            self.reccalls['poxVms'] = self.thenetwork.pox.getrecVms
        elif self.thenetwork.pox.classlabel == 'SPICellUnit':         
            self.reccalls['poxvs'] = self.thenetwork.pox.getrecvs                    
        elif self.thenetwork.pox.classlabel == 'IzhiUnit':         
            self.reccalls['poxVs'] = self.thenetwork.pox.getrecVs                         
        else:  # IntFire1Unit
            self.reccalls['poxms'] = self.thenetwork.pox.getrecms            
        self.reccalls['STDPcurve'] = self.thenetwork.learnnlc.getSTDPcurve
        
    def setdefaultparams(self):
        # Set the simulation duration (in ms).
        self.simparams['simduration'] = 900.0 
        
        # Set up whether to train using LTP or LTD.
        self.simparams['trainLTP'] = True
        
        # Set the stimulus start time (in ms).
        self.simparams['stimstarttime'] = 10.0
        
        # Set the spiking interval for the NetStimUnits (in ms).
        self.simparams['stimspkinterval'] = 50.0
        
        # Set the phase lead / lag interval (in ms).
        self.simparams['stimsleadlaginterval'] = 3.0
                    
    def globalfinits(self):
        # Run superclass method.
        super(STDPTestSim,self).globalfinits()
                                                      
    def disablenst2(self):
        self.thenetwork.nst2.disable()
        
    def enablenst2(self):
        self.thenetwork.nst2.enable()   
        
    def disableprx(self):
        self.thenetwork.prx.disable()
        
    def enableprx(self):
        self.thenetwork.prx.enable()   
               
    def save(self):
        fname = 'data/stdptests/stdptests-STDPonly_'
        if self.thenetwork.netparams['prxexcitatory']:
            fname += 'EtoX_'
        else:
            fname += 'ItoX_'
        if self.simparams['trainLTP']:
            fname += 'LTP_'
        else:
            fname += 'LTD_'
        if self.thenetwork.netparams['prxunittype'] == 'INTF6Unit':
            fname += 'INTF6prx_'
        elif self.thenetwork.netparams['prxunittype'] == 'SPICellUnit':
            fname += 'SPIprx_'          
        elif self.thenetwork.netparams['prxunittype'] == 'IzhiUnit':
            fname += 'Izhiprx_'          
        else:
            fname += 'IF1prx_'
        if self.thenetwork.netparams['poxunittype'] == 'INTF6Unit':
            fname += 'INTF6pox.sim'
        elif self.thenetwork.netparams['poxunittype'] == 'SPICellUnit':
            fname += 'SPIpox.sim'           
        elif self.thenetwork.netparams['poxunittype'] == 'IzhiUnit':
            fname += 'Izhipox.sim'          
        else:
            fname += 'IF1pox.sim'
                
        # Do the file save.
        print 'Saving to %s' % fname
        super(STDPTestSim,self).save(fname)
                                                                                                                        
    def showresults(self):
        #
        # Show the spike and membrane value results
        #
        print 'Showing the results...\n'
        
        print 'Model spikes:'
        print '     nst1 ->' , self.recdata['nst1spks']
        print '     nst2 ->' , self.recdata['nst2spks']
        print '      prx ->' , self.recdata['prxspks']
        print '      pox ->' , self.recdata['poxspks']
        print '  nst1nlc ->' , self.recdata['nst1nlcspks']
        print '  nst2nlc ->' , self.recdata['nst2nlcspks']
        print ' learnnlc ->' , self.recdata['learnnlcspks']
        
        # Plot the STDP curve.
        figure()
        self.plotSTDPcurve()
                
        # Plot unit spike times.
        #figure()
        #self.plotunitspikes('nst1')
        #figure()
        #self.plotunitspikes('nst2')        
        #figure()
        #self.plotunitspikes('prx')
        #figure()
        #self.plotunitspikes('pox')
                
        # Plot the learnnlc weights over time.
        figure()
        self.plotweightchange()
        
        # Plot the unit voltages and stimulator spikes.
        figure()
        self.plotunitvoltages('prx', stimname='nst1')
        figure()
        self.plotunitvoltages('pox', stimname='nst2')
                                                
    def EtoXtrainingfig(self, saveFile=False, savedir=''):
        # Load the save directory.
        if (savedir == ''):
            savedir = '.'
           
        # Make the figure.
        figure(figsize=(12,8))
        #subplots_adjust(hspace = 0.3)
        subplots_adjust(wspace = 0.3)
        
        # Make sure we're set to the LTP case.
        thesim.load('data/stdptests/stdptests-EtoX_LTP_900ms.sim')          
                     
        # Plot the STDP curve.
        subplot(2,3,1)
        self.plotSTDPcurve()
        title('STDP Curve')

        # Plot the learnnlc weights over time.
        subplot(2,3,2)        
        self.plotweightchange()  
        title('LTP Training')
        xlabel('')
        axvline(self.simparams['simduration'] / 3.0, color='r')
        axvline(self.simparams['simduration'] * 2.0 / 3.0, color='r')
                
        # Plot the pox activity with stimuli.
        subplot(2,3,5)        
        self.plotpostcelllearning()      
        title('')
        axvline(self.simparams['simduration'] / 3.0, color='r')
        axvline(self.simparams['simduration'] * 2.0 / 3.0, color='r')
                                            
        # Switch to the LTD case.
        thesim.load('data/stdptests/stdptests-EtoX_LTD_900ms.sim')

        # Plot the learnnlc weights over time.
        subplot(2,3,3)        
        self.plotweightchange()    
        title('LTD Training') 
        xlabel('')
        ylabel('')
        axvline(self.simparams['simduration'] / 3.0, color='r')
        axvline(self.simparams['simduration'] * 2.0 / 3.0, color='r')
                         
        # Plot the pox activity with stimuli.
        subplot(2,3,6)        
        self.plotpostcelllearning()   
        title('')
        ylabel('')
        axvline(self.simparams['simduration'] / 3.0, color='r')
        axvline(self.simparams['simduration'] * 2.0 / 3.0, color='r')
        
        # Save the file if we want to.
        if (saveFile):
            savefig('%s/STDPtests.png' % savedir)        
                                  
    def plotSTDPcurve(self):
        xmarg = 1     # x display margin
        ymarg =     0.1 # y display margin
        tvec, vec = self.recdata['STDPcurve']
        fill_between(tvec,vec,0)
        title('learnnlc STDP Curve')
        xlabel('Pre- to post-synaptic delay (ms)')
        ylabel('Weight change')
        xlim((-xmarg + tvec.min(), xmarg + tvec.max()))
        ylim((-ymarg + vec.min(), ymarg + vec.max()))
        
    def plotunitspikes(self, unitname):
        print type(unitname)        
        xmarg = 10   # x display margin
        spktimes = self.recdata['%sspks' % unitname]
        if len(spktimes) > 0:
            plot_spike_times(spktimes)
        title('%s Spike Times' % unitname)
        xlabel('Time (ms)')
        xlim((-xmarg, self.simparams['simduration'] + xmarg))
                    
    def plotweightchange(self):
        xmarg = 10      # x display margin
        ymarg =     1     # y display margin        
        tvec, vec = self.recdata['learnnlcwts']
        plot(tvec, vec)
        title('learnnlc Weight Modification')
        xlabel('Time (ms)')
        ylabel('Weight')
        xlim((-xmarg, self.simparams['simduration'] + xmarg))
        ylim((-ymarg + vec.min(), ymarg + vec.max()))
        
    def plotunitvoltages(self, unitname, stimname=None):
        xmarg = 10    # x display margin
        ymarg = 0.1   # y display margin
        if unitname == 'prx':
            unittype = self.networkparams['prxunittype']
        elif unitname == 'pox':
            unittype = self.networkparams['poxunittype']        
        else:   
            unittype = eval('self.thenetwork.%s.classlabel' % unitname)
        #unitobj = eval('self.thenetwork.%s' % unitname)
        if unittype == 'INTF6Unit':
            tvec, vec = self.recdata['%sVms' % unitname]
        elif unittype == 'SPICellUnit':
            tvec, vec = self.recdata['%svs' % unitname]            
        elif unittype == 'IzhiUnit':
            tvec, vec = self.recdata['%sVs' % unitname]                       
        else:  # IntFire1Unit
            tvec, vec = self.recdata['%sms' % unitname]         
        plot(tvec, vec)
        title('%s Membrane Potentials' % unitname)
        xlabel('Time (ms)')
        if unittype == 'INTF6Unit':
            ylabel('Vm (mV)')
        elif unittype == 'SPICellUnit':
            ylabel('v (mV)')             
        elif unittype == 'IzhiUnit':
            ylabel('V (mV)')             
        else:  # IntFire1Unit
            ylabel('m value')
            ylim(-ymarg-1.0, 2.0 + ymarg)
        xlim((-xmarg, self.simparams['simduration'] + xmarg))
        spktimes = self.recdata['%sspks' % unitname]
        for ind in range(len(spktimes)):
            axvline(spktimes[ind], color='r')   
        if stimname != None:
            for stimtime in self.recdata['%sspks' % stimname]:
                if unittype == 'INTF6Unit':
                    plot(stimtime, -70.0, 'go')  
                elif unittype == 'SPICellUnit':
                    plot(stimtime, -110.0, 'go')                     
                elif unittype == 'IzhiUnit':
                    plot(stimtime, -65.0, 'go')                                     
                else:  # IntFire1Unit
                    plot(stimtime, -0.02, 'go') 
        # This is commented out because it doesn't work properly with saved 
        # NeuralSimulations.  To make it work correctly, I may need to save the
        # values that are to be plotted.
        #if unittype == 'INTF6Unit':      
            #axhline(unitobj.get_VRMP(), color='k',linestyle=':')
            #axhline(unitobj.get_VTH(), color='k',linestyle='--')
            #axhline(unitobj.get_Vblock(), color='r',linestyle='--')
                                        
    def plotpostcelllearning(self):
        xmarg = 10   # x display margin
        ymarg = 20   # y display margin        
        tvec, vec = self.recdata['poxVms']
        plot(tvec, vec)
        title('pox Membrane Potentials')
        xlabel('Time (ms)')
        ylabel('Vm (mV)')
        xlim((-xmarg, self.simparams['simduration'] + xmarg))
        ylim((-ymarg + vec.min(), ymarg + vec.max()))
        for stimtime in self.recdata['nst1spks']:
            plot(stimtime, -74.0, 'go')
        for stimtime in self.recdata['nst2spks']:
            plot(stimtime, -70.0, 'ro')
                                    
#
# Globals
#

# Set up directory and file names.
datadir = 'data'
#simdir =  '%s/testsim' % datadir
#simfstem = '%s/testsim-' % simdir
simdir =  '%s/stdptests' % datadir
simfstem = '%s/stdptests-' % simdir
simfilename = '%ssim' % simfstem

#
# Script code to run.
#

#
# Set up network and simulation parameters.
#

# Create an empty dictionary for all of the parameters.
theparams = {}

# Set whether we turn off even initializing the network on initial run.
theparams['noinit'] = False

# Set whether we turn off running the sim.
theparams['norun'] = False

# Set whether we turn off showing the results of the sim.
theparams['noshow'] = False

# Set whether we turn off showing the results of the sim.
theparams['dosave'] = False

# Set file to load a sim from.
theparams['loadsim'] = None

# Read in what parameters exist on the command line.
cmdlineparams(theparams)

# Set up what type of unit is used for prx.
#theparams['prxunittype'] = 'INTF6Unit'

# Set up what type of unit is used for pox.
#theparams['poxunittype'] = 'INTF6Unit'

# Set whether the pre-synaptic STDP test cell is excitatory.
#theparams['prxexcitatory'] = False # True

# Set whether soft-thresholding is used for weight adjustments.
#theparams['wtadjsoftthresh'] = False


# Set the simulation duration (in ms).
#theparams['simduration'] = 900.0

# Set up whether to train using LTP or LTD.
#theparams['trainLTP'] = False # True

# Set the stimulus start time (in ms).
#theparams['stimstarttime'] = 10.0

# Set the spiking interval for the NetStimUnits (in ms).
#theparams['stimspkinterval'] = 50.0

# Set the phase lead / lag interval (in ms).
#theparams['stimsleadlaginterval'] = 3.0

# If we are loading from an existing sim...
if theparams['loadsim'] != None:
    # Load the simulation from the specified file.
    #thesimfile = theparams['loadsim']
    thesimfile = '%s/%s' % (simdir, theparams['loadsim'])
    thesim = STDPTestSim(params=theparams, filename=thesimfile)
    
    # Show the results, if we haven't disabled.
    if not theparams['noshow']: 
        thesim.showresults()
    
# If we're not disabling initialization...
elif not theparams['noinit']:
    # Initialize the network.
    print 'Initializing network...'
    #thenet = STDPTestModel()  # just using default parameters
    thenet = STDPTestModel(theparams)  # passing in parameters
    print
    
    # Initialize a new simulation.
    #thesim = STDPTestSim(thenet)  # just using default parameters
    thesim = STDPTestSim(thenet, theparams)
    
    # Turn on the verbosity for the learning weight.
    thenet.learnnlc.wtadjobj.verbose = 1

    # If we're not disabling running...
    if not theparams['norun']:    
        # Run the simulation
        print 'Running the simulation...\n'
        thesim.run()
        print
        
        # Show the results, if we haven't disabled.
        if not theparams['noshow']: 
            thesim.showresults()
            
        # Do a save, if we want to.
        if theparams['dosave']:
            thesim.save()
