"""
SPIcell.py -- module for wrapping SPICell neural units, which are multi-
  compartmental units developed from Ben Suter's model of cortico-spinal
  pyramidal cells in motor cortex

Classes for SPICell neural units:
  SPICellUnit (subclass of SpikeUnit)
    
Initial design and implementation: 
  George Chadderdon, Ashutosh Mohan, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import neural unit module.
import neuralunit as nun

# Load the neural connections module.
import neuralconn as nconn

# Load the stdp module.
import stdp

# Load the STDP RL module.
import stdp_rl as stdprl

# Import numpy.
import numpy

# Import mlab.
import matplotlib.mlab

class SPICellUnit(nun.SpikeUnit):
    """
    SPICell unit for the model.  For non-plastic connections with a post-synaptic 
    SPICellUnit type, you can use NeuralConn, and for plastic connections, it's 
    okay to use STDPConn or STDPRLConn.

    Methods:
      __init__(self, etc.): constructor
      getinconnattachpoint(self): return a hoc object for NetCons using this unit
        as a destination       
      getoutconnattachpoint(self): return a hoc object for NetCons using this unit
        as a source      
      turnonrecspks(self): turns on the recording of spikes during the sim in recspks        
      turnonrecvs(self,recinterval): turn on recording of membrane potentials every
        interval
      getrecvs(self): return tvec and amplitudes for the membrane potentials
      disable(self): disables the unit so no output comes out
      enable(self): enables or reenables the unit's output       
      isdisabled(self): returns whether the unit is disabled
                                                     
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances
      hobj: hoc SPICell object
      recvsinterval: recording interval for voltage values in ms.       
      recvs: hoc Vector object for recorded membrane voltages
      proj2soma: does this project to the soma?  (default means projects to dendrite)
      recsegment: NEURON segment that output recording is done from

    Usage:
    - Uses: 
      PTcell.hoc (load_file() this) 
      
    >>>sp1 = spic.SPICellUnit()
    """

    classlabel = 'SPICellUnit'
    numCreated = 0
    allUnits = []    
                
    def __init__(self, NTs=['GLU'], proj2soma=False, label=''):
        # Set superclass parameters.
        super(SPICellUnit,self).__init__(NTs, label)

        # Create the hoc object for the SPI cell.
        self.hobj = h.PTcell(0, 0)   # zloc, type values (not sure what these ought to be)

        # Set whether it projects to the soma.
        self.proj2soma = proj2soma
        
        # Set up the output recording compartment.
        self.recsegment = self.hobj.axon(1)
        
        # Set the output connection Section to the axon.
        self.outconnsect = self.hobj.axon 
         
        # Set the output threshold for NetCons.
        self.outconnthresh = 0.0
        
        # Set up a hoc Vectors for Vm recording.
        self.recvs = h.Vector()

        # Increment the instance count.
        SPICellUnit.numCreated += 1
        SPICellUnit.allUnits.append(self)

    def getinconnattachpoint(self, src):
        # Attach to the appropriate (excitatory or inhibitory) synapse
        # mechanism.
        if src.isexcitatory():
            return self.hobj.esyn
        else:
            return self.hobj.isyn
        
    def getoutconnattachpoint(self):
        # Return the recording segment.    
        return self.hobj.axon(1)._ref_v
        
    def turnonrecspks(self):
        if not self.isrecording():
            # Set up a recording NetCon.
            self.recnetcon = h.NetCon(self.getoutconnattachpoint(), None, \
                sec=self.outconnsect)
            self.recnetcon.threshold = self.outconnthresh
        
            # Set up a hoc Vector for spike recording.
            self.recspks = h.Vector()
        
            # Turn on the recording.
            self.recnetcon.record(self.recspks)
                       
    def turnonrecvs(self, recinterval):
        # Turn on the recording for a particular recording interval.
        #self.recvs.record(self.hobj.soma(0.5)._ref_v, recinterval)
        self.recvs.record(self.recsegment._ref_v, recinterval)
        
        # Remember the recording interval.
        self.recvsinterval = recinterval

    def getrecvs(self):
        recvs = numpy.array(self.recvs)
        tvec = numpy.linspace(0.0, self.recvsinterval * len(recvs), \
             num=len(recvs), endpoint=False)
        return tvec, recvs
                     
    def disable(self):
        pass
        
    def enable(self):
        pass
        
    def isdisabled(self):
        return False              
