"""
database.py -- module for classes for databases (e.g. sqlite3, NQS)

Classes for databases:
  Database (base class)
      NQSDB
      DictDB
      
Initial design and implementation: George Chadderdon, 2013
 
Last update: 7/29/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import numpy.
import numpy

# Load functions for interfacing with hoc data structures.
from hocinterface import *

class Database(object):
    """
    Generic database.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      instlabel: instance label (usable for tagging)

    Usage:
    >>>db1 = database.Database()
    """

    classlabel = 'GenDatabase'
    numCreated = 0

    def __init__(self, label=''):
        # Set a NULL instance label.
        self.instlabel = label

        # Increment the instance count.
        Database.numCreated += 1

class NQSDB(Database):
    """
    NQS (hoc object) based database.

    Methods:
      __init__(self): constructor
      [added lots of stuff...]

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      hobj: hoc NQS object
      colnames: list of strings of column names

    Usage:
    - vecst.mod needs to be installed in special.
    - Uses: 
      /usr/site/nrniv/local/hoc/nqs.hoc (load_file() this is sufficient)
      /usr/site/nrniv/local/hoc/decvec.hoc
      /usr/site/nrniv/local/hoc/declist.hoc
      /usr/site/nrniv/simctrl/hoc/setup.hoc

    >>>h.load_file("nqs.hoc")  # load NQS code
    >>>nq = database.NQSDB()

    NOTE: The element values are assumed to be only doubles here, though NQS
    Also allows string and object objects to be stored.
    """

    classlabel = 'NQSDatabase'
    numCreated = 0

    def __init__(self, label='', colnames=['rowid','col1'], fname=''):
        # Set superclass parameters.
        super(NQSDB,self).__init__(label)

        # Save the list of column names.
        self.colnames = colnames

        # If we have a filename, init and then load from the file.
        if (fname != ''):
            self.hobj = h.NQS()
            self.load(fname)

        # Otherwise, create the hoc NQS object using the column names passed in.
        else:
            # If we have <= 20 columns...
            if len(colnames) <= 20:
                # Init them as normal.
                colargs = '"%s"' % colnames[0]
                for colname in colnames[1:]:
                    colargs += ',"%s"' % colname
                exec('self.hobj = h.NQS(%s)' % colargs)
            # Otherwise, if we have too many...
            else:
                # Do the first 20 as normal.
                colargs = '"%s"' % colnames[0]
                for colname in colnames[1:20]:
                    colargs += ',"%s"' % colname
                exec('self.hobj = h.NQS(%s)' % colargs)

                # Do the rest with resize commands.
                for colname in colnames[20:]:
                    self.hobj.resize(colname)

        # Increment the instance count.
        NQSDB.numCreated += 1

    def getnumrows(self):
        self.hobj.verbose = 0  # still doesn't get rid of display :(
        numrows = int(self.hobj.size())
        self.hobj.verbose = 1
        return numrows

    def getnumcols(self):
        return len(self.colnames)

    def showcolheaders(self):
        self.hobj.gethdrs()

    def showall(self):
        self.hobj.pr()

    def addrow(self,numlist):
        numargs = '%f' % numlist[0]
        for ind in range(len(numlist) - 1):
            numargs += ',%f' % numlist[ind + 1]
        exec('self.hobj.append(%s)' % numargs)

    def getcol(self,colid='rowid'):
        return hv2narr(self.hobj.getcol(colid))

    def getrow(self,rownum=0):
        # Select the desired row (index 0-based).
        self.hobj.select("IND_",rownum)

        # Extract the row numbers.
        rownums = []
        for ind in range(self.getnumcols()):
            rownums.append(self.getcol(self.colnames[ind])[0])

        # Toggle back to the non-selected DB.
        self.hobj.tog()

        # Return the vector as a numpy array.
        return numpy.array(rownums)

    def getrowcolelem(self,rownum,colid):
        # Get the column.
        thecol = self.getcol(colid)

        # Return the rownum element of the column.
        return thecol[rownum]

    def setrowcolelem(self,theval,rownum,colid):
        # Get the column index number.
        colnum = self.colnames.index(colid)

        # Set the value.
        self.hobj.v[colnum].x[rownum] = theval

    def select(self,selectlist):
        selectargs = '"%s"' % selectlist[0]   # always is a string
        for ind in range(len(selectlist) - 1):
            if isinstance(selectlist[ind + 1], (int, long, float)):
                selectargs += ', %f' % selectlist[ind + 1]
            else:
                selectargs += ',"%s"' % selectlist[ind + 1]
        exec('numfoundrows = self.hobj.select(%s)' % selectargs)
      
        # Return the number of rows found.
        return int(numfoundrows)

    def switchfulldb(self):
        self.hobj.tog('db')

    def save(self,fname='mynq.nqs'):
        self.hobj.sv(fname)

    def load(self,fname='mynq.nqs'):
        # Read the file in.
        self.hobj.rd(fname)

        # Get the number of columns.
        numcols = int(self.hobj.m[0])

        # Rebuild the column names list.
        self.colnames = []
        for ind in range(numcols):
            self.colnames.append(self.hobj.s[ind].s)


class DictDB(Database):
    """
    A simple Python dictionary-based database.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      columns: Python dictionary for the columns

    Usage:
    >>>dictdb = database.NQSDB()
    """

    classlabel = 'DictDatabase'
    numCreated = 0

    def __init__(self, label=''):
        # Set superclass parameters.
        super(DictDB,self).__init__(label)

        # Create the main dictionary.
        self.columns = {}

        # Increment the instance count.
        DictDB.numCreated += 1
