"""
intf6.py -- module for wrapping INTF6 neural units

Classes for INTF6 neural units:
  INTF6Unit (subclass of SpikeUnit)
  INTF6Conn (subclass of NeuralConn)
    INTF6STDPConn (also inherits STDPConn)
    INTF6STDPRLConn (also inherits STDPRLConn)
        
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import neural unit module.
import neuralunit as nun

# Load the neural connection module.
import neuralconn as nconn

# Load the stdp module.
import stdp

# Load the STDP RL module.
import stdp_rl as stdprl

# Import numpy.
import numpy

# Import mlab.
import matplotlib.mlab

class INTF6Unit(nun.SpikeUnit):
    """
    INTF6 unit for the model.  For non-plastic connections with a post-synaptic 
    INTF6Unit type, you should use INTF6Conn instead of NeuralConn, and
    for plastic connections, INTF6STDPConn or INTF6STDPRLConn.

    Methods:
      get_EAM() (class): get AMPA reversal potential relative to V_RMP
      set_EAM(value) (class): set AMPA reversal potential relative to V_RMP 
      get_ENM() (class): get NMDA reversal potential relative to V_RMP
      set_ENM(value) (class): set NMDA reversal potential relative to V_RMP 
      get_EGA() (class): get GABA reversal potential relative to V_RMP
      set_EGA(value) (class): set GABA reversal potential relative to V_RMP 
      get_mg() (class): get Mg++ - related parameter
      set_mg(value) (class): set Mg++ - related parameter
      __init__(self): constructor
      turnonrecVms(self,recinterval): turn on recording of membrane potentials every
        interval
      getrecVms(self): return tvec and amplitudes for the membrane potentials
      get_VRMP(self): get the resting membrane voltage (in mV) of the unit
      set_VRMP(self,value): set the resting membrane voltage (in mV) of the unit
      get_VTH(self): get the firing threshold voltage (mV) of the unit
      set_VTH(self,value): set the firing threshold voltage (mV) of the unit
      get_Vblock(self): get the depolarization blockade voltage (mV) of the unit
      set_Vblock(self,value): set the depolarization blockade voltage (mV) of the unit
      get_ahpwt(self): get the after-hyperpolarization weight of the unit
      set_ahpwt(self,value): set the after-hyperpolarization weight of the unit     
      get_tauahp(self): get the after-hyperpolarization decay time constant (in ms) of the unit
      set_tauahp(self,value): set the after-hyperpolarization decay time constant (in ms) of the unit
      get_refrac(self): get the absolute refractory period (in ms) of the unit
      set_refrac(self,value): set the absolute refractory period (in ms) of the unit
      get_RRWght(self): get the weight for the relative refractory threshold adjustment of the unit
      set_RRWght(self,value): set the weight for the relative refractory threshold adjustment of the unit      
      get_tauRR(self): get the decay time constant (in ms) for the relative refractory threshold adjustment of the unit
      set_tauRR(self,value): set the decay time constant (in ms) for the relative refractory threshold adjustment of the unit  
      get_tauAM(self): get the somatic AMPA synapse decay time constant (in ms) of the unit
      set_tauAM(self,value): set the somatic AMPA synapse decay time constant (in ms) of the unit  
      get_tauAM2(self): get the dendritic AMPA synapse decay time constant (in ms) of the unit
      set_tauAM2(self,value): set the dendritic AMPA synapse decay time constant (in ms) of the unit  
      get_tauNM(self): get the somatic NMDA synapse decay time constant (in ms) of the unit
      set_tauNM(self,value): set the somatic NMDA synapse decay time constant (in ms) of the unit  
      get_tauNM2(self): get the dendritic NMDA synapse decay time constant (in ms) of the unit
      set_tauNM2(self,value): set the dendritic NMDA synapse decay time constant (in ms) of the unit  
      get_tauGA(self): get the somatic GABA A synapse decay time constant (in ms) of the unit
      set_tauGA(self,value): set the somatic GABA A synapse decay time constant (in ms) of the unit  
      get_tauGA2(self): get the dendritic GABA A synapse decay time constant (in ms) of the unit
      set_tauGA2(self,value): set the dendritic GABA A synapse decay time constant (in ms) of the unit 
      disable(self): disables the unit so no output comes out
      enable(self): enables or reenables the unit's output       
      isdisabled(self): returns whether the unit is disabled
                                                     
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances
      recVmsinterval (class): recording interval for Vm values in ms.  This is a 
        class variable because vdt is a GLOBAL in intf6.mod.
      hobj: hoc INTF6 object
      recVms: hoc Vector object for recorded membrane voltages (Vm)
      recVmstvec: hoc Vector object for times for recorded membrane voltages (Vm)
      proj2soma: does this project to the soma?  (default means projects to dendrite)

    Usage:
    - vecst.mod needs to be installed in special.
    - stats.mod needs to be installed in special.
    - intf6.mod needs to be installed in special.
    - Uses: 
      /usr/site/nrniv/local/hoc/declist.hoc (load_file() this) 
      /usr/site/nrniv/local/hoc/decvec.hoc (load_file() this)    
      /usr/site/nrniv/local/hoc/labels.hoc (load_file() this)
      
    >>>i1 = intf6.INTF6Unit(h.INTF6[0])
    """

    classlabel = 'INTF6Unit'
    numCreated = 0
    allUnits = []    
    recVmsinterval = h.vdt_INTF6
    
    @staticmethod
    def get_EAM():
        return h.EAM_INTF6
        
    @staticmethod
    def set_EAM(value):
        h.EAM_INTF6 = value

    @staticmethod
    def get_ENM():
        return h.ENM_INTF6
        
    @staticmethod
    def set_ENM(value):
        h.ENM_INTF6 = value
        
    @staticmethod
    def get_EGA():
        return h.EGA_INTF6
        
    @staticmethod
    def set_EGA(value):
        h.EGA_INTF6 = value
        
    @staticmethod
    def get_mg():
        return h.mg_INTF6
        
    @staticmethod
    def set_mg(value):
        h.mg_INTF6 = value
                
    def __init__(self,thecell=None,NTs=['GLU'],proj2soma=False,celltype=h.E5B,\
        VRMP=-65.0,VTH=-37.5,Vblock=-25.0,ahpwt=1.0,tauahp=10.0,refrac=5.0,\
        RRWght=0.25,tauRR=8.0,tauAM=10,tauAM2=20.0,tauNM=300.0,tauNM2=300.0,\
        tauGA=10.0,tauGA2=20.0,\
        label=''):
        # Set superclass parameters.
        super(INTF6Unit,self).__init__(NTs, label)

        # If no cell is passed in, make a new one.
        if thecell == None:
            # Set up parameters and create the new cell object.
            cellid = 0
            if self.isinhibitory():
                cellinhib = 1
            else:
                cellinhib = 0
            columnid = INTF6Unit.numCreated  + 1 # take off +1 when not running in code with existing column
            self.hobj = h.INTF6(cellid,celltype,cellinhib,columnid)   # cell id, cell type, inhibitory?, column id
            
            # Set up an independent INTF6 column for the single cell.  
            # Hopefully this will allow parallelization.
            self.clist = h.List()
            self.clist.append(self.hobj)
            h('double ix%d[50]' % columnid)
            exec('self.ix = h.ix%d' % columnid)
            h('double ixe%d[50]' % columnid)     
            exec('self.ixe = h.ixe%d' % columnid)                  
            h('double numc%d[50]' % columnid)
            exec('self.numc = h.numc%d' % columnid)        
            self.numc[int(celltype)] = 1    
            h('objref xo')
            h.xo = self.hobj
            h('objref ce')
            h.ce = self.clist
            h('xo.cinit(ce,%d, &ix%d, &ixe%d, &numc%d)' % (columnid, columnid, columnid, columnid))   
            # This didn't work...       
            #self.hobj.cinit(self.clist,columnid,self._ref_ix,self._ref_ixe,self._ref_numc)
        # Otherwise, set the hoc cell to what's passed in.           
        else:
            self.hobj = thecell

        # Turn off Jitcon for the cell.
        self.hobj.flag("jcn",0)

        # Enable NetCon events in the cell.
        self.hobj.flag("out",1)

        # Set whether it projects to the soma.
        self.proj2soma = proj2soma

        # Set up the unit parameters if a new cell was created.
        if thecell == None:
            self.set_VRMP(VRMP)
            self.set_VTH(VTH)
            self.set_Vblock(Vblock)
            self.set_ahpwt(ahpwt)
            self.set_tauahp(tauahp)
            self.set_refrac(refrac)
            self.set_RRWght(RRWght)
            self.set_tauRR(tauRR)
            self.set_tauAM(tauAM)
            self.set_tauAM2(tauAM2)
            self.set_tauNM(tauNM)
            self.set_tauNM2(tauNM2)
            self.set_tauGA(tauGA)
            self.set_tauGA2(tauGA2)
        
        # Set up a hoc Vectors for Vm recording.
        self.recVms = h.Vector()
        self.recVmstvec = h.Vector()

        # Increment the instance count.
        INTF6Unit.numCreated += 1
        INTF6Unit.allUnits.append(self)
        
    def postnrnrun(self):
        self.hobj.fini()
                   
    def turnonrecVms(self, simduration):
        # Resize the recVms vectors so they are right size for initrec().
        npacsz = 1.2  # not sure what significance of this scale factor is
        self.recVms.resize(npacsz * simduration / INTF6Unit.recVmsinterval)
        self.recVmstvec.resize(npacsz * simduration / INTF6Unit.recVmsinterval)

        # Turn on the recording for a particular recording interval.
        self.hobj.initrec('time', self.recVmstvec)  # tvec
        self.hobj.initrec('VM', self.recVms)        # vec

        # This method doesn't quite work correctly because it doesn't show Vm
        # dynamics between unit state updates.
#        self.recVms.record(self.hobj._ref_Vm, INTF6Unit.recVmsinterval)

    def getrecVms(self):
        recVms = numpy.array(self.recVms)
        tvec = numpy.array(self.recVmstvec)

        # If we have zero trailers from the run, trim them off.
        if len(matplotlib.mlab.find(tvec == 0.0)) > 1:
            trimind = matplotlib.mlab.find(tvec == 0.0)[1]
            recVms = recVms[0:trimind]
            tvec = tvec[0:trimind]

        # old method for tvec, when it was assumed no tvec values were recorded
#        tvec = numpy.linspace(0.0, INTF6Unit.recVmsinterval * len(recVms), \
#             num=len(recVms), endpoint=False)
        return tvec, recVms
        
    def get_VRMP(self):
        return self.hobj.RMP
        
    def set_VRMP(self, value):
        self.hobj.RMP = value
        
    def get_VTH(self):
        return self.hobj.VTH
        
    def set_VTH(self, value):
        self.hobj.VTH = value
        
    def get_Vblock(self):
        return self.hobj.Vblock
        
    def set_Vblock(self, value):
        self.hobj.Vblock = value
        
    def get_ahpwt(self):
        return self.hobj.ahpwt
        
    def set_ahpwt(self, value):
        self.hobj.ahpwt = value
                
    def get_tauahp(self):
        return self.hobj.tauahp
        
    def set_tauahp(self, value):
        self.hobj.tauahp = value
                
    def get_refrac(self):
        return self.hobj.refrac
        
    def set_refrac(self, value):
        self.hobj.refrac = value
        
    def get_RRWght(self):
        return self.hobj.RRWght
        
    def set_RRWght(self, value):
        self.hobj.RRWght = value
        
    def get_tauRR(self):
        return self.hobj.tauRR
        
    def set_tauRR(self, value):
        self.hobj.tauRR = value
        
    def get_tauAM(self):
        return self.hobj.tauAM
        
    def set_tauAM(self, value):
        self.hobj.tauAM = value       
                
    def get_tauAM2(self):
        return self.hobj.tauAM2
        
    def set_tauAM2(self, value):
        self.hobj.tauAM2 = value     
        
    def get_tauNM(self):
        return self.hobj.tauNM
        
    def set_tauNM(self, value):
        self.hobj.tauNM = value      
               
    def get_tauNM2(self):
        return self.hobj.tauNM2
        
    def set_tauNM2(self, value):
        self.hobj.tauNM2 = value            
 
    def get_tauGA(self):
        return self.hobj.tauGA
        
    def set_tauGA(self, value):
        self.hobj.tauGA = value       
                
    def get_tauGA2(self):
        return self.hobj.tauGA2
        
    def set_tauGA2(self, value):
        self.hobj.tauGA2 = value   
             
    def disable(self):
        self.hobj.flag('dead',1)
        
    def enable(self):
        self.hobj.flag('dead',0)    
        
    def isdisabled(self):
        return (self.hobj.isdead() == 1.0)
                  
class INTF6Conn(nconn.NeuralConn):
    """
    Neural connection converging into INTF6Unit for the model.

    Methods:
      __init__(self): constructor
      getwt(self): returns the primary weight
      setwt(self,weight): sets the primary weight
      getwt2(self): returns the secondary weight
      setwt2(self,weight): sets the secondary weight
      getwtncind(self): returns the proper primary weight index for the NetCon
      getwt2ncind(self): returns the proper secondary weight index for the NetCon
      
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances      
      proj2soma: does this project to the soma of the target?  
        (default means projects to dendrite)
      
    Usage:
    >>>nlc1 = intf6.INTF6Conn(prx,pox)
    """

    classlabel = 'INTF6Conn'
    numCreated = 0
    allConns = []
    
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        label='', proj2soma=False):
        # Set superclass parameters.
        nconn.NeuralConn.__init__(self, src, dest, weight, delay, label)
        
        # Clear the first weight (set by the NeuralConn constructor).
        self.hobj.weight[0] = 0.0

        # If the source unit is an INTF6Unit, use the source unit to decide 
        # whether it projects to the soma.
        if (self.src.classlabel == INTF6Unit.classlabel):
            self.proj2soma = self.src.proj2soma
        # Otherwise, use the parameter passed into the constructor.
        else:
            self.proj2soma = proj2soma
        
        # Set the weights according to the specific rules.
        self.setwt(weight)
        self.setwt2(weight2)

        # Increment the instance count.
        INTF6Conn.numCreated += 1
        INTF6Conn.allConns.append(self)
        
    def getwt(self):
        return self.hobj.weight[self.getwtncind()]

    def setwt(self,weight):
        self.hobj.weight[self.getwtncind()] = weight

    def getwt2(self):
        wtind = self.getwt2ncind()
        if wtind != -1:
            return self.hobj.weight[wtind]
        else:
            return -1  # error condition (no valid weight secondary weight)

    def setwt2(self,weight):
        wtind = self.getwt2ncind()
        if wtind != -1:
            self.hobj.weight[wtind] = weight

    def getwtncind(self):
        # If the cell projects to soma rather than dendrite...
        if self.proj2soma:
            # If cell is excitatory, return AM index.
            if self.src.isexcitatory():
                return int(h.AM) 
            # Else, inhibitory so return GA index.
            else:
                return int(h.GA)             
        # Otherwise (cell projects to dendrite)...
        else:
            # If cell is excitatory, return AM2 index.
            if self.src.isexcitatory():
                return int(h.AM2) 
            # Else, inhibitory so return GA2 index.
            else:
                return int(h.GA2)

    def getwt2ncind(self):
        # If the cell projects to soma rather than dendrite...
        if self.proj2soma:
            # If cell is excitatory, return NM index.
            if self.src.isexcitatory():
                return int(h.NM)
            # Else, inhibitory so return -1 (error).
            else:
                return -1                
        # Otherwise (cell projects to dendrite)...
        else:
            # If cell is excitatory, return NM2 index.
            if self.src.isexcitatory():
                return int(h.NM2)
            # Else, inhibitory so return -1 (error).
            else:
                return -1

class INTF6STDPConn(INTF6Conn,stdp.STDPConn):
    """
    STDP-using INTF6Conn for the model.

    Methods:
      __init__(self): constructor
      turnonrecwts(self,recinterval): turns on recording of weights every interval

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances            

    Usage:
    >>>nlc1 = intf6.INTF6STDPConn(prx,pox)
    """

    classlabel = 'INTF6STDPConn'
    numCreated = 0
    allConns = []
  
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=15.0, soft_thresh=False, \
        label='', proj2soma=False):
        # Set INTF6Conn parameters.
        # (Don't try to use super() method here because we have different 
        # __init()__ function signatures for the superclasses.)
        INTF6Conn.__init__(self, src, dest, weight, weight2, delay, label, proj2soma)

        # Set PlasticConn init stuff.
        nconn.PlasticConn.uniqueinit(self, weight)
        
        #
        # Set STDPConn init stuff, but according to INTF6 situation.
        #
        
        # Set the invariant STDPConn init stuff.
        stdp.STDPConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh)
            
        # Find the weight index.
        wtind = self.getwtncind()

        # Set the pointer using the appropriate weight index.
        h.setpointer(self.hobj._ref_weight[wtind],'wsyn',self.wtadjobj)
         
        #
        # Set INTF6STDPConn-specific stuff.
        #

        # Increment the instance count.
        INTF6STDPConn.numCreated += 1
        INTF6STDPConn.allConns.append(self)

    def turnonrecwts(self, recinterval):
        import neuron

        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[self.getwtncind()], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval

class INTF6STDPRLConn(INTF6Conn,stdprl.STDPRLConn):
    """
    STDP- and RL-using INTF6Conn for the model.

    Methods:
      __init__(self): constructor
      turnonrecwts(self,recinterval): turns on recording of weights every interval

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances            

    Usage:
    >>>nlc1 = intf6.INTF6STDPRLConn(prx,pox)
    """

    classlabel = 'INTF6STDPRLConn'
    numCreated = 0
    allConns = []
  
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=15.0, soft_thresh=False, \
        maxeligwind_Hebb=10.0, maxeligwind_antiHebb=10.0, \
        eliglength_Hebb=100.0, eliglength_antiHebb=100.0, \
        maxeliggain_Hebb=1.0, maxeliggain_antiHebb=-1.0, \
        expeligdecay=False, label='', proj2soma=False):
        # Set INTF6Conn parameters.
        # (Don't try to use super() method here because we have different 
        # __init()__ function signatures for the superclasses.)
        INTF6Conn.__init__(self, src, dest, weight, weight2, delay, label, proj2soma)

        # Set PlasticConn init stuff.
        nconn.PlasticConn.uniqueinit(self, weight)
 
        #
        # Set STDPRLConn init stuff, but according to INTF6 situation.
        #
        
        # Set the invariant STDPRLConn init stuff.
        stdprl.STDPRLConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh, maxeligwind_Hebb, \
            maxeligwind_antiHebb, eliglength_Hebb, eliglength_antiHebb, \
            maxeliggain_Hebb, maxeliggain_antiHebb, expeligdecay)
            
        # Find the weight index.
        wtind = self.getwtncind()

        # Set the pointer using the appropriate weight index.
        h.setpointer(self.hobj._ref_weight[wtind],'wsyn',self.wtadjobj)
        
        #
        # Set INTF6STDPRLConn-specific stuff.
        #

        # Increment the instance count.
        INTF6STDPRLConn.numCreated += 1
        INTF6STDPRLConn.allConns.append(self)

    def turnonrecwts(self, recinterval):
        import neuron

        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[self.getwtncind()], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval
