"""
intfire1.py -- module for wrapping IntFire1 neural units

Classes for IntFire1 neural units:
  IntFire1Unit (subclass of SpikeUnit)
  IntFire1Conn (subclass of NeuralConn)
    IntFire1STDPConn (also inherits STDPConn)
    IntFire1STDPRLConn (also inherits STDPRLConn)
        
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import neural unit module.
import neuralunit as nun

# Import the neural simulation module.
import neuralsim as nsim

# Import the timed callbacks module.
import timedcallback as tcb

# Load the neural connections module.
import neuralconn as nconn

# Load the stdp module.
import stdp

# Load the STDP RL module.
import stdp_rl as stdprl

# Import numpy.
import numpy

# Import mlab.
import matplotlib.mlab

class IntFire1Unit(nun.SpikeUnit):
    """
    IntFire1 unit for the model.  For non-plastic connections with a post-synaptic 
    IntFire1Unit type, you should use IntFire1Conn instead of NeuralConn, 
    and for plastic connections, IntFire1STDPConn or IntFire1STDPRLConn.

    Methods:
      __init__(self): constructor
      get_tau(self): get the "membrane voltage" decay time constant
      set_tau(self,value): set the "membrane voltage" decay time constant
      get_refrac(self): get the unit refractory period (in ms)
      set_refrac(self,value): set the unit refractory period (in ms)  
      get_verbose(self): get the unit verbose setting
      set_verbose(self,value): set the unit verbose setting           
      turnonrecms(recinterval): turn on recording of membrane potentials every 
        interval
      getrecms(self): return tvec and amplitudes for the membrane potentials
      getmval(self): get the current M value from the unit (call to M()) 
      recmval(self): calls getmval() and appends result to list of recorded m 
        values
      disable(self): disables the unit so no output comes out, nor input in
      enable(self): enables or reenables the unit's output and input      
      isdisabled(self): returns whether the unit is disabled
                  
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances     
      hobj: hoc IntFire1 object
      recms: hoc Vector object for recorded membrane voltages (m)
      recmsinterval: recording interval for m values in ms

    Usage:
    >>>i1 = if1.IntFire1Unit()
    """

    classlabel = 'IntFire1Unit'
    numCreated = 0
    allUnits = []
    
    def __init__(self, NTs=['GLU'], tau=10.0, refrac=5.0, verbose=0, label=''):
        # Set superclass parameters.
        super(IntFire1Unit,self).__init__(NTs, label)

        # Create the hoc IntFire1 cell object.
        self.hobj = h.IntFire1b()
        self.hobj.tau = tau
        self.hobj.refrac = refrac
        self.hobj.verbose = verbose

        # Set up a hoc nil object at first for m recording.
        self.recms = None

        # Increment the instance count.
        IntFire1Unit.numCreated += 1
        IntFire1Unit.allUnits.append(self)
        
    def prenrnrun(self):
        # Clear out recms if it is not None.
        if not self.recms == None:
            self.recms.resize(0)
        
    def get_tau(self):
        return self.hobj.tau
        
    def set_tau(self, value):
        self.hobj.tau = value        
        
    def get_refrac(self):
        return self.hobj.refrac
        
    def set_refrac(self, value):
        self.hobj.refrac = value     

    def get_verbose(self):
        return self.hobj.verbose
        
    def set_verbose(self, value):
        self.hobj.verbose = value     
                       
    def turnonrecms(self, thesim, recinterval):
        # Set up the hoc Vector for recording m values.
        self.recms = h.Vector()
        
        # Remember the recording interval.
        self.recmsinterval = recinterval
        
        # Create a repeating callback that starts at the beginning of the sim
        # and calls recmval every recording interval.
        self.recmscb = tcb.TimedCallback(self.recmval, thesim, 
            0.0, callrepeatperiod=recinterval, label='recm')

    def getrecms(self):
        recms = numpy.array(self.recms)
        tvec = numpy.linspace(0.0, self.recmsinterval * len(recms), \
             num=len(recms), endpoint=False)
        return tvec, recms
        
    def getmval(self):
        return self.hobj.M()
        
    def recmval(self):
        self.recms.append(self.getmval())
        
    def disable(self):
        if not self.isdisabled():       
            self.hobj.disabled = 1
        
    def enable(self):
        if self.isdisabled(): 
            self.hobj.disabled = 0
            
    def isdisabled(self):
        return (self.hobj.disabled == 1)
                 
class IntFire1Conn(nconn.NeuralConn):
    """
    Neural connection converging into IntFire1Unit for the model.

    Methods:
      __init__(self): constructor
      getwt(self): returns the primary weight
      setwt(self,weight): sets the primary weight
      
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances      
      
    Usage:
    >>>nlc1 = if1.IntFire1Conn(prx,pox)
    """

    classlabel = 'IntFire1Conn'
    numCreated = 0
    allConns = []
    
    def __init__(self, src, dest, weight=0.0, delay=0.0, label=''):
        # Set superclass parameters.
        nconn.NeuralConn.__init__(self, src, dest, weight, delay, label)
        
        # Negate the weight if the src is inhibitory.
        if src.isinhibitory():
            self.hobj.weight[0] = -weight

        # Increment the instance count.
        IntFire1Conn.numCreated += 1
        IntFire1Conn.allConns.append(self)
        
    def getwt(self):
        # Return the absolute value since negative weights in the NetCons are
        # an implementational detail.
        return abs(self.hobj.weight[0])

    def setwt(self, weight):
        # Negate the weight if the src is inhibitory, set regular otherwise.
        if self.src.isinhibitory():
            self.hobj.weight[0] = -weight
        else:
            self.hobj.weight[0] = weight

class IntFire1STDPConn(IntFire1Conn, stdp.STDPConn):
    """
    STDP-using IntFire1Conn for the model.

    Methods:
      __init__(self): constructor
      turnonrecwts(self,recinterval): turns on recording of weights every interval
      getrecwts(self): get recorded weights, returning absolute values for the
        NetCon weight values
        
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances            

    Usage:
    >>>nlc1 = if1.IntFire1STDPConn(prx,pox)
    """

    classlabel = 'IntFire1STDPConn'
    numCreated = 0
    allConns = []
  
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=1.0, soft_thresh=False, \
        label=''):
        # Set IntFire1Conn parameters.
        # (Don't try to use super() method here because we have different 
        # __init()__ function signatures for the superclasses.)
        IntFire1Conn.__init__(self, src, dest, weight, delay, label)

        # Set PlasticConn init stuff.
        nconn.PlasticConn.uniqueinit(self, weight)
        
        #
        # Set STDPConn init stuff, but according to IntFire1 situation.
        #
        
        # Set the invariant STDPConn init stuff.
        stdp.STDPConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh)
        
        # Set the wsyn pointer for the weight adjuster object using the 
        # appropriate weight index.
        h.setpointer(self.hobj._ref_weight[0],'wsyn',self.wtadjobj)

        # Negate the weight if the src is inhibitory.
        if src.isinhibitory():
            self.wtadjobj.negweight = 1
                    
        #
        # Set IntFire1STDPConn-specific stuff.
        #

        # Increment the instance count.
        IntFire1STDPConn.numCreated += 1
        IntFire1STDPConn.allConns.append(self)

    def turnonrecwts(self, recinterval):
        import neuron

        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[0], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval
        
    def getrecwts(self):
        recwts = numpy.array(self.recwts)
        tvec = numpy.linspace(0.0, self.recwtsinterval * len(recwts), \
             num=len(recwts), endpoint=False)
             
        # Return the absolute value because we don't want the user to be
        # concerned with negative weights.  (They are implicit in the fact that
        # the pre-synaptic NeuralUnit is inhibitory.)
        return tvec, numpy.abs(recwts)
        
class IntFire1STDPRLConn(IntFire1Conn, stdprl.STDPRLConn):
    """
    STDP- and RL-using IntFire1Conn for the model.

    Methods:
      __init__(self): constructor
      turnonrecwts(self,recinterval): turns on recording of weights every interval
      getrecwts(self): get recorded weights, returning absolute values for the
        NetCon weight values
        
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances            

    Usage:
    >>>nlc1 = if1.IntFire1STDPRLConn(prx,pox)
    """

    classlabel = 'IntFire1STDPRLConn'
    numCreated = 0
    allConns = []
  
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=1.0, soft_thresh=False, \
        maxeligwind_Hebb=10.0, maxeligwind_antiHebb=10.0, \
        eliglength_Hebb=100.0, eliglength_antiHebb=100.0, \
        maxeliggain_Hebb=1.0, maxeliggain_antiHebb=-1.0, \
        expeligdecay=False, label=''):
        # Set IntFire1Conn parameters.
        # (Don't try to use super() method here because we have different 
        # __init()__ function signatures for the superclasses.)
        IntFire1Conn.__init__(self, src, dest, weight, delay, label)

        # Set PlasticConn init stuff.
        nconn.PlasticConn.uniqueinit(self, weight)
        
        #
        # Set STDPRLConn init stuff, but according to IntFire1 situation.
        #
        
        # Set the invariant STDPRLConn init stuff.
        stdprl.STDPRLConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh, maxeligwind_Hebb, \
            maxeligwind_antiHebb, eliglength_Hebb, eliglength_antiHebb, \
            maxeliggain_Hebb, maxeliggain_antiHebb, expeligdecay)
        
        # Set the wsyn pointer for the weight adjuster object using the 
        # appropriate weight index.
        h.setpointer(self.hobj._ref_weight[0],'wsyn',self.wtadjobj)

        # Negate the weight if the src is inhibitory.
        if src.isinhibitory():
            self.wtadjobj.negweight = 1
        
        #
        # Set IntFire1STDPRLConn-specific stuff.
        #

        # Increment the instance count.
        IntFire1STDPRLConn.numCreated += 1
        IntFire1STDPRLConn.allConns.append(self)

    def turnonrecwts(self, recinterval):
        import neuron

        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[0], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval
        
    def getrecwts(self):
        recwts = numpy.array(self.recwts)
        tvec = numpy.linspace(0.0, self.recwtsinterval * len(recwts), \
             num=len(recwts), endpoint=False)
             
        # Return the absolute value because we don't want the user to be
        # concerned with negative weights.  (They are implicit in the fact that
        # the pre-synaptic NeuralUnit is inhibitory.)
        return tvec, numpy.abs(recwts)
