"""
izhiunit.py -- module for wrapping Izhikevich neural units, as implemented 
  by Cliff Kerr in izhi.mod

Classes for Izhikevich neural units:
  IzhiUnit (subclass of SpikeUnit)
  IzhiConn (subclass of NeuralConn)
    IzhiSTDPConn (also inherits STDPConn)
    IzhiSTDPRLConn (also inherits STDPRLConn)
        
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import neural unit module.
import neuralunit as nun

# Import the neural simulation module.
import neuralsim as nsim

# Import the timed callbacks module.
import timedcallback as tcb

# Load the neural connections module.
import neuralconn as nconn

# Load the stdp module.
import stdp

# Load the STDP RL module.
import stdp_rl as stdprl

# Import numpy.
import numpy

# Import mlab.
import matplotlib.mlab

class IzhiUnit(nun.SpikeUnit):
    """
    Izhikevich unit for the model.  For non-plastic connections with a post-synaptic 
    IzhiUnit type, you should use IzhiConn instead of NeuralConn, 
    and for plastic connections, IzhiSTDPConn or IzhiSTDPRLConn.
    
    The Izhikevich model is taken from 
      Izhikevich EM, Edelman GM (2008).
      "Large-scale model of mammalian thalamocortical systems." 
      PNAS 105(9) 3593-3598.
       
    The dynamics formulae maintain two variables V (membrane potential in mV)
    and a recovery variable u (net voltage-gated outward current):
    
      (1) C * v-dot = k * (v - vr) * (v - vt) - u + I
      (2) u-dot = a * (b * (v - vr) - u)
    
    where vr is the cells resting membrane potential, vt is the threshold 
    potential, I is the (inward flowing positive) current, and k, a, and b 
    are gain constants for the equations.  Additionally, when v drifts over 
    a voltage vpeak, v gets instantaneously set to parameter c and u gets 
    set to u + d.
    
    Some commentary on the equations to try to illuminate intuitions... 
     
    (1) is similar to standard integrate-and-fire equations in that injected 
    current increases the voltage with capacitance setting the time-constant 
    for the change.  However the first 2 terms modify this and are worth 
    further exploration.  The first term will be positive when v is either 
    less than or greater than both vt and vr and will be negative when it 
    falls between the two thresholds.  That means that the influence of this
    term will drive v towards vr, until v > vt.  When the latter happens, 
    this term will drive runaway v increase until v reaches vpeak (the 
    allowed height of the spike).  The second term, however, subtracts u 
    from the voltage increase.  u represents the outward (positive) current 
    drive the cell back towards resting potential.
    
    (2) is the dynamical equation for u.  a sets the overall gain on the 
    change of u.  Two terms compete for effect in this equation.  The first
      b * (v - vr)
    will cause will drive more u increase the more depolarized the cell is 
    with respect to resting potential.  The second term (-u) will encourage
    decay of u the greater its value.  Gain b, then, sets the degree to 
    which the voltage excitation of the recovery variable u prevails.
    
    When v > vpeak happens, it can be considered analogous to sudden activation
    of outward (positive) current channels causing the cell to polarize 
    again.  v gets set to c, making c the reset voltage.  (Frequently c is 
    set equal to vr.) u has d added to it which corresponds to a sudden outflux 
    of current due to opening of polarizing voltage-gated channels.  It 
    is worth noting that the immediate effect of these two things on (2) is 
    that u is driven very high, but will decay very rapidly because v = vr.  
    In fact, u is likely to drive v under vr which will lead to quicker decay 
    of u.
    
    Methods:
      __init__(self,...): constructor     
      get_C(self): get the membrane capacitance of the unit
      set_C(self,value): set the membrane capacitance of the unit
      get_k(self): get the gain of voltage-gated drive of v
      set_k(self,value): set the gain of voltage-gated drive of v
      get_vr(self): get the membrane resting potential (in mV)
      set_vr(self,value): set the membrane resting potential (in mV)
      get_vt(self): get the membrane threshold potential (in mV)
      set_vt(self,value): set the membrane threshold potential (in mV)     
      get_vpeak(self): get peak voltage of v (in mV)
      set_vpeak(self,value): set the peak voltage of v (in mV)
      get_a(self): get the overall gain of u change
      set_a(self,value): set the overall gain of u change
      get_b(self): get the gain of the voltage-base u drive term
      set_b(self,value): set the gain of the voltage-base u drive term
      get_c(self): get the spike voltage reset voltage (in mV)
      set_c(self,value): set the spike voltage reset voltage (in mV)
      get_d(self): get the spike recovery current outflux
      set_d(self,value): set the spike recovery current outflux     
      get_tauAMPA(self): get the AMPA synaptic decay time constant (in ms)
      set_tauAMPA(self,value): set the AMPA synaptic decay time constant (in ms)
      get_tauNMDA(self): get the NMDA synaptic decay time constant (in ms)
      set_tauNMDA(self,value): set the NMDA synaptic decay time constant (in ms)
      get_tauGABAA(self): get the GABAA synaptic decay time constant (in ms)
      set_tauGABAA(self,value): set the GABAA synaptic decay time constant (in ms)
      get_tauGABAB(self): get the GABAB synaptic decay time constant (in ms)
      set_tauGABAB(self,value): set the GABAB synaptic decay time constant (in ms)      
      get_xloc(self): get the unit's x location
      set_xloc(self,value): set the unit's x location
      get_yloc(self): get the unit's y location
      set_yloc(self,value): set the unit's y location
      get_zloc(self): get the unit's z location
      set_zloc(self,value): set the unit's z location
      get_flag(self): get the flag value
      set_flag(self,value): set the flag value                 
      # get_verbose(self): get the unit verbose setting
      # set_verbose(self,value): set the unit verbose setting           
      turnonrecVs(recinterval): turn on recording of membrane potentials every 
        interval
      getrecVs(self): return tvec and amplitudes for the membrane potentials
      getVval(self): get the current V value from the unit
      recVval(self): calls getVval() and appends result to list of recorded V 
        values
      # disable(self): disables the unit so no output comes out, nor input in
      # enable(self): enables or reenables the unit's output and input      
      # isdisabled(self): returns whether the unit is disabled
                  
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances   
      dummysection (class): hoc object which is a dummy Section to be used by all the 
        Izhikevich units (needed because it is a POINT_PROCESS)        
      hobj: hoc IntFire1 object
      recVs: hoc Vector object for recorded membrane voltages (V)
      recVsinterval: recording interval for V values in ms

    Usage:
    - izhi.mod needs to be installed in special.

    >>>i1 = izhi.IzhiUnit()
    """

    classlabel = 'IzhiUnit'
    numCreated = 0
    allUnits = []
    dummysection = h.Section()
        
    # NOTE: The default values are for pyramidal cells.
    def __init__(self, NTs=['GLU'], C=100, k=3, vr=-60, vt=-50, vpeak=30, \
        a=0.01, b=5, c=-60, d=400, tauAMPA=5, tauNMDA=150, tauGABAA=6, \
        tauGABAB=150, xloc=0, yloc=0, zloc=0, flag=-1, \
        verbose=0, label=''):
        # Set superclass parameters.
        super(IzhiUnit,self).__init__(NTs, label)

        # Create the hoc Izhikevich cell object.
        self.hobj = h.Izhi(0.5, sec=IzhiUnit.dummysection)
        self.hobj.C = C
        self.hobj.k = k
        self.hobj.vr = vr  
        self.hobj.vt = vt  
        self.hobj.vpeak = vpeak  
        self.hobj.a = a  
        self.hobj.b = b  
        self.hobj.c = c  
        self.hobj.d = d  
        self.hobj.tauAMPA = tauAMPA  
        self.hobj.tauNMDA = tauNMDA  
        self.hobj.tauGABAA = tauGABAA  
        self.hobj.tauGABAB = tauGABAB          
        self.hobj.xloc = xloc  
        self.hobj.yloc = yloc  
        self.hobj.zloc = zloc  
        self.hobj.flag = flag                       
        #self.hobj.verbose = verbose

        # Set up a hoc nil object at first for V recording.
        # NOTE: we probably want something for the membrane restorative 
        # variable U, too.
        self.recVs = None

        # Increment the instance count.
        IzhiUnit.numCreated += 1
        IzhiUnit.allUnits.append(self)
        
    def prenrnrun(self):
        # Clear out recVs if it is not None.
        if not self.recVs == None:
            self.recVs.resize(0)
        
    def get_C(self):
        return self.hobj.C
        
    def set_C(self, value):
        self.hobj.C = value        
        
    def get_k(self):
        return self.hobj.k
        
    def set_k(self, value):
        self.hobj.k = value     

    def get_vr(self):
        return self.hobj.vr
        
    def set_vr(self, value):
        self.hobj.vr = value     

    def get_vt(self):
        return self.hobj.vt
        
    def set_vt(self, value):
        self.hobj.vt = value     

    def get_vpeak(self):
        return self.hobj.vpeak
        
    def set_vpeak(self, value):
        self.hobj.vpeak = value     

    def get_a(self):
        return self.hobj.a
        
    def set_a(self, value):
        self.hobj.a = value     

    def get_b(self):
        return self.hobj.b
        
    def set_b(self, value):
        self.hobj.b = value     

    def get_c(self):
        return self.hobj.c
        
    def set_c(self, value):
        self.hobj.c = value     

    def get_d(self):
        return self.hobj.d
        
    def set_d(self, value):
        self.hobj.d = value     

    def get_tauAMPA(self):
        return self.hobj.tauAMPA
        
    def set_tauAMPA(self, value):
        self.hobj.tauAMPA = value     

    def get_tauNMDA(self):
        return self.hobj.tauNMDA
        
    def set_tauNMDA(self, value):
        self.hobj.tauNMDA = value     

    def get_tauGABAA(self):
        return self.hobj.tauGABAA
        
    def set_tauGABAA(self, value):
        self.hobj.tauGABAA = value     

    def get_tauGABAB(self):
        return self.hobj.tauGABAB
        
    def set_tauGABAB(self, value):
        self.hobj.tauGABAB = value     

    def get_xloc(self):
        return self.hobj.xloc
        
    def set_xloc(self, value):
        self.hobj.xloc = value     

    def get_yloc(self):
        return self.hobj.yloc
        
    def set_yloc(self, value):
        self.hobj.yloc = value     

    def get_zloc(self):
        return self.hobj.zloc
        
    def set_zloc(self, value):
        self.hobj.zloc = value     

    def get_flag(self):
        return self.hobj.flag
        
    def set_flag(self, value):
        self.hobj.flag = value     

    # Want to add verbose flag to izhi.mod
    #def get_verbose(self):
        #return self.hobj.verbose
        
    #def set_verbose(self, value):
        #self.hobj.verbose = value     
                       
    def turnonrecVs(self, thesim, recinterval):
        # Set up the hoc Vector for recording V values.
        self.recVs = h.Vector()
        
        # Remember the recording interval.
        self.recVsinterval = recinterval
        
		# This method may not quite work correctly because it may not show Vm
        # dynamics between unit state updates (BUT NEED TO TEST).
        self.recVs.record(self.hobj._ref_V, self.recVsinterval)
        
        # THE CALL BELOW MAY BE NEEDED INSTEAD OF THE ONE ABOVE.
        # Create a repeating callback that starts at the beginning of the sim
        # and calls recVval every recording interval.
        # IF THIS CALL IS DEFINITELY NOT NEEDED, WE SHOULD PROBABLY 
        # GET RID OF THE thesim PARAMETER IN THIS FUNCTION.
        #self.recVscb = tcb.TimedCallback(self.recVval, thesim, 
            #0.0, callrepeatperiod=recinterval, label='recV')

    def getrecVs(self):
        recVs = numpy.array(self.recVs)
        tvec = numpy.linspace(0.0, self.recVsinterval * len(recVs), \
             num=len(recVs), endpoint=False)
        return tvec, recVs
        
    def getVval(self):
        return self.hobj.V
        
    def recVval(self):
        self.recVs.append(self.getVval())
        
    # Want to add disabled flag to izhi.mod      
    #def disable(self):
        #if not self.isdisabled():       
            #self.hobj.disabled = 1
        
    #def enable(self):
        #if self.isdisabled(): 
            #self.hobj.disabled = 0
            
    #def isdisabled(self):
        #return (self.hobj.disabled == 1)
                 
class IzhiConn(nconn.NeuralConn):
    """
    Neural connection converging into IzhiUnit for the model.

    Methods:
      __init__(self): constructor
      getwt(self): returns the primary weight
      setwt(self,weight): sets the primary weight
      getwt2(self): returns the secondary weight
      setwt2(self,weight): sets the secondary weight
      getwtncind(self): returns the proper primary weight index for the NetCon
      getwt2ncind(self): returns the proper secondary weight index for the NetCon
      
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances      
      
    Usage:
    >>>nlc1 = izhi.IzhiConn(prx,pox)
    """

    classlabel = 'IzhiConn'
    numCreated = 0
    allConns = []
    
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, label=''):
        # Set superclass parameters.
        nconn.NeuralConn.__init__(self, src, dest, weight, delay, label)
        
        # Clear the first weight (set by the NeuralConn constructor).
        self.hobj.weight[0] = 0.0
        
        # Set the weights according to the specific rules.
        self.setwt(weight)
        self.setwt2(weight2)

        # Increment the instance count.
        IzhiConn.numCreated += 1
        IzhiConn.allConns.append(self)
        
    def getwt(self):
        return self.hobj.weight[self.getwtncind()]

    def setwt(self,weight):
        self.hobj.weight[self.getwtncind()] = weight

    def getwt2(self):
        wtind = self.getwt2ncind()
        if wtind != -1:
            return self.hobj.weight[wtind]
        else:
            return -1  # error condition (no valid weight secondary weight)

    def setwt2(self,weight):
        wtind = self.getwt2ncind()
        if wtind != -1:
            self.hobj.weight[wtind] = weight

    def getwtncind(self):
		# If cell is excitatory, return AMPA index.
		if self.src.isexcitatory():
			return 0 # AMPA index
		# Else, inhibitory so return GABA_A index.
		else:
			return 2 # GABA_A index

    def getwt2ncind(self):
		# If cell is excitatory, return NMDA index.
		if self.src.isexcitatory():
			return 1 # NMDA index
		# Else, inhibitory so return GABA_B index.
		else:
			return 3 # GABA_B index

class IzhiSTDPConn(IzhiConn, stdp.STDPConn):
    """
    STDP-using IzhiConn for the model.

    Methods:
      __init__(self): constructor
      turnonrecwts(self,recinterval): turns on recording of weights every interval
      getrecwts(self): get recorded weights, returning absolute values for the
        NetCon weight values
        
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances            

    Usage:
    >>>nlc1 = izhi.IzhiSTDPConn(prx,pox)
    """

    classlabel = 'IzhiSTDPConn'
    numCreated = 0
    allConns = []
  
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=1.0, soft_thresh=False, \
        label=''):
        # Set IzhiConn parameters.
        # (Don't try to use super() method here because we have different 
        # __init()__ function signatures for the superclasses.)
        IzhiConn.__init__(self, src, dest, weight, weight2, delay, label)

        # Set PlasticConn init stuff.
        nconn.PlasticConn.uniqueinit(self, weight)
        
        #
        # Set STDPConn init stuff, but according to Izhi situation.
        #
        
        # Set the invariant STDPConn init stuff.
        stdp.STDPConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh)
        
        # Find the weight index.
        wtind = self.getwtncind()

        # Set the pointer using the appropriate weight index.
        h.setpointer(self.hobj._ref_weight[wtind],'wsyn',self.wtadjobj)
                    
        #
        # Set IzhiSTDPConn-specific stuff.
        #

        # Increment the instance count.
        IzhiSTDPConn.numCreated += 1
        IzhiSTDPConn.allConns.append(self)

    def turnonrecwts(self, recinterval):
        import neuron

        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[self.getwtncind()], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval
                
class IzhiSTDPRLConn(IzhiConn, stdprl.STDPRLConn):
    """
    STDP- and RL-using IzhiConn for the model.

    Methods:
      __init__(self): constructor
      turnonrecwts(self,recinterval): turns on recording of weights every interval
      getrecwts(self): get recorded weights, returning absolute values for the
        NetCon weight values
        
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances            

    Usage:
    >>>nlc1 = izhi.IzhiSTDPRLConn(prx,pox)
    """

    classlabel = 'IzhiSTDPRLConn'
    numCreated = 0
    allConns = []
  
    def __init__(self, src, dest, weight=0.0, weight2=0.0, delay=0.0, \
        tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=1.0, soft_thresh=False, \
        maxeligwind_Hebb=10.0, maxeligwind_antiHebb=10.0, \
        eliglength_Hebb=100.0, eliglength_antiHebb=100.0, \
        maxeliggain_Hebb=1.0, maxeliggain_antiHebb=-1.0, \
        expeligdecay=False, label=''):
        # Set IzhiConn parameters.
        # (Don't try to use super() method here because we have different 
        # __init()__ function signatures for the superclasses.)
        IzhiConn.__init__(self, src, dest, weight, weight2, delay, label)

        # Set PlasticConn init stuff.
        nconn.PlasticConn.uniqueinit(self, weight)
        
        #
        # Set STDPRLConn init stuff, but according to Izhi situation.
        #
        
        # Set the invariant STDPRLConn init stuff.
        stdprl.STDPRLConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh, maxeligwind_Hebb, \
            maxeligwind_antiHebb, eliglength_Hebb, eliglength_antiHebb, \
            maxeliggain_Hebb, maxeliggain_antiHebb, expeligdecay)
        
        # Find the weight index.
        wtind = self.getwtncind()

        # Set the pointer using the appropriate weight index.
        h.setpointer(self.hobj._ref_weight[wtind],'wsyn',self.wtadjobj)
        
        #
        # Set IzhiSTDPRLConn-specific stuff.
        #

        # Increment the instance count.
        IzhiSTDPRLConn.numCreated += 1
        IzhiSTDPRLConn.allConns.append(self)

    def turnonrecwts(self, recinterval):
        import neuron

        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[self.getwtncind()], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval
        
