"""
neuralconn.py -- module for classes for model neural connections

Classes for neural connections:
  NeuralConn (base class)
      PlasticConn
  ConnGroup (base class)
 
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import numpy.
import numpy

class NeuralConn(object):
    """
    Generic neural connection for the model.  This can be used to connect any 
    two NeuralUnits, but generally you want to use the class XConn where X is 
    the class of the post-synaptic NeuralUnit.

    Methods:
      __init__(self): constructor
      prenrnrun(self): stub for code that needs to be called before h.run()
      postnrnrun(self): stub for code that needs to be called after h.run()      
      getwt(self): returns the weight (assume to be weight[0])
      setwt(self,weight): sets the weight
      turnonrecspks(self): turns on spikes recording
      isrecording(self): returns whether we are recording spikes      
      getrecspks(self): returns a numpy array of the recorded spike times
      disable(self): disables the connection 
      enable(self): enables or reenables the connection     
      isdisabled(self): returns whether the connection is disabled  
         
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances
      hobj: hoc NetCon object
      src: pointer to source NeuralUnit
      dest: pointer to destination NeuralUnit
      recspks: hoc Vector object for recorded spikes
      instlabel: instance label (usable for tagging)

    Usage:
    >>>nlc1 = neuralconn.NeuralConn(nst,i1)
    """

    classlabel = 'NeuralConn'
    numCreated = 0
    allConns = []

    def __init__(self, src, dest, weight=0.0, delay=0.0, label=''):
        # Create the NetCon hoc object.
        if src.outconnsect == None:
            self.hobj = h.NetCon(src.getoutconnattachpoint(), \
                dest.getinconnattachpoint(src))
        else:
            self.hobj = h.NetCon(src.getoutconnattachpoint(), \
                dest.getinconnattachpoint(src), sec=src.outconnsect)           
        self.hobj.delay = delay
        
        # Set the threshold as appropriate.
        if src.outconnthresh != None:
            self.hobj.threshold = src.outconnthresh

        # Set the first weight.
        self.hobj.weight[0] = weight

        # Set the source and destination pointers to the NeuralUnits.
        self.src = src
        self.dest = dest

        # Set the instance label.
        self.instlabel = label

        # Increment the instance count.
        NeuralConn.numCreated += 1
        NeuralConn.allConns.append(self)
        
    def prenrnrun(self):
        # Should be instantiated with code for the particular NeuralConn 
        # subclass. 
        pass
        
    def postnrnrun(self):
        # Should be instantiated with code for the particular NeuralConn 
        #subclass.        
        pass 
      
    def getwt(self):
        return self.hobj.weight[0]

    def setwt(self,weight):
        self.hobj.weight[0] = weight

    def turnonrecspks(self):
        # Turn on recording for the pre-synaptic unit (if not already turned
        # on).
        self.src.turnonrecspks()
              
    def isrecording(self):
        # We are recording if and only if our pre-synaptic cell is recording.
        return self.src.isrecording()
                
    def getrecspks(self):
        # Give an error and exit if recording is not enabled.
        if not self.isrecording():
            print 'ERROR in %s NeuralConn: Recording not enabled in src unit.' %\
                self.instlabel   
            return numpy.array([])
                    
        # Return the source unit's recorded spikes.  This is done this way
        # because attempts to record from the NetCon directly will conflict 
        # with any attempts to record from other NetCons with the same 
        # pre-synaptic unit.
        return self.src.getrecspks()
        
    def disable(self):
        self.hobj.active(0)
        
    def enable(self):
        self.hobj.active(1)    
        
    def isdisabled(self):
        return (self.hobj.active() == 0.0)   
                 
class PlasticConn(NeuralConn):
    """
    Generic learning neural connection for the model.  If a specialized 
    XNeuralConn is need for a post-synaptic NeuralUnit of type X, usually 
    you want to use a subclass which inherits PlasticConn functionality, 
    e.g., INTF6STDPConn for INTF6Units.

    Methods:
      __init__(self,...): constructor
      uniqueinit(self,weight): constructor code unique to our class.  This 
        function is needed to deal with multiple inheritance
      resetweight(self): resets the unit's weight to the initial value
      turnonlearning(self): turns on learning
      turnofflearning(self): turns off learning
      turnonrecwts(self,recinterval): turns on recording of weights every interval
      getrecwts(self): returns tvec and amplitudes for the weights

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances      
      initialwt: initial value of the weight (saved to allow possibility of
        rerunning sim with different parameters, but start weights)
      learningon: is learning turned on?
      recwts: hoc Vector object for recorded weights
      recwtsinterval: recording interval for weights in ms

    Usage:
    >>>nlc1 = neuralconn.PlasticConn(nst,i1)
    """

    classlabel = 'PlasticConn'
    numCreated = 0
    allConns = []
    
    def __init__(self, src, dest, weight=0.0, delay=0.0, label=''):
        # Set superclass parameters.
        super(PlasticConn,self).__init__(src, dest, weight, delay, label)

        # Do init specific to PlasticConn.  We use PlasticConn. prefix because 
        # object not constructed yet.
        PlasticConn.uniqueinit(self, weight)
        
    def uniqueinit(self, weight=0.0):
        # Set learning to be on by default.
        self.learningon = True

        # Save the initial weight so we have something to reset to if we need
        # to restart the sim from the same initial state, but some parameters 
        # different.
        self.initialwt = weight
        
        # Set up a hoc Vector for weights recording.
        self.recwts = h.Vector()

        # Increment the instance count.
        PlasticConn.numCreated += 1
        PlasticConn.allConns.append(self)      
        
    def resetweight(self):
        self.setwt(self.initialwt)
        
    def turnonlearning(self):
        self.learningon = True
 
    def turnofflearning(self):
        self.learningon = False   

    def turnonrecwts(self, recinterval):
        # Turn on the recording for a particular recording interval.
        self.recwts.record(self.hobj._ref_weight[0], recinterval)

        # Remember the recording interval.
        self.recwtsinterval = recinterval

    def getrecwts(self):
        recwts = numpy.array(self.recwts)
        tvec = numpy.linspace(0.0, self.recwtsinterval * len(recwts), \
             num=len(recwts), endpoint=False)
        return tvec, recwts

class ConnGroup(object):
    """
    Generic group of neural connections for the model.  This can be a mixed group 
    consisting of several-typed units.

    Methods:
      __init__(self): constructor
      __iter__(self): returns the iterator for the encapsulated list
      __len__(self): gets the size of the group
      __getitem__(self, index): allows array indexing of the group
      getconnbyinstlabel(self, label): gets the first connection matching the 
        instance label
      getconnsbyclasslabel(self, label): makes a new ConnGroup containing 
        connections in the old group that match by classlabel      
      append(self,theconn): appends a connection object to the group
      disable(self): disables all of the connections
      enable(self): enables or reenables all of the connections      
      turnonlearning(self): turns on learning for all of the connections
      turnofflearning(self): turns off learning for all of the connections       
      resetwts(self): resets all of the weights to the initial values  
      getwts(self): returns a list of the current (primary) weights
      getwts2(self): returns a list of the current secondary weights (if these
        exist)
      turnonrecspks(self): turns on the recording of spikes for all of the 
        connections
      isrecording(self): returns whether we are recording spikes from the 
        connections
      getrecspks(self): returns a numpy array of the recorded spike times
                
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allGroups (class): list of all instances           
      conns: connections the group contains (references only)
      instlabel: instance label (usable for tagging)

    Usage:
    >>>nlcg1 = neuralconn.ConnGroup()
    """

    classlabel = 'ConnGroup'
    numCreated = 0
    allGroups = []
    
    def __init__(self, theconns=[], label=''):
        # Create an empty group as a Python list.
        self.conns = []
        
        # Add the units that were passed in.
        self.conns += theconns

        # Set the instance label.
        self.instlabel = label

        # Increment the instance count.
        ConnGroup.numCreated += 1
        ConnGroup.allGroups.append(self)
         
    def __iter__(self):
        return iter(self.conns)                 
        
    def __len__(self):
        return len(self.conns)
        
    def __getitem__(self, index):
        return self.conns[index]
        
    def getconnbyinstlabel(self, label):
        for conn in self.conns:
            if conn.instlabel == label:
                return conn
                
    def getconnsbyclasslabel(self, label):
        # Create a new ConnGroup to hold the matches.
        cgrp = ConnGroup()
        
        # Put matches by class in the new ConnGroup.
        for conn in self.conns:      
            if conn.classlabel == label:
                cgrp.append(conn)
                
        # Return the group. 
        return cgrp
                
    def append(self,theconn):
        self.conns.append(theconn)
        
    def disable(self):
        for conn in self.conns:
            conn.disable()
            
    def enable(self):
        for conn in self.conns:
            conn.enable()
            
    def turnonlearning(self):
        for conn in self.conns:
            conn.turnonlearning()
 
    def turnofflearning(self):
        for conn in self.conns:
            conn.turnofflearning()    
                  
    def resetwts(self):
        for conn in self.conns:
            conn.resetweight()
            
    def getwts(self):
        return numpy.array([conn.getwt() for conn in self])
        
    def getwts2(self):
        return numpy.array([conn.getwt2() for conn in self])      
        
    def turnonrecspks(self):
        if not self.isrecording():      
            for conn in self.conns:
                conn.turnonrecspks()
        
    def turnoffrecspks(self):
        if self.isrecording:
            for conn in self.conns:
                conn.turnoffrecspks()
            
    def isrecording(self):
        # Return False if we see any connection that is not recording.
        for conn in self.conns:
            if not conn.isrecording():
                return False
                
        # Return True.  We made it through the whole list.
        return True 
        
    def getrecspks(self):
        # Give an error and exit if recording is not enabled.
        if not self.isrecording():
            print 'ERROR in %s ConnGroup: Recording not enabled.' %\
                self.instlabel      
            return numpy.array([])
              
        # Set up a raster list to hold numpy arrays for each of the connection's 
        # spikes.
        rasterlist = []
        for conn in self.conns:
            rasterlist.append(numpy.array(conn.getrecspks()))
            
        # Return a tuple for the raster list.
        return tuple(rasterlist)
