"""
neuralnetwork.py -- module for classes for an entire network of neural units

Classes for groups of neural units:
  NeuralNetwork (base class)
  
Initial design and implementation: George Chadderdon, 2013
 
Last update: 7/22/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# NOTE: This should be ultimately set up as a "singleton" so you can't 
# instantiate multiple objects.
class NeuralNetwork(object):
    """
    Class for an entire NEURON neural network model.  Only a single 
    instance of this should be set up at any given time or weird interferences
    may happen between the networks when (supposedly) one sim is run.
    You should rerun a new Python session when you want to initialize a 
    fresh NeuralNetwork object.  (The code will let you bind new 
    NeuralNetwork objects, or unbind old ones, but the results may be 
    unpredictable.)
    
    Methods:
      __init__(self): constructor, should initialize the static structure of the 
        network
      setdefaultparams(self): set default values for the network parameters
      prenrnrun(self): stub for code that needs to be called before h.run()
      postnrnrun(self): stub for code that needs to be called after h.run()
    
    Attributes:
      classlabel (class): text label for the class
      instlabel: instance label (usable for tagging)
      netparams: dictionary of parameters for the model
        'randseed': random number seed for network wiring
      hasrun: has the simulation run yet?
      randobj: hoc Random object for randomizing wiring
          
    Usage:
    >>>thenet = neuralnetwork.NeuralNetwork(label='M1MultilayerModel')
    """

    classlabel = 'NeuralNetwork'

    def __init__(self, params={}, label='theNetwork'):      
        # Create an empty dictionary for the network parameters.
        self.netparams = {}
        
        # Set default versions of the network parameters.
        self.setdefaultparams()
        
        # For each of our parameters, if the key is in the passed in parameters,
        # overwrite our parameter using the latter.
        for thekey in self.netparams.keys():
            if thekey in params:
                self.netparams[thekey] = params[thekey]
        
        # Set the network as not having run yet.
        self.hasrun = False
        
        # Create a hoc Random object for random wiring.
        self.randobj = h.Random()
        
        # Set the instance label.
        self.instlabel = label

    def setdefaultparams(self):
        # Should be instantiated with parameter settings used for the 
        # particular model.         
        pass
        
    def prenrnrun(self):
        # Should be instantiated with code for the particular model. 
        pass
        
    def postnrnrun(self):
        # Should be instantiated with code for the particular model.        
        pass 
