"""
neuralsim.py -- module for classes for a simulation run of a neural network

Classes for groups of neural units:
  NeuralSimulation (base class)
  
Initial design and implementation: George Chadderdon, 2013

Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import numpy.
import numpy

# Import the time module.
import time

# Import the pickle module.
import pickle

# Import the timed callbacks module.
import timedcallback as tcb

# Load the neural unit module.
import neuralunit as nun

# Load the neural connections module.
import neuralconn as nconn

class NeuralSimulation(object):
    """
    Class encapsulating a particular simulation run for an entire NEURON neural 
    network model.   One or more instances may be maintained of this on any run 
    of the code.

    Methods:
      __init__(self): constructor, sets up an un-run simulation
      setdefaultparams(self): set default values for the simulation parameters      
      globalfinits(self): FInitializeHandler called in beginning stages of run
      postcallback(self, simtime, callbackfunc): posts a timed callback for 
        NEURON to trigger at simtime ms
      recordcallbackevent(self, eventname): record an event time in the 
        recevent dictionary under the key defined by the particular eventname
      samplecallback(self): a sample callback function for testing
      run(self): run the simulation on the network
      save(self, filename): save the important sim info to a file
      load(self, filename): read the important sim info from a file
      showresults(self): show results collected from the sim (e.g. spike times
        and plots of activity)

    Attributes:
      classlabel (class): text label for the class
      instlabel: instance label (usable for tagging)
      thenetwork: pointer to the NeuralNetwork object whose sim is being run
      networkparams: the network parameters that were (last) used for the run
      simparams: dictionary of simulation (as opposed to network) parameters
        'simduration': simulation duration (in ms)           
      reccalls: dictionary storing the functions to pull out the desired data 
        after a sim run
      recdata: dictionary storing the data itself after the run
      recevents: dictionary storing times for events in the run (including 
        callbacks)
      runtimestamp: time of the end of the last run of the simulation (in 
        second since UNIX epoch: time.time() method)
      initnewnetwork: a flag used to keep track of when we need to call a 
        NeuralNetwork constructor with networkparams passed in
      fih: FInitializeHandler hoc object for the simulation
      theactivesim: are we the sim being run at the moment?

    Usage:
    >>>thesim = neuralsim.NeuralSimulation(label='M1MultilayerSim')
    """

    classlabel = 'NeuralSimulation'
    numCreated = 0
    allSims = []
    
    def __init__(self, thenet=None, params={}, filename='', label='currentSim'):    
        # Create an empty dictionary for the simulation parameters.
        self.simparams = {}
        
        # Set default versions of the simulation parameters.
        self.setdefaultparams()
        
        # For each of our parameters, if the key is in the passed in parameters,
        # overwrite our parameter using the latter.
        for thekey in self.simparams.keys():
            if thekey in params:
                self.simparams[thekey] = params[thekey]
                 
        # Set the pointer to the NeuralNetwork object.
        self.thenetwork = thenet  
        
        # Create empty dictionaries for the record functions, data, and events.
        self.reccalls = {}
        self.recdata = {}
        self.recevents = {}
        
        # Set the instance label.
        self.instlabel = label
        
        # Default to not initializing a new NeuralNetwork object.
        self.initnewnetwork = False
        
        # If we weren't passed a proper NeuralNetwork object, and we have
        # a filename to load from...
        if ((self.thenetwork == None) and (filename != '')):
            # Load parameters from the file.
            self.load(filename)
            
            # Remember that we need to initialize the network using the loaded
            # parameters.
            self.initnewnetwork = True
            
        # Set up the FInitiializeHandler.  self.globalfinits() gets called after
        # mechanism INITIAL blocks.
        self.fih = h.FInitializeHandler(1, self.globalfinits)
        
        # Increment the instance count.
        NeuralSimulation.numCreated += 1
        NeuralSimulation.allSims.append(self)
        
    def setdefaultparams(self):
        # Should be instantiated with parameter settings used for the 
        # particular class of simulations.   
        # Example settings:
        #     self.simparams['simduration'] = 1500.0
        #     self.simparams['stimstart'] = 100.0              
        pass      
          
    def globalfinits(self):
        # Exit if we are not the active sim.
        if not self.theactivesim:
            return
            
        # Install all of the callbacks that belong to us on the list.
        for thecb in tcb.TimedCallback.allCBs:
            if thecb.thesim == self:
                thecb.install()
                    
        # Miscellaneous callbacks for the simulation should be posted in the 
        # the subclass.
        # Example setting:
        #self.postcallback(50.0, self.samplecallback) 
    
    # NOTE: This must be called during a sim, not before or after.    
    def postcallback(self, simtime, callbackfunc):
        h.cvode.event(simtime, callbackfunc)
        
    def recordcallbackevent(self, eventname):
        if not eventname in self.recevents:
            self.recevents[eventname] = numpy.array([])
        self.recevents[eventname] = \
            numpy.append(self.recevents[eventname], h.t)        
        
    def samplecallback(self):
        # Show the event happening.
        print 'samplecallback() triggered at t = %f' % h.t
        
        # Stuff that needs to happen goes here for real callbacks...
        
        # Add the time of the event to the recevents dictionary.
        self.recordcallbackevent('samplecallbackcbs')
           
    # NOTE: There is presently some weird behavior that can happen when you 
    # rerun a NeuralSimulation object that has already been run.  Some work 
    # needs to be done to make sure that repeated run() results make sense.
    def run(self):
        # Inactivate all of the sims but ourselves, so only one globalfinits()
        # gets executed.
        for asim in NeuralSimulation.allSims:
            if asim == self:
                asim.theactivesim = True
            else:
                asim.theactivesim = False
        
        # Set the events dictionary to being empty.
        self.recevents = {}
        
        # Save the network parameters we're using.
        self.networkparams = self.thenetwork.netparams.copy()
        
        # Set NEURON to stop at the simulation duration.
        h.tstop =  self.simparams['simduration']

        # Run any pre-NEURON-run() code that needs to be run for the units.
        for unit in nun.NeuralUnit.allUnits:
            unit.prenrnrun()
            
        # Run any pre-NEURON-run() code that needs to be run for the connections.
        for conn in nconn.NeuralConn.allConns:
            conn.prenrnrun()
            
        # Run any pre-NEURON-run() code that needs to be run for the model.
        self.thenetwork.prenrnrun()
        
        # Run the simulation.
        h.run()
        
        # Run any post-NEURON-run() code that needs to be run for the units.
        for unit in nun.NeuralUnit.allUnits:
            unit.postnrnrun()
            
        # Run any post-NEURON-run() code that needs to be run for the weights.
        for conn in nconn.NeuralConn.allConns:
            conn.postnrnrun()     
        
        # Run any post-NEURON-run() code that needs to be run for the model.
        self.thenetwork.postnrnrun()       
         
        # Save the end of run timestamp.
        self.runtimestamp = time.time()
        
        # Flag the network as having run.
        self.thenetwork.hasrun = True

        # Get the recorded data from the reccalls dictionary (by running the 
        # functions pointed to by the reccalls dictionary).
        for reckey in self.reccalls.keys():
            self.recdata[reckey] = self.reccalls[reckey]()
            
    def save(self, filename):
        # Open the output file for binary writing.
        outfile = open(filename, 'wb')
        
        # Pickle the sim and network parameters, the recorded (results) data, 
        # events, the instance label, and the end of run timestamp.
        # (2 = new binary protocol.)
        pickle.dump(self.simparams, outfile, 2)        
        pickle.dump(self.networkparams, outfile, 2)
        pickle.dump(self.recdata, outfile, 2)
        pickle.dump(self.recevents, outfile, 2)        
        pickle.dump(self.instlabel, outfile, 2)
        pickle.dump(self.runtimestamp, outfile, 2)
        
        # Close the file.
        outfile.close()
            
    # NOTE: This does NOT (yet) update the state of the NeuralNetwork object 
    # to reflect the simulation's state during or after the saved simulation.
    def load(self, filename):
        # Open the input file for binary reading.
        infile = open(filename, 'rb')
        
        # Unpickle the sim and network parameters and the recorded (results) data,
        # events, the instance label, and the end of run timestamp.
        # (2 = new binary protocol.)
        self.simparams = pickle.load(infile)        
        self.networkparams = pickle.load(infile)
        self.recdata = pickle.load(infile)
        self.recevents = pickle.load(infile)        
        self.instlabel = pickle.load(infile)
        self.runtimestamp = pickle.load(infile)
        
        # Close the file.
        infile.close()
        
    def showresults(self):
        # Should be instantiated with code for the particular model. 
        pass
