"""
neuralunit.py -- module for classes for model neural units

N.B.: Tentative name for the framework:
  PYNE-NEDL
    PY_thon / 
    NE_URON
    NE_twork
    D_esign
    L_ibrary
    
Classes for neural units:
  NeuralUnit (base class)
      SpikeUnit
  UnitLocation (base class)
 
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/30/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import numpy.
import numpy

# Load the neural connections module.
import neuralconn as nconn

class NeuralUnit(object):
    """
    Generic neural unit for the model.

    Methods:
      __init__(self,...): constructor
      prenrnrun(self): stub for code that needs to be called before h.run()
      postnrnrun(self): stub for code that needs to be called after h.run()      
      isexcitatory(self): returns whether the unit is excitatory (based on the NTs)
      isinhibitory(self): returns whether the unit is excitatory (based on the NTs)
      disable(self): disables the unit so no output comes out
      enable(self): enables or reenables the unit's output
      isdisabled(self): returns whether the unit is disabled
      getinconnattachpoint(self, src): return a hoc object for NetCons using 
        this unit as a destination.  src is passed in so that the unit may, for 
        example, decide where to attach to based on whether the src unit
        is excitatory or inhibitory 
      getoutconnattachpoint(self): return a hoc object for NetCons using this unit
        as a source
      getfaninconns(self,connpool): return a ConnGroup containing convergent 
        connections from a chosen connpool into the unit (the default pool is the 
        list of all connections)
      getfanoutconns(self,connpool): return a ConnGroup containing divergent 
        connections from a chosen connpool out of the unit (the default pool is the 
        list of all connections)
      distfromunit(self, otherunit): return the distance from another NeuralUnit 
        (None if it's not valid)
              
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances
      instlabel: instance label (usable for tagging the unit)
      NTs: neurotransmitters released by the unit (list of strings)
        'GLU' -- glutamate
        'GABA' -- GABA
        'DA' -- dopamine
        '5HT' -- serotonin
        'NE' -- norepinephrine
        'ACH' -- acetylcholine
        having as list allows co-transmitters (e.g. ['DA','GLU'])
      loc: UnitLocation of cell (None means none assigned yet.)
      outconnsect: hoc Section which which NetCons can use as source
        (None means that nothing is appropriate here.)
      outconnthresh: threshold NetCons use that use this unit as a source
        (None means the threshold is not to be adjusted.)

    Usage:
    >>>nu = neuralunit.NeuralUnit()
    """

    classlabel = 'NeuralUnit'
    numCreated = 0
    allUnits = []

    def __init__(self, NTs=['GLU'], label=''):
        # Set the neurotransmitters released (determines excitatory vs. inhibitory)
        self.NTs = NTs

        # Set the instance label.
        self.instlabel = label

        # Initialize to no location.
        self.loc = None
        
        # Set the output connection Section to nothing.  This should be 
        # overridden for units that have compartments.
        self.outconnsect = None
        
        # Set the output connection threshold for NetCons to None.
        self.outconnthresh = None
        
        # Increment the instance count.
        NeuralUnit.numCreated += 1
        NeuralUnit.allUnits.append(self)
        
    def prenrnrun(self):
        # Should be instantiated with code for the particular NeuralUnit subclass. 
        pass
        
    def postnrnrun(self):
        # Should be instantiated with code for the particular NeuralUnit subclass.        
        pass 

    def isexcitatory(self):
        if 'GLU' in self.NTs:
            return True
        return False

    def isinhibitory(self):
        if 'GABA' in self.NTs:
            return True
        return False
        
    def disable(self):
        print '%s NeuralUnit: no effect: disable() needs to be instantiated by the specific class.' %\
            self.instlabel
        
    def enable(self):
        print '%s NeuralUnit: no effect: enable() needs to be instantiated by the specific class.' %\
            self.instlabel
            
    def isdisabled(self):
        print '%s NeuralUnit: no effect: isdisabled() needs to be instantiated by the specific class.' %\
            self.instlabel            
            
    def getinconnattachpoint(self, src):
        # NOTE: Subclasses should override returning the desired NetCon attach 
        # object.
        return None
        
    def getoutconnattachpoint(self):
        # NOTE: Subclasses should override returning the desired NetCon attach 
        # object.       
        return None
                
    def getfaninconns(self,connpool=None,incdisabled=False):
        # If no connection pool was selected to search from, default to a pool 
        # containing all NeuralConns.
        if connpool == None:
            connpool = nconn.ConnGroup(nconn.NeuralConn.allConns)
            
        # Create a new ConnGroup for the convergent connections from the pool.
        cgrp = nconn.ConnGroup()
        
        # For all connections in the pool...
        for conn in connpool:
            # If the destination address matches us...
            if conn.dest == self:
                # If the connection is not disabled or we are including disabled
                # connections, add to the list.
                if (not conn.isdisabled()) or incdisabled:
                    cgrp.append(conn)
        
        return cgrp

    def getfanoutconns(self,connpool=None,incdisabled=False):
        # If no connection pool was selected to search from, default to a pool 
        # containing all NeuralConns.
        if connpool == None:
            connpool = nconn.ConnGroup(nconn.NeuralConn.allConns)
            
        # Create a new ConnGroup for the divergent connections from the pool.
        cgrp = nconn.ConnGroup()
        
        # For all connections in the pool...
        for conn in connpool:
            # If the source addresses matches us...
            if conn.src == self:
                # If the connection is not disabled or we are including disabled
                # connections, add to the list.
                if (not conn.isdisabled()) or incdisabled:
                    cgrp.append(conn)
        
        return cgrp
        
    def distfromunit(self, otherunit):
        if otherunit.loc == None:
            return None
        else:
            return self.loc.distfrom(otherunit.loc)
                
class SpikeUnit(NeuralUnit):
    """
    Generic spiking unit (spikes for output) for the model.

    Methods:
      __init__(self,...): constructor
      getinconnattachpoint(self, src): return a hoc object for NetCons using 
        this unit as a destination.  src is passed in so that the unit may, for 
        example, decide where to attach to based on whether the src unit
        is excitatory or inhibitory 
      getoutconnattachpoint(self): return a hoc object for NetCons using this unit
        as a source      
      turnonrecspks(self): turns on the recording of spikes during the sim in recspks
      turnoffrecspks(self): turns off the recording of spikes during the sim in recspks
      isrecording(self): returns whether we are recording spikes from the unit
      getrecspks(self): returns a numpy array of the recorded spike times
      
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances      
      hobj: hoc object (of unspecified type) attached to the unit
      recspks: hoc Vector object for recorded spikes
      recnetcon: NetCon hoc object spawned to record spikes during a sim

    Usage:
    >>>spku = neuralunit.SpikeUnit()
    """

    classlabel = 'SpikeUnit'
    numCreated = 0
    allUnits = []

    def __init__(self, NTs=['GLU'], label=''):
        # Set superclass parameters.
        super(SpikeUnit,self).__init__(NTs,label)

        # Set a nil hoc object for the unit.
        self.hobj = None
        
        # Set up a nil hoc object for the recording NetCon.
        self.recnetcon = None
        
        # Set up nil hoc object for spike recording.
        self.recspks = None

        # Increment the instance count.
        SpikeUnit.numCreated += 1
        SpikeUnit.allUnits.append(self)
        
    def getinconnattachpoint(self, src):
        # Return the hoc object itself (good for ARTCELLs, but not 
        # compartmental cells).             
        return self.hobj
        
    def getoutconnattachpoint(self):
        # Return the hoc object itself (good for ARTCELLs, but not 
        # compartmental cells).     
        return self.hobj
                
    def turnonrecspks(self):
        if not self.isrecording():
            # Set up a recording NetCon.
            self.recnetcon = h.NetCon(self.hobj, None)
        
            # Set up a hoc Vector for spike recording.
            self.recspks = h.Vector()
        
            # Turn on the recording.
            self.recnetcon.record(self.recspks)
        
    def turnoffrecspks(self):
        # Null out the recording NetCon and recording vectors.
        self.recnetcon = None
        self.recspks = None
       
    def isrecording(self):
        return (not self.recnetcon == None)
         
    def getrecspks(self):
        # Give an error and exit if recording is not enabled.
        if not self.isrecording():
            print 'ERROR in %s NeuralUnit: Recording not enabled.' %\
                self.instlabel      
            return numpy.array([])
              
        # Get a numpy array version of the recorded spikes.
        return numpy.array(self.recspks)
        
class UnitLocation(object):
    """
    Location of a NeuralUnit.  By default, this has an x, y, and z location.
    These values default to 0.0, so you can use 2 out of 3 of them and the 
    Euclidean distances calculated by distfrom should work out fine.  If you 
    want to use a different distance metric however or add other location 
    parameters, you should subclass this class.

    Methods:
      dist2locs(loc1,loc2) (class): get the (Euclidean) distance between 
        two locations    
      __init__(self,...): constructor
      distfrom(otherloc): distance from another UnitLocation (Euclidean here)
              
    Attributes:
      x -- x location
      y -- y location
      z -- z location

    Usage:
    >>>loc = neuralunit.UnitLocation(x=2.0,y=5.0,z=3.0)
    """
    
    @staticmethod
    def dist2locs(loc1, loc2):
        dx = loc2.x - loc1.x
        dy = loc2.y - loc1.y
        dz = loc2.z - loc1.z
        return numpy.sqrt(dx*dx + dy*dy + dz*dz)

    def __init__(self, x=0.0, y=0.0, z=0.0):
        # Set the values.
        self.x = x
        self.y = y
        self.z = z
        
    def distfrom(self, otherloc):
        dx = otherloc.x - self.x
        dy = otherloc.y - self.y
        dz = otherloc.z - self.z
        return numpy.sqrt(dx*dx + dy*dy + dz*dz)
