"""
stdp.py -- module for STDP-using neural connections

Classes for STDP neural connections:
  STDPConn (PlasticConn)
  
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import the neural connections module.
import neuralconn as nconn

# Import numpy.
import numpy

class STDPConn(nconn.PlasticConn):
    """
    STDP-using neural connection for the model.

    Methods:
      __init__(self,...): constructor
      uniqueinit(self, ...): constructor code unique to our class.  This 
        function is needed to deal with multiple inheritance      
      get_tau_Hebb(self): get Hebbian decay time constant
      set_tau_Hebb(self,value): set Hebbian decay time constant  
      get_tau_antiHebb(self): get anti-Hebbian decay time constant
      set_tau_antiHebb(self,value): set anti-Hebbian decay time constant     
      get_maxadj_Hebb(self): get Hebbian maximum weight adjustment
      set_maxadj_Hebb(self,value): set Hebbian maximum weight adjustment  
      get_maxadj_antiHebb(self): get anti-Hebbian maximum weight adjustment
      set_maxadj_antiHebb(self,value): set anti-Hebbian maximum weight 
        adjustment     
      get_maweight(self): get maximum weight value
      set_maxweight(self,value): set maximum weight value  
      getSTDPcurve(self): return tvec / vec arrays giving the STDP curve
      turnonsoft_thresh(self): turn on soft-thresholding
      turnoffsoft_thresh(self): turn on soft-thresholding      
      usingsoft_thresh(self): are we using soft-thresholding?      
      turnonlearning(self): turns on learning
      turnofflearning(self): turns off learning

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances      
      dummysection (class): hoc object which is a dummy Section to be used by all the 
        weight adjusters (hopefully saving memory per synapse)
      wtadjobj: hoc object for the weight adjuster
      prewtadjnc: NetCon object for pre-synaptic cell to weight adjuster signal
      postwtadjnc: NetCon object for post-synaptic cell to weight adjuster signal

    Usage:
    - stdp_wtadj_lastspks.mod needs to be installed in special.    
        
    >>>nlc1 = stdp.STDPConn(prx,pox)
    """

    classlabel = 'STDPConn'
    numCreated = 0
    allConns = []    
    dummysection = h.Section()
    
    def __init__(self, src, dest, weight=0.0, delay=0.0, tau_Hebb=10.0, \
        tau_antiHebb=10.0, maxadj_Hebb=1.0, maxadj_antiHebb=-1.0, \
        maxweight=15.0, soft_thresh=False, label=''):
        # Set superclass parameters.
        super(STDPConn,self).__init__(src, dest, weight, delay, label)
        
        # Do init specific to STDPConn.  We use STDPConn. prefix because 
        # object no constructed yet.
        STDPConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh)

        # Set the pointer using the appropriate weight index.
        h.setpointer(self.hobj._ref_weight[0],'wsyn',self.wtadjobj)
 
    def uniqueinit(self, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=15.0, soft_thresh=False):
        # Create a dummy Section object to allow proper initialization.
        # Only need this if the class variable won't work for separate instances.
        # I'm not sure yet if that's the case.
        #self.dummysection = h.Section()

        # Create the weight adjuster object.  The pointer back to the weight 
        # should be set elsewhere.
#        self.wtadjobj = h.StdpWtAdj(0.5,sec=self.dummysection)  # use only if necessary
        self.wtadjobj = h.StdpWtAdjLastSpks(0.5,sec=STDPConn.dummysection)
#        self.wtadjobj = h.StdpWtAdjTwoElig(0.5,sec=STDPConn.dummysection)   

        # Create the pre-synaptic NetCon and set its weight to positive.
        if self.src.outconnsect == None:
            self.prewtadjnc = h.NetCon(self.src.getoutconnattachpoint(), \
                self.wtadjobj)
        else:
            self.prewtadjnc = h.NetCon(self.src.getoutconnattachpoint(), \
                self.wtadjobj, sec=self.src.outconnsect)
        self.prewtadjnc.delay = 0.0
        self.prewtadjnc.weight[0] = 1.0
        if self.src.outconnthresh != None:
            self.prewtadjnc.threshold = self.src.outconnthresh
            
        # Create the post-synaptic NetCon and set its weight to negative.
        if self.dest.outconnsect == None:
            self.postwtadjnc = h.NetCon(self.dest.getoutconnattachpoint(), \
                self.wtadjobj)
        else:
            self.postwtadjnc = h.NetCon(self.dest.getoutconnattachpoint(), \
                self.wtadjobj, sec=self.dest.outconnsect)
        self.postwtadjnc.delay = 0.0
        self.postwtadjnc.weight[0] = -1.0
        if self.dest.outconnthresh != None:
            self.prewtadjnc.threshold = self.dest.outconnthresh

        # Set the parameters for the weight adjuster.
        self.set_tau_Hebb(tau_Hebb)
        self.set_tau_antiHebb(tau_antiHebb) 
        self.set_maxadj_Hebb(maxadj_Hebb)
        self.set_maxadj_antiHebb(maxadj_antiHebb)    
        self.set_maxweight(maxweight)
        if soft_thresh:
            self.turnonsoft_thresh()
        else:
            self.turnoffsoft_thresh()  
              
        # Increment the instance count.
        STDPConn.numCreated += 1
        STDPConn.allConns.append(self)
        
    def get_tau_Hebb(self):
        return self.wtadjobj.tau_Hebb
                
    def set_tau_Hebb(self, value):
        self.wtadjobj.tau_Hebb = value
                    
    def get_tau_antiHebb(self):
        return self.wtadjobj.tau_antiHebb
        
    def set_tau_antiHebb(self, value):
        self.wtadjobj.tau_antiHebb = value   
        
    def get_maxadj_Hebb(self):
        return self.wtadjobj.maxadj_Hebb
                
    def set_maxadj_Hebb(self, value):
        self.wtadjobj.maxadj_Hebb = value
                    
    def get_maxadj_antiHebb(self):
        return self.wtadjobj.maxadj_antiHebb
        
    def set_maxadj_antiHebb(self, value):
        self.wtadjobj.maxadj_antiHebb = value   
      
    def get_maxweight(self):
        return self.wtadjobj.wmax
        
    def set_maxweight(self, value):
        self.wtadjobj.wmax = value            
              
    def getSTDPcurve(self):
        # Set sample period (in ms)
        sampinterval = 0.1      
        
        # Set how many time constants to plot out to.
        timeconsts = 3.5   
        
        # Create tvec, vec, with vec zeroed for starts.     
        mint = -timeconsts * self.get_tau_antiHebb()
        maxt = timeconsts * self.get_tau_Hebb()     
        numsamps = int ((maxt - mint) / sampinterval)
        tvec = numpy.linspace(mint, maxt, numsamps, endpoint=False)
        vec = numpy.zeros(len(tvec))
        
        # Generate the Hebbian (right) side of curve.
        for tind, time in enumerate(tvec):
            if time > 0:
                vec[tind] = self.get_maxadj_Hebb() * numpy.exp(-time / \
                    self.get_tau_Hebb())
                
        # Generate the anti-Hebbian (left) side of curve.
        for tind, time in enumerate(tvec):
            if time < 0:
                vec[tind] = self.get_maxadj_antiHebb() * numpy.exp(time / \
                    self.get_tau_antiHebb())
                        
        return tvec, vec
                      
    def turnonsoft_thresh(self):
        self.wtadjobj.adj_soft_thresh = 1
 
    def turnoffsoft_thresh(self):
        self.wtadjobj.adj_soft_thresh = 0 
        
    def usingsoft_thresh(self):
        return (self.wtadjobj.adj_soft_thresh == 1)
                          
    def turnonlearning(self):
        nconn.PlasticConn.turnonlearning(self)
        self.wtadjobj.on = 1
         
    def turnofflearning(self):
        nconn.PlasticConn.turnofflearning(self)
        self.wtadjobj.on = 0   
