"""
stdp_rl.py -- module for STDP- and reinforcement learning (RL) using neural 
  connections

Classes for STDP RL neural connections:
  STDPRLConn (PlasticConn)
  
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/26/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import the neural simulation module.
import neuralsim as nsim

# Import the timed callbacks module.
import timedcallback as tcb

# Import the neural connections module.
import neuralconn as nconn

# Import numpy.
import numpy

# Import the find function.
from pylab import find

class STDPRLConn(nconn.PlasticConn):
    """
    STDP- and RL-using neural connection for the model.

    Methods:
      __init__(self,...): constructor
      uniqueinit(self,...): constructor code unique to our class.  This 
        function is needed to deal with multiple inheritance      
      get_tau_Hebb_STDP(self): get Hebbian STDP decay time constant
      set_tau_Hebb_STDP(self,value): set Hebbian STDP decay time constant  
      get_tau_antiHebb_STDP(self): get anti-Hebbian STDP decay time constant
      set_tau_antiHebb_STDP(self,value): set anti-Hebbian STDP decay time constant     
      get_maxadj_Hebb(self): get Hebbian maximum weight adjustment
      set_maxadj_Hebb(self,value): set Hebbian maximum weight adjustment  
      get_maxadj_antiHebb(self): get anti-Hebbian maximum weight adjustment
      set_maxadj_antiHebb(self,value): set anti-Hebbian maximum weight 
        adjustment     
      get_maxeligwind_Hebb(self): get Hebbian maximum eligibility time window 
        (in ms)
      set_maxeligwind_Hebb(self,value): set Hebbian maximum eligibility time 
        window (in ms) 
      get_maxeligwind_antiHebb(self): get anti-Hebbian maximum eligibility time 
        window (in ms)
      set_maxeligwind_antiHebb(self,value): set anti-Hebbian eligibility time 
        window (in ms)
      get_eliglength_Hebb(self): get Hebbian length of eligibility trace 
        (in ms)
      set_eliglength_Hebb(self,value): set Hebbian length of eligibility trace
        (in ms) 
      get_eliglength_antiHebb(self): get anti-Hebbian length of eligibility
        trace (in ms)
      set_eliglength_antiHebb(self,value): set anti-Hebbian length of 
        eligibility trace (in ms)   
      get_maxeliggain_Hebb(self): get Hebbian maximum eligibility trace gain
      set_maxeliggain_Hebb(self,value): set Hebbian maximum eligibility trace
        gain 
      get_maxeliggain_antiHebb(self): get anti-Hebbian maximum eligibility trace 
        gain
      set_maxeliggain_antiHebb(self,value): set anti-Hebbian eligibility trace 
        gain
      get_maweight(self): get maximum weight value
      set_maxweight(self,value): set maximum weight value  
      getSTDPcurve(self): return tvec / vec arrays giving the STDP curve
      getHebbeliggaincurve(self): return tvec / vec arrays giving the curve for
        the Hebbian eligibility trace
      getantiHebbeliggaincurve(self): return tvec / vec arrays giving the curve
        for the ant-Hebbian eligibility trace
      turnonsoft_thresh(self): turn on soft-thresholding
      turnoffsoft_thresh(self): turn on soft-thresholding      
      usingsoft_thresh(self): are we using soft-thresholding?      
      turnonlearning(self): turns on learning
      turnofflearning(self): turns off learning
      turnonSTDP(self): turns on STDP learning
      turnoffSTDP(self): turns off STDP learning
      turnonRL(self): turns on reinforcement learning
      turnoffRL(self): turns off reinforcement learning
      turnonexpdecay_elig(self): turns on exponential decay eligibility time 
        window 
      turnoffexpdecay_elig(self): turns off exponential decay eligibility time 
        window             
      turnonHebbelig(self): turns on Hebbian eligibility traces
      turnoffHebbelig(self): turns off Hebbian eligibility traces
      turnonantiHebbelig(self): turns on anti-Hebbian eligibility traces
      turnoffantiHebbelig(self): turns off anti-Hebbian eligibility traces   
      turnonrecHebbelig(self, thesim, recinterval): turns on recording of 
        Hebbian eligibility traces
      turnonrecantiHebbelig(self, thesim, recinterval): turns on recording of 
        anti-Hebbian eligibility traces            
      getrecHebbelig(self): get tvec and vec for the Hebbian eligibility trace
        gains
      getrecantiHebbelig(self): get tvec and vec for the anti-Hebbian eligibility 
        trace gains                   
      getHebbeliggain(self): get the current Hebbian eligibility trace gain    
      getantiHebbeliggain(self): get the current anti-Hebbian eligibility trace 
        gain                
      recHebbeliggain(self): callback "payload function" for recording the 
        Hebbian eligibility trace gains
      recantiHebbeliggain(self): callback "payload function" for recording the 
        anti-Hebbian eligibility trace gains            
      rewardpunish(self, reinf_signal): send a reward or punisher to the unit 
        for reinforcement learning update

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allConns (class): list of all instances      
      dummysection (class): hoc object which is a dummy Section to be used by all the 
        weight adjusters (hopefully saving memory per synapse)
      wtadjobj: hoc object for the weight adjuster
      prewtadjnc: NetCon object for pre-synaptic cell to weight adjuster signal
      postwtadjnc: NetCon object for post-synaptic cell to weight adjuster signal    
      recHebbelig: hoc Vector object for storing recorded Hebbian eligibility 
        trace gains
      recantiHebbelig: hoc Vector object for storing recorded anti-Hebbian 
        eligibility trace gains
      recHebbeliginterval: recording interval for storing Hebbian eligibility
        trace gains
      recantiHebbeliginterval: recording interval for storing anti-Hebbian 
        eligibility trace gains        
      recHebbeligcb: callback for Hebbian eligibility gain recording
      recantiHebbeligcb: callback for anti-Hebbian eligibility gain recording
              
    Usage:
    - stdp_rl_wtadj.mod needs to be installed in special.    
        
    >>>nlc1 = stdprl.STDPRLConn(prx,pox)
    """

    classlabel = 'STDPRLConn'
    numCreated = 0
    allConns = []    
    dummysection = h.Section()
    
    def __init__(self, src, dest, weight=0.0, delay=0.0, tau_Hebb=10.0, \
        tau_antiHebb=10.0, maxadj_Hebb=1.0, maxadj_antiHebb=-1.0, \
        maxweight=15.0, soft_thresh=False, maxeligwind_Hebb=10.0, \
        maxeligwind_antiHebb=10.0, eliglength_Hebb=100.0, \
        eliglength_antiHebb=100.0, maxeliggain_Hebb=1.0, \
        maxeliggain_antiHebb=-1.0, expeligdecay=False, label=''):
        # Set superclass parameters.
        super(STDPRLConn,self).__init__(src, dest, weight, delay, label)

        # Do init specific to STDPRLConn.  We use STDPRLConn. prefix because 
        # object not constructed yet.
        STDPRLConn.uniqueinit(self, tau_Hebb, tau_antiHebb, maxadj_Hebb, \
            maxadj_antiHebb, maxweight, soft_thresh, maxeligwind_Hebb, \
            maxeligwind_antiHebb, eliglength_Hebb, eliglength_antiHebb, \
            maxeliggain_Hebb, maxeliggain_antiHebb, expeligdecay)

        # Set the pointer using the appropriate weight index.
        h.setpointer(self.hobj._ref_weight[0],'wsyn',self.wtadjobj)
        
    def uniqueinit(self, tau_Hebb=10.0, tau_antiHebb=10.0, maxadj_Hebb=1.0, \
        maxadj_antiHebb=-1.0, maxweight=15.0, soft_thresh=False, \
        maxeligwind_Hebb=10.0, maxeligwind_antiHebb=10.0, \
        eliglength_Hebb=100.0, eliglength_antiHebb=100.0, \
        maxeliggain_Hebb=1.0, maxeliggain_antiHebb=-1.0, \
        expeligdecay=False):
        # Create a dummy Section object to allow proper initialization.
        # Only need this if the class variable won't work for separate instances.
        # I'm not sure yet if that's the case.
        #self.dummysection = h.Section()

        # Create the weight adjuster object.  The pointer back to the weight 
        # should be set elsewhere.
#        self.wtadjobj = h.StdpRLWtAdj(0.5,sec=self.dummysection)  # use only if necessary
        self.wtadjobj = h.StdpRLWtAdj(0.5,sec=STDPRLConn.dummysection)      

        # Create the pre-synaptic NetCon and set its weight to positive.
        if self.src.outconnsect == None:
            self.prewtadjnc = h.NetCon(self.src.getoutconnattachpoint(), \
                self.wtadjobj)
        else:
            self.prewtadjnc = h.NetCon(self.src.getoutconnattachpoint(), \
                self.wtadjobj, sec=self.src.outconnsect)
        self.prewtadjnc.delay = 0.0
        self.prewtadjnc.weight[0] = 1.0
        if self.src.outconnthresh != None:
            self.prewtadjnc.threshold = self.src.outconnthresh
            
        # Create the post-synaptic NetCon and set its weight to negative.
        if self.dest.outconnsect == None:
            self.postwtadjnc = h.NetCon(self.dest.getoutconnattachpoint(), \
                self.wtadjobj)
        else:
            self.postwtadjnc = h.NetCon(self.dest.getoutconnattachpoint(), \
                self.wtadjobj, sec=self.dest.outconnsect)
        self.postwtadjnc.delay = 0.0
        self.postwtadjnc.weight[0] = -1.0
        if self.dest.outconnthresh != None:
            self.prewtadjnc.threshold = self.dest.outconnthresh
            
        # Set the parameters for the weight adjuster.
        self.set_tau_Hebb_STDP(tau_Hebb)
        self.set_tau_antiHebb_STDP(tau_antiHebb) 
        self.set_maxadj_Hebb(maxadj_Hebb)
        self.set_maxadj_antiHebb(maxadj_antiHebb)    
        self.set_maxweight(maxweight)
        if soft_thresh:
            self.turnonsoft_thresh()
        else:
            self.turnoffsoft_thresh()  
        self.set_maxeligwind_Hebb(maxeligwind_Hebb)
        self.set_maxeligwind_antiHebb(maxeligwind_antiHebb)
        self.set_eliglength_Hebb(eliglength_Hebb)
        self.set_eliglength_antiHebb(eliglength_antiHebb)
        self.set_maxeliggain_Hebb(maxeliggain_Hebb)
        self.set_maxeliggain_antiHebb(maxeliggain_antiHebb)
        if expeligdecay:        
            self.turnonexpdecay_elig()
        else:
            self.turnoffexpdecay_elig()
                        
        # Increment the instance count.
        STDPRLConn.numCreated += 1
        STDPRLConn.allConns.append(self)
        
    def get_tau_Hebb_STDP(self):
        return self.wtadjobj.tau_Hebb_STDP
                
    def set_tau_Hebb_STDP(self, value):
        self.wtadjobj.tau_Hebb_STDP = value
                    
    def get_tau_antiHebb_STDP(self):
        return self.wtadjobj.tau_antiHebb_STDP
        
    def set_tau_antiHebb_STDP(self, value):
        self.wtadjobj.tau_antiHebb_STDP = value   
                   
    def get_maxadj_Hebb(self):
        return self.wtadjobj.maxadj_Hebb
                
    def set_maxadj_Hebb(self, value):
        self.wtadjobj.maxadj_Hebb = value
                    
    def get_maxadj_antiHebb(self):
        return self.wtadjobj.maxadj_antiHebb
        
    def set_maxadj_antiHebb(self, value):
        self.wtadjobj.maxadj_antiHebb = value   

    def get_maxeligwind_Hebb(self):
        return self.wtadjobj.maxeligwind_Hebb
                
    def set_maxeligwind_Hebb(self, value):
        self.wtadjobj.maxeligwind_Hebb = value
                    
    def get_maxeligwind_antiHebb(self):
        return self.wtadjobj.maxeligwind_antiHebb
        
    def set_maxeligwind_antiHebb(self, value):
        self.wtadjobj.maxeligwind_antiHebb = value 
   
    def get_eliglength_Hebb(self):
        return self.wtadjobj.eliglength_Hebb
                
    def set_eliglength_Hebb(self, value):
        self.wtadjobj.eliglength_Hebb = value
                    
    def get_eliglength_antiHebb(self):
        return self.wtadjobj.eliglength_antiHebb
        
    def set_eliglength_antiHebb(self, value):
        self.wtadjobj.eliglength_antiHebb = value   

    def get_maxeliggain_Hebb(self):
        return self.wtadjobj.maxeliggain_Hebb
                
    def set_maxeliggain_Hebb(self, value):
        self.wtadjobj.maxeliggain_Hebb = value
                    
    def get_maxeliggain_antiHebb(self):
        return self.wtadjobj.maxeliggain_antiHebb
        
    def set_maxeliggain_antiHebb(self, value):
        self.wtadjobj.maxeliggain_antiHebb = value   

    def get_maxweight(self):
        return self.wtadjobj.wmax
        
    def set_maxweight(self, value):
        self.wtadjobj.wmax = value            
              
    def getSTDPcurve(self):
        # Set sample period (in ms)
        sampinterval = 0.1      
        
        # Set how many time constants to plot out to.
        timeconsts = 3.5   
        
        # Create tvec, vec, with vec zeroed for starts.     
        mint = -timeconsts * self.get_tau_antiHebb_STDP()
        maxt = timeconsts * self.get_tau_Hebb_STDP()     
        numsamps = int ((maxt - mint) / sampinterval)
        tvec = numpy.linspace(mint, maxt, numsamps, endpoint=False)
        vec = numpy.zeros(len(tvec))
        
        # Generate the Hebbian (right) side of curve.
        for tind, time in enumerate(tvec):
            if time > 0:
                vec[tind] = self.get_maxadj_Hebb() * numpy.exp(-time / \
                    self.get_tau_Hebb_STDP())
                
        # Generate the anti-Hebbian (left) side of curve.
        for tind, time in enumerate(tvec):
            if time < 0:
                vec[tind] = self.get_maxadj_antiHebb() * numpy.exp(time / \
                    self.get_tau_antiHebb_STDP())
                        
        return tvec, vec
             
    def  getHebbeliggaincurve(self):
        # Set sample period (in ms)
        sampinterval = 0.1      
        
        # Set how many time constants to plot out to in the case of exponential 
        # decay.
        timeconsts = 3.5   
        
        # Create tvec, vec, with vec zeroed for starts.     
        mint = 0.0
        if self.wtadjobj.expdecay_elig:
            maxt = timeconsts * self.get_eliglength_Hebb()   
        else:
            maxt = 1.2 * self.get_eliglength_Hebb()
        numsamps = int ((maxt - mint) / sampinterval)
        tvec = numpy.linspace(mint, maxt, numsamps, endpoint=False)
        vec = numpy.zeros(len(tvec))
        
        # If the trace uses exponential decay...
        if self.wtadjobj.expdecay_elig == 1.0:
            # Set the values according to e(-t  / decay_time_constant).
            for ind, t in enumerate(tvec):
                vec[ind] = self.get_maxeliggain_Hebb() * \
                    numpy.exp(-t / self.get_eliglength_Hebb())               
        # Else (the trace is a square-wave)...
        else:
            # Set all of the values within the eligibility trace length to the 
            # gain value.
            chnginds = find(tvec <= self.get_eliglength_Hebb())
            vec[chnginds] = self.get_maxeliggain_Hebb()
                    
        return tvec, vec

    def  getantiHebbeliggaincurve(self):
        # Set sample period (in ms)
        sampinterval = 0.1      
        
        # Set how many time constants to plot out to in the case of exponential 
        # decay.
        timeconsts = 3.5   
        
        # Create tvec, vec, with vec zeroed for starts.     
        mint = 0.0
        if self.wtadjobj.expdecay_elig:
            maxt = timeconsts * self.get_eliglength_antiHebb()   
        else:
            maxt = 1.2 * self.get_eliglength_antiHebb()
        numsamps = int ((maxt - mint) / sampinterval)
        tvec = numpy.linspace(mint, maxt, numsamps, endpoint=False)
        vec = numpy.zeros(len(tvec))
        
        # If the trace uses exponential decay...
        if self.wtadjobj.expdecay_elig == 1.0:
            # Set the values according to e(-t  / decay_time_constant).
            for ind, t in enumerate(tvec):
                vec[ind] = self.get_maxeliggain_antiHebb() * \
                    numpy.exp(-t / self.get_eliglength_antiHebb())               
        # Else (the trace is a square-wave)...
        else:
            # Set all of the values within the eligibility trace length to the 
            # gain value.
            chnginds = find(tvec <= self.get_eliglength_antiHebb())
            vec[chnginds] = self.get_maxeliggain_antiHebb()
        
        return tvec, vec
                 
    def turnonsoft_thresh(self):
        self.wtadjobj.adj_soft_thresh = 1
 
    def turnoffsoft_thresh(self):
        self.wtadjobj.adj_soft_thresh = 0 
        
    def usingsoft_thresh(self):
        return (self.wtadjobj.adj_soft_thresh == 1)
                          
    def turnonlearning(self):
        nconn.PlasticConn.turnonlearning(self)
        self.wtadjobj.on = 1
         
    def turnofflearning(self):
        nconn.PlasticConn.turnofflearning(self)
        self.wtadjobj.on = 0  
        
    def turnonSTDP(self):
        self.wtadjobj.STDPon = 1
 
    def turnoffSTDP(self):
        self.wtadjobj.STDPon = 0  
        
    def turnonRL(self):
        self.wtadjobj.RLon = 1
 
    def turnoffRL(self):
        self.wtadjobj.RLon = 0  
                     
    def turnonexpdecay_elig(self):
        self.wtadjobj.expdecay_elig = 1
 
    def turnoffexpdecay_elig(self):
        self.wtadjobj.expdecay_elig = 0  
 
    def turnonHebbelig(self):
        self.wtadjobj.elig_Hebb_on = 1
 
    def turnoffHebbelig(self):
        self.wtadjobj.elig_Hebb_on = 0  
        
    def turnonantiHebbelig(self):
        self.wtadjobj.elig_antiHebb_on = 1
 
    def turnoffantiHebbelig(self):
        self.wtadjobj.elig_antiHebb_on = 0                            
                    
    def turnonrecHebbelig(self, thesim, recinterval):
        # Create a hoc vector for the the Hebbian eligibility trace reading.
        self.recHebbelig = h.Vector()

        # Remember the recording interval.
        self.recHebbeliginterval = recinterval
                
        # Create a repeating callback that starts at the beginning of the sim
        # and calls recHebbeliggain every recording interval.
        self.recHebbeligcb = tcb.TimedCallback(self.recHebbeliggain, thesim, 
            0.0, callrepeatperiod=recinterval, label='recHebbeliggain')
            
    def turnonrecantiHebbelig(self, thesim, recinterval):
        # Create a hoc vector for the the anti-Hebbian eligibility trace reading.
        self.recantiHebbelig = h.Vector()

        # Remember the recording interval.
        self.recantiHebbeliginterval = recinterval
                
        # Create a repeating callback that starts at the beginning of the sim
        # and calls recantiHebbeliggain every recording interval.
        self.recantiHebbeligcb = tcb.TimedCallback(self.recantiHebbeliggain, 
            thesim, 0.0, callrepeatperiod=recinterval, label='recantiHebbeliggain')
                        
    def getrecHebbelig(self):
        receliggains = numpy.array(self.recHebbelig)
        tvec = numpy.linspace(0.0, self.recHebbeliginterval * len(receliggains), \
             num=len(receliggains), endpoint=False)
        return tvec, receliggains   
                  
    def getrecantiHebbelig(self):
        receliggains = numpy.array(self.recantiHebbelig)
        tvec = numpy.linspace(0.0, self.recantiHebbeliginterval * len(receliggains), \
             num=len(receliggains), endpoint=False)
        return tvec, receliggains  
                          
    def getHebbeliggain(self):
        return self.wtadjobj.Hebb_elig_gain()
        
    def getantiHebbeliggain(self):
        return self.wtadjobj.antiHebb_elig_gain()       
                  
    def recHebbeliggain(self):
        self.recHebbelig.append(self.getHebbeliggain())
                
    def recantiHebbeliggain(self):
        self.recantiHebbelig.append(self.getantiHebbeliggain())
                        
    def rewardpunish(self, reinf_signal):
        self.wtadjobj.reward_punish(reinf_signal) 
