"""
stim.py -- module for classes for stimuli to be delivered to the model

Classes for stimulus units:
  SpikeStimUnit (subclass of SpikeUnit)
      SpikeCallbackStimUnit
      NetStimUnit
      VecStimUnit
      
Initial design and implementation: George Chadderdon, 2013
 
Last update: 9/12/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import neural unit module.
import neuralunit as nun

# Import the timed callbacks module.
import timedcallback as tcb

class SpikeStimUnit(nun.SpikeUnit):
    """
    Generic spike stimulus unit for the model.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances
      
    Usage:
    >>>st = stim.SpikeStimUnit()
    """

    classlabel = 'SpikeStimUnit'
    numCreated = 0
    allUnits = []
    
    def __init__(self,  NTs=['GLU'], label=''):
        # Set superclass parameters.
        super(SpikeStimUnit,self).__init__(NTs, label)

        # Increment the instance count.
        SpikeStimUnit.numCreated += 1
        SpikeStimUnit.allUnits.append(self)
      
class SpikeCallbackStimUnit(SpikeStimUnit):
    """
    TimedCallback-based stimulus unit for the model.

    Methods:
      __init__(self): constructor
      disable(self): disables the unit so no output comes out
      enable(self): enables or reenables the unit's output      
      isdisabled(self): returns whether the unit is disabled
      turnonrecspks(self): turns on the recording of spikes during the sim in recspks
      isrecording(self): returns whether we are recording spikes from the unit 
      addspktimes(self, spktime): add TimedCallbacks for a set of spike times     
      firespike(self, thetime): post an output spike event at a particular time
                  
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances
      disabled: is the unit disabled?
      cblist: TimedCallback list for the individual spike events

    Usage:
    >>>nst = stim.SpikeCallbackStimUnit()
    """

    classlabel = 'SpikeCallbackStimUnit'
    numCreated = 0
    allUnits = []
    
    def __init__(self, NTs=['GLU'], label=''):
        # Set superclass parameters.
        super(SpikeCallbackStimUnit,self).__init__(NTs, label)
          
        # Set the unit to not disabled.
        self.disabled = False
        
        # Set up an empty callbacks list.
        self.cblist = []
        
        # Increment the instance count.
        SpikeCallbackStimUnit.numCreated += 1
        SpikeCallbackStimUnit.allUnits.append(self)
        
    def disable(self):
        self.disabled = True
        
    def enable(self):
        self.disabled = False
            
    def isdisabled(self):
        return self.disabled
        
    def turnonrecspks(self):
        if not self.isrecording():
            # Set up a hoc Vector for spike recording.
            self.recspks = h.Vector()
            
    def isrecording(self):
        return (not self.recspks == None)
                    
    def addspktimes(self, thesim, spktimes):
        for spktime in spktimes:
            self.cblist.append(tcb.TimedCallback(self.firespike, thesim, 
                spktime, label='firespike'))
        
    def firespike(self):
        # If the unit is not disabled...
        if not self.isdisabled():
            # For all of the fanout weights, send a spiking event with the current
            # time + the NetCon delay.
            for outwt in self.getfanoutwts():
                outwt.hobj.event(h.t + outwt.hobj.delay)
                
            # Add the current time to the spike recording list.
            if self.isrecording():
                self.recspks.append(h.t)
              
class NetStimUnit(SpikeStimUnit):
    """
    Regular NetStim-based stimulus unit for the model.

    Methods:
      __init__(self): constructor
      
      get_interval(self): get the period interval for spike generation (in ms)
      set_interval(self,value): set the period interval for spike generation 
        (in ms)
      get_number(self): get the number of spikes to be produced
      set_number(self,value): set the number of spikes to be produced        
      get_start(self): get the first spike time (in ms)
      set_start(self,value): set the first spike time (in ms)   
      get_noise(self): get the noise amount (0 to 1)
      set_noise(self,value): set the noise amount (0 to 1)              
      disable(self): disables the unit so no output comes out
      enable(self): enables or reenables the unit's output      
      isdisabled(self): returns whether the unit is disabled
            
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances      
      hobj: hoc NetStim object
      randobj: hoc Random object for randomizing spike times
      randseed: random number seed for spike times

    Usage:
    >>>nst = stim.NetStimUnit()
    """

    classlabel = 'NetStimUnit'
    numCreated = 0
    allUnits = []
    
    def __init__(self, NTs=['GLU'], label='', interval=10, number=5, start=0.0, \
        noise=0.0, randseed=1):   # NOTE: passing in randseed=0 seems to make random
        # Set superclass parameters.
        super(NetStimUnit,self).__init__(NTs, label)

        # Create the hoc NetStim object.  (Use the netstim2.mod version of 
        # NetStim rather than the default.
        self.hobj = h.NetStim2(0.5)
        self.hobj.interval = interval
        self.hobj.number = number
        self.hobj.start = start
        self.hobj.noise = noise
        
        # Remember the random number seed.
        self.randseed = randseed
        
        # Create a hoc Random object for random stimuli.
        self.randobj = h.Random()
        
        # Set a negative exponential distribution with the Poisson-distributed
        # interval set to 1.
        self.randobj.negexp(1)
        
        # Use the MCellRan4 version of the generator.  [not sure what the 
        # significance of this is]
        self.randobj.MCellRan4(self.randseed,self.randseed)
  
        # Set the NetStim to use this random number generator.
        self.hobj.noiseFromRandom(self.randobj) 
        
        # Increment the instance count.
        NetStimUnit.numCreated += 1
        NetStimUnit.allUnits.append(self)
        
    def get_interval(self):
        return self.hobj.interval
        
    def set_interval(self, value):
        self.hobj.interval = value        
        
    def get_number(self):
        return self.hobj.number
        
    def set_number(self, value):
        self.hobj.number = value 
        
    def get_start(self):
        return self.hobj.start
        
    def set_start(self, value):
        self.hobj.start = value        
        
    def get_noise(self):
        return self.hobj.noise
        
    def set_noise(self, value):
        self.hobj.noise = value 
               
    def disable(self):
        if not self.isdisabled():       
            self.hobj.disabled = 1
        
    def enable(self):
        if self.isdisabled(): 
            self.hobj.disabled = 0
            
    def isdisabled(self):
        return (self.hobj.disabled == 1)

class VecStimUnit(SpikeStimUnit):
    """
    VecStim-based stimulus unit for the model.  This can be used to play spikes
    at particular times.

    Methods:
      __init__(self): constructor
      disable(self): disables the unit so no output comes out
      enable(self): enables or reenables the unit's output      
      isdisabled(self): returns whether the unit is disabled
            
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allUnits (class): list of all instances      
      hobj: hoc NetStim object
      spktimes: hoc Vector object containing spike times
      savedspktimes: saved spike times while the unit is disabled

    Usage:
    >>>vst = stim.VecStimUnit([40, 50, 100, 150, 170])
    """

    classlabel = 'VecStimUnit'
    numCreated = 0
    allUnits = []
    
    def __init__(self, spktimes, NTs=['GLU'], label=''):
        # Set superclass parameters.
        super(VecStimUnit,self).__init__(NTs, label)

        # Create the hoc VecStim object.
        self.hobj = h.VecStim(0.5)

        # Save the spike times.
        self.spktimes = h.Vector(spktimes)
        
        # Set savedspktimes to nothing.
        self.savedspktimes = None
        
        # Set the spike times for playback.
        self.hobj.play(self.spktimes)
        
        # Increment the instance count.
        VecStimUnit.numCreated += 1
        VecStimUnit.allUnits.append(self)
        
    def disable(self):
        if not self.isdisabled():
            self.savedspktimes = h.Vector()
            self.savedspktimes.copy(self.spktimes)
            self.spktimes = h.Vector([])
            self.hobj.play(self.spktimes)
        
    def enable(self):
        if self.isdisabled(): 
            self.spktimes.copy(self.savedspktimes)
            self.hobj.play(self.spktimes)            
            self.savedspktimes = None
            
    def isdisabled(self):
        return (not self.savedspktimes == None)
