"""
timedcallback.py -- module for timed NEURON callbacks

Classes for timed callback:
  TimedCallback (base class)
  
Initial design and implementation: George Chadderdon, 2013
 
Last update: 7/8/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

class TimedCallback(object):
    """
    Class for a timed callback in NEURON.  Each instance encapsulates
    code to be run at particular scheduled simulation-times.
    
    Methods:
      __init__(self, callbackfunc, thesim, firstcalltime, callrepeatperiod, label): 
        constructor for the callback
      install(self): installs the callback during the globalfinits() call from the 
        NeuralSimulation object
      cbexec(self): runs the actual callback and handles the other generic 
        callback functionality along with the "payload"
      disable(self): disables the callback 
      enable(self): enables or reenables the callback    
      turnonreccalltimes(self): turns on recording of the callback times
      turnoffreccalltimes(self): turns off recording of the callback times
                
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allCBs (class): list of all instances
      instlabel: instance label (usable for tagging)
      thecb: the function (pointer) to the "payload" of this callback
      thesim: pointer to the simulation all the events are running in
      firstcalltime: the simulation time (in ms) when the callback first 
        executes
      callrepeatperiod: the repetition period (in ms) for when the 
        callback executes again (-1 means there is no repetition)
      verbose: verbosity of debugging information about the events
      disabled: is the callback disabled?
      reccalltimes: are we recording the callback times?
          
    Usage:
    >>>thecb = tcb.TimedCallback(myfunc, thesim, label='MyFunc')
    """

    classlabel = 'TimedCallback'
    numCreated = 0
    allCBs = []
    
    def __init__(self, callbackfunc, thesim, firstcalltime=0.0, \
        callrepeatperiod=-1, label=''):
        # Set the instance label.
        self.instlabel = label

        # Set the callback function.
        self.thecb = callbackfunc
        
        # Set the simulation we're running in.
        self.thesim = thesim
        
        # Set the first callback simulation-time.
        self.firstcalltime = firstcalltime
        
        # Set the callback repeat period (default -1 means no repeat).
        self.callrepeatperiod = callrepeatperiod
        
        # Set the verbosity to 0 by default.
        self.verbose = 0
        
        # Set disabling off by default.
        self.disabled = False
        
        # Set recording of call times by default.
        self.reccalltimes = True
        
        # Increment the instance count.
        TimedCallback.numCreated += 1
        TimedCallback.allCBs.append(self)
        
    def install(self):
        h.cvode.event(self.firstcalltime, self.cbexec)
        
    def cbexec(self):
        # Return immediately if the event is disabled.
        if self.disabled:
            return
            
        # If verbose > 0, show the event happening.
        if self.verbose > 0:
            print '%s() triggered at t = %f' % (self.instlabel, h.t)
            
        # Call the functional "payload."
        self.thecb()
        
        # Have the simulation record the event time if recording is turned on.
        if self.reccalltimes:
            self.thesim.recordcallbackevent('%scbs' % self.instlabel)
        
        # Schedule the next call of the callback.
        if self.callrepeatperiod > 0:
            h.cvode.event(h.t + self.callrepeatperiod, self.cbexec)
            
    def disable(self):
        self.disabled = True
        
    def enable(self):
        self.disabled = False    
        
    def turnonreccalltimes(self):
        self.reccalltimes = True
 
    def turnoffreccalltimes(self):
        self.reccalltimes = False
        
