"""
unitgroups.py -- module for classes for groups of neural units

Classes for groups of neural units:
  UnitGroup (base class)
      NetStimGroup
      IntFire1Group
      
Initial design and implementation: George Chadderdon, 2013
 
Last update: 6/27/13 (georgec)
"""

# Import hoc object from neuron package.
from neuron import h

# Import numpy.
import numpy

# Import neural unit module.
import neuralunit as nun

# Load the stimulus class module.
import stim

class UnitGroup(object):
    """
    Generic group of neural units for the model.  This can be a mixed group 
    consisting of several-typed units.

    Methods:
      __init__(self): constructor
      __iter__(self): returns the iterator for the encapsulated list     
      __len__(self): gets the size of the group
      __getitem__(self, index): allows array indexing of the group      
      getunitbyinstlabel(self, label): gets the first unit matching the instance
        label
      getunitsbyclasslabel(self, label): makes a new UnitGroup containing 
        units in the old group that match by classlabel
      append(self,theunit): appends a unit to the group
      disable(self): disables the units so no output comes out
      enable(self): enables or reenables the units' output      
      turnonrecspks(self): turns on the recording of spikes for all of the units
      turnoffrecspks(self): turns off the recording of spikes for all of the units     
      isrecording(self): returns whether we are recording spikes from the 
        units
      getrecspks(self): returns a numpy array of the recorded spike times
            
    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created
      allGroups (class): list of all instances     
      unitlist: units the group contains (references only)
      instlabel: instance label (usable for tagging)

    Usage:
    >>>ug1 = unitgroups.UnitGroup()
    """

    classlabel = 'UnitGroup'
    numCreated = 0
    allGroups = []
    
    def __init__(self, theunits=[], label=''):
        # Create an empty group as a Python list.
        self.unitlist = []
        
        # Add the units that were passed in.
        self.unitlist += theunits

        # Set the instance label.
        self.instlabel = label

        # Increment the instance count.
        UnitGroup.numCreated += 1
        UnitGroup.allGroups.append(self)
                
    def __iter__(self):
        return iter(self.unitlist)
        
    def __len__(self):
        return len(self.unitlist)
        
    def __getitem__(self, index):
        return self.unitlist[index]    
            
    def getunitbyinstlabel(self, label):
        for unit in self.unitlist:
            if unit.instlabel == label:
                return unit
                
    def getunitsbyclasslabel(self, label):
        # Create a new UnitGroup to hold the matches.
        ugrp = UnitGroup()
        
        # Put matches by class in the new UnitGroup.
        for unit in self.unitlist:      
            if unit.classlabel == label:
                ugrp.append(unit)
                
        # Return the group. 
        return ugrp
        
    def append(self,theunit):
        self.unitlist.append(theunit)
        
    def disable(self):
        for unit in self.unitlist:
            unit.disable()
            
    def enable(self):
        for unit in self.unitlist:
            unit.enable()
            
    def turnonrecspks(self):
        if not self.isrecording():      
            for unit in self.unitlist:
                unit.turnonrecspks()
        
    def turnoffrecspks(self):
        if self.isrecording:
            for unit in self.unitlist:
                unit.turnoffrecspks()
            
    def isrecording(self):
        # Return False if we see any unit that is not recording.
        for unit in self.unitlist:
            if not unit.isrecording():
                return False
                
        # Return True.  We made it through the whole list.
        return True 
        
    def getrecspks(self):
        # Give an error and exit if recording is not enabled.
        if not self.isrecording():
            print 'ERROR in %s UnitGroup: Recording not enabled.' %\
                self.instlabel      
            return numpy.array([])
              
        # Set up a raster list to hold numpy arrays for each of the unit's 
        # spikes.
        rasterlist = []
        for unit in self.unitlist:
            rasterlist.append(numpy.array(unit.getrecspks()))
            
        # Return a tuple for the raster list.
        return tuple(rasterlist)
                         
# NOTE: not sure if this subclass is really useful for not.
class NetStimGroup(UnitGroup):
    """
    Group of NetStimUnits for the model.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created

    Usage:
    >>>ug1 = unitgroups.NetStimGroup()
    """

    classlabel = 'NetStimGroup'
    numCreated = 0

    def __init__(self, theunits=[], groupsize=1, label='', tNTs=['GLU'], tinterval=10,\
        tnumber=5,tstart=0.0,tnoise=0.0):
        # Set superclass parameters with a null list.
        super(NetStimGroup,self).__init__([], label)

        # If we passed in units, add them as normal.
        if len(theunits) > 0:
            self.unitlist += theunits
        # Otherwise, make the units.
        for ind in range(groupsize):
            ilab = '%s_%d' % (label, ind)
            self.unitlist.append(stim.NetStimUnit(NTs=tNTs,interval=tinterval, \
                 number=tnumber, start=tstart, noise=tnoise, label=ilab))

        # Increment the instance count.
        NetStimGroup.numCreated += 1

# NOTE: not sure if this subclass is really useful for not.
class IntFire1Group(UnitGroup):
    """
    Group of IntFire1Units for the model.

    Methods:
      __init__(self): constructor

    Attributes:
      classlabel (class): text label for the class
      numCreated (class): number of instances created

    Usage:
    >>>ug1 = unitgroups.IntFire1Group()
    """

    classlabel = 'IntFire1Group'
    numCreated = 0

    def __init__(self, theunits=[], groupsize=1, label='', tNTs=['GLU'], ttau=10.0, \
        trefrac=5.0):
        # Set superclass parameters with a null list.
        super(IntFire1Group,self).__init__([], label)

        # If we passed in units, add them as normal.
        if len(theunits) > 0:
            self.unitlist += theunits
        # Otherwise, make the units.
        for ind in range(groupsize):
            ilab = '%s_%d' % (label, ind)
            self.unitlist.append(nun.IntFire1Unit(NTs=tNTs, tau=ttau, \
                refrac=trefrac, label=ilab))

        # Increment the instance count.
        IntFire1Group.numCreated += 1
